/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.config;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashSet;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import rasmus_torkel.config.exception.ConfigSyntaxException;
import rasmus_torkel.config.exception.ConfigIoException;
import rasmus_torkel.config.exception.ConfigReferenceException;
import rasmus_torkel.misc.Constants;
import rasmus_torkel.text.read.exception.ReadException;

/**
 * This class is for reading one or more configuration files. This class offers a number of
 * options as to what to do with the read data. The most useful of thing to do is to obtain
 * a {@link PropertiesView} object via
 * {@link #propertiesViewFromArgs(java.lang.String[], int, int)}. 
 * 
 * @author Rasmus Torkel
 *
 */
public class ConfigFileReader extends ConfigProcessor
{
    protected static final Pattern _includePattern = Pattern.compile("^.*\\&include\\s+(.*\\S)\\s*$");
    protected static final Pattern _callPattern = Pattern.compile("^.*\\&call\\s+(.*\\S)\\s*$");
    protected static final Pattern _nonLinePattern = Pattern.compile("^\\s*(\\#.*)?$");
    
    protected HashSet<String> _includeHs = new HashSet<String>();
    
    public
    ConfigFileReader(File file) throws ReadException
    {
        this(new File[]{file});
    }
    
    /**
     * Constructor.
     * @param fileNames Names of the files from which configuration is to be read.
     * @throws ConfigException
     */
    public
    ConfigFileReader(String[] fileNames) throws ReadException
    {
        this(fileNamesToFiles(fileNames), fileNamesToCheckedCanonicalFiles(fileNames));
    }

    private static File[]
    fileNamesToCheckedCanonicalFiles(String[] fileNames) throws ReadException
    {
        File[] files = fileNamesToFiles(fileNames);
        return filesToCheckedCanonicalFiles(files);
    }
    
    private static File[]
    fileNamesToFiles(String[] fileNames) throws ReadException
    {
        File[] files = new File[fileNames.length];
        for (int i = 0; i < files.length; i++)
        {
            files[i] = new File(fileNames[i]);
        }
        return files;
    }
    
    private static File[]
    filesToCheckedCanonicalFiles(File[] files) throws ReadException
    {
        File[] canonicalFiles = new File[files.length];
        for (int i = 0; i < canonicalFiles.length; i++)
        {
            try
            {
                canonicalFiles[i] = files[i].getCanonicalFile();
            }
            catch (IOException e)
            {
                throw new ConfigIoException("File " + files[i] + " can not be converted to canonical form ", e);
            }
            if (!canonicalFiles[i].exists())
            {
                throw new ConfigReferenceException("File " + files[i] + " does not exist");
            }
            if (canonicalFiles[i].isDirectory())
            {
                throw new ConfigReferenceException("File " + files[i] + " is a directory");
            }
            if (!canonicalFiles[i].isFile())
            {
                throw new ConfigReferenceException("File " + files[i] + " is not a normal file");
            }
        }
        return canonicalFiles;
    }
    
    /**
     * Constructor.
     * @param files The files from which configuration is to be read.
     * @throws ReadException
     */
    public
    ConfigFileReader(File[] files) throws ReadException
    {
        this(files, filesToCheckedCanonicalFiles(files));
    }
    
    private
    ConfigFileReader(File[] originalFiles,
                     File[] canonicalFiles) throws ReadException
    {
        super(new PropertySourceDescription(originalFiles));
        for (int i = 0; i < canonicalFiles.length; i++)
        {
            ConfigInvokeStack invokeStack = new ConfigInvokeStack();
            readFile(originalFiles[i], canonicalFiles[i], invokeStack);
        }
    }
    
    private void
    readFile(File originalFile, File canonicalFile, ConfigInvokeStack invokeStack) throws ReadException
    {
        int lineNr = 0;
        FileReader fReader;
        try
        {
            fReader = new FileReader(canonicalFile);
        }
        catch (FileNotFoundException e)
        {
            throw new ConfigIoException("Making file reader", canonicalFile, e);
        }
        BufferedReader bReader = new BufferedReader(fReader);
        try
        {
            String line = null;
            do
            {
                lineNr++;
                try
                {
                    line = bReader.readLine();
                }
                catch (IOException e)
                {
                    throw new ConfigIoException("Reading line", canonicalFile, e);
                }
                if (line != null)
                {
                    if (   !tryNonLine(line)
                        && !tryInclude(line, originalFile, canonicalFile, lineNr, invokeStack)
                        && !tryCall(line, originalFile, canonicalFile, lineNr, invokeStack)
                        && !tryParam(line, canonicalFile, lineNr))
                    {
                        throw new ConfigSyntaxException("line did not match any allowable pattern", canonicalFile, lineNr);
                    }
                }
            } while (line != null);
        }
        finally
        {
            try
            {
                bReader.close();
            }
            catch (IOException e)
            {
            }
        }
    }
    
    // comment or blank
    private boolean
    tryNonLine(String line)
    {
        Matcher nonLineMatcher = _nonLinePattern.matcher(line);
        if (nonLineMatcher.matches())
        {
            return true;
        }
        return false;
    }
    
    private boolean
    tryInclude(String            line,
               File              originalReferencingFile,
               File              canonicalReferencingFile,
               int               lineNr,
               ConfigInvokeStack invokeStack) throws ReadException
    {
        Matcher includeMatcher = _includePattern.matcher(line);
        if (!includeMatcher.matches())
        {
            return false;
        }
        String fileName = includeMatcher.group(1);
        doFileReference(
                fileName, originalReferencingFile, canonicalReferencingFile, lineNr, invokeStack, true);
        return true;
    }
    
    private boolean
    tryCall(String            line,
            File              originalReferencingFile,
            File              canonicalReferencingFile,
            int               lineNr,
            ConfigInvokeStack invokeStack) throws ReadException
    {
        Matcher callMatcher = _callPattern.matcher(line);
        if (!callMatcher.matches())
        {
            return false;
        }
        String fileName = callMatcher.group(1);
        doFileReference(
                fileName, originalReferencingFile, canonicalReferencingFile, lineNr, invokeStack, false);
        return true;
    }
    
    private void
    doFileReference(String            fileName,
                    File              originalReferencingFile,
                    File              canonicalReferencingFile,
                    int               lineNr,
                    ConfigInvokeStack invokeStack,
                    boolean           avoidReread) throws ReadException
    {
        ConfigInvokeStack.Element stackElement =
                new ConfigInvokeStack.Element(
                        originalReferencingFile, canonicalReferencingFile, lineNr);
        invokeStack.add(stackElement);
        try
        {
            doFileReference(fileName, invokeStack, avoidReread);
        }
        finally
        {
            invokeStack.pop();
        }
    }
    
    private void
    doFileReference(String            fileName,
                    ConfigInvokeStack invokeStack,
                    boolean           avoidReread) throws ReadException
    {
        File originalFile = new File(fileName);
        File canonicalFile;
        if (originalFile.isAbsolute())
        {
            try
            {
                canonicalFile = originalFile.getCanonicalFile();
            }
            catch (IOException e)
            {
                throw new ConfigIoException("converting to canonical file", originalFile, e);
            }
        }
        else
        {
            canonicalFile = resolveFileReference(fileName, invokeStack);
        }
        if (avoidReread)
        {
            String canonicalFileName = canonicalFile.toString();
            if (_includeHs.contains(canonicalFileName))
            {
                //System.out.println("Already did: " + canonicalFileName);
                return;
            }
            //System.out.println("First time: " + canonicalFileName);
            _includeHs.add(canonicalFileName);
        }
        readFile(originalFile, canonicalFile, invokeStack);
    }
    
    private File
    resolveFileReference(String            referencedFileName,
                         ConfigInvokeStack invokeStack) throws ReadException
    {
        File referencedFile = new File(referencedFileName);
        if (referencedFile.isAbsolute())
        {
            try
            {
                File referencedCanonicalFile = referencedFile.getCanonicalFile();
                return referencedCanonicalFile;
            }
            catch (IOException e)
            {
                throw new ConfigIoException("making canonical file from ", referencedFile, invokeStack, e);
            }
        }
        else
        {
            ConfigInvokeStack.Element referencing = invokeStack.last();
            File referencedFileInProgress = referencing._canonicalFile.getParentFile();
            StringTokenizer tokeniser = new StringTokenizer(referencedFileName, Constants.FILE_SEP);
            while (tokeniser.hasMoreTokens())
            {
                String nextPiece = tokeniser.nextToken();
                if (nextPiece.equals(".") || nextPiece.equals("."))
                { 
                }
                else if (nextPiece.equals(".."))
                {
                    referencedFileInProgress = referencedFileInProgress.getParentFile();
                    if (referencedFileInProgress == null)
                    {
                        throw new ConfigReferenceException(
                                "Relative file name " + referencedFileName +
                                invokeStack.referencedFromString() +
                                " contains to many .. components");
                    }
                }
                else
                {
                    referencedFileInProgress = new File(referencedFileInProgress, nextPiece);
                    if (!referencedFileInProgress.exists())
                    {
                        throw new ConfigReferenceException(
                                "Relative file name " + referencedFileName +
                                invokeStack.referencedFromString() +
                                " leads to file " + referencedFileInProgress +
                                " which does not exist");
                    }
                    if (!referencedFileInProgress.isDirectory() && tokeniser.hasMoreTokens())
                    {
                        throw new ConfigReferenceException(
                                "Relative file name " + referencedFileName +
                                invokeStack.referencedFromString() +
                                " leads to file " + referencedFileInProgress +
                                " which should be a directory but is not");
                    }
                }
            }
            if (!referencedFileInProgress.isFile())
            {
                throw new ConfigReferenceException(
                        "Relative file name " + referencedFileName +
                        invokeStack.referencedFromString() +
                        " combines with referencing file to file " + referencedFileInProgress +
                        " which should be a normal file but is not");
            }
            try
            {
                referencedFile = referencedFileInProgress.getCanonicalFile();
            }
            catch (IOException e)
            {
                // I am not sure that this can even happen, but it's better to be safe than sorry.
                throw new ConfigIoException(
                        "File name " + referencedFileName +
                        invokeStack.referencedFromString() +
                        ", leads to " + referencedFileInProgress +
                        " which can not be converted to canonical file",
                        e);
            }
        }
        return referencedFile;
    }
    
    /**
     * Initialises an instance from an array of configuration file names.
     * @param args             Contains the names of the configuration files. Typically,
     *                         they will come from the command line arguments.
     * @param minConfigFileQty throws ReadException if args.length is less than this
     * @param maxConfigFileQty throws ReadException if args.length is more than this
     * @return
     * @throws ReadException
     */
    public static ConfigFileReader
    fromArgs(String[] args,
             int      minConfigFileQty,
             int      maxConfigFileQty) throws ReadException
    {
        if (args.length < minConfigFileQty || args.length > maxConfigFileQty)
        {
            String qtyRangeString;
            if (minConfigFileQty == maxConfigFileQty)
            {
                qtyRangeString = "" + minConfigFileQty;
            }
            else if (minConfigFileQty + 1 == maxConfigFileQty)
            {
                qtyRangeString = minConfigFileQty + " or " + maxConfigFileQty;
            }
            else
            {
                qtyRangeString = minConfigFileQty + " to " + maxConfigFileQty;
            }
            throw new ConfigReferenceException(
                    "number of expected configuration files is " + qtyRangeString +
                    " but number of arguments is " + args.length);
        }
        ConfigFileReader reader = new ConfigFileReader(args);
        return reader;
    }
    
    /**
     * Initialises a PropertiesView object from an array of configuration files.
     * @param args             Contains the names of the configuration files. Typically,
     *                         they will come from the command line arguments.
     * @param minConfigFileQty throws ReadException if args.length is less than this
     * @param maxConfigFileQty throws ReadException if args.length is more than this
     * @return
     * @throws ReadException
     */
    public static PropertiesView
    propertiesViewFromArgs(String[] args,
                           int      minConfigFileQty,
                           int      maxConfigFileQty) throws ReadException
    {
        ConfigFileReader reader = fromArgs(args, minConfigFileQty, maxConfigFileQty);
        return reader.makePropertiesView();
    }
    
    public static PropertiesView
    propertiesViewFromFile(File file) throws ReadException
    {
        ConfigFileReader reader = new ConfigFileReader(file);
        return reader.makePropertiesView();
    }
}
