/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.config;

import java.io.File;
import java.util.ArrayList;

/**
 * Stack for file references. People who need to merely invoke configuration
 * reading and retrieval do not need to know this. It is only public because
 * it is referred to in the exception package.
 * 
 * @author Rasmus Torkel
 */
public class ConfigInvokeStack extends ArrayList<ConfigInvokeStack.Element>
{
    /**
     * Represents one file or file reference in the stack of invocations.
     */
    public static class Element
    {
        public final File _originalFile;
        public final File _canonicalFile;
        public final int  _lineNr;
        public
        Element(File originalFile,
                File canonicalFile,
                int  lineNr)
        {
            _originalFile = originalFile;
            _canonicalFile = canonicalFile;
            _lineNr = lineNr;
        }
        
        @Override
        public String
        toString()
        {
            String oString = _originalFile.toString();
            String cString = _canonicalFile.toString();
            if (oString.equals(cString))
            {
                return oString + ", " + _lineNr;
            }
            else
            {
                return oString + ", " + _lineNr + " (canonical name is " + cString + ")"; 
            }
        }
    }
    
    public void
    pop()
    {
        verifyStackNotEmpty();
        int lastIndex = size() - 1;
        //Element last = get(lastIndex);
        remove(lastIndex);
        //return last;
    }
    
    public Element
    last()
    {
        verifyStackNotEmpty();
        int lastIndex = size() - 1;
        Element last = get(lastIndex);
        return last;
    }
    
    private void
    verifyStackNotEmpty()
    {
        if (size() < 1)
        {
            throw new RuntimeException("Internal error, stack is unexpectedly empty");
        }
    }
    
    public String
    referencedFromString()
    {
        Element last = last();
        return " referenced from " + last._canonicalFile + ", line " + last._lineNr;
    }
    
    @Override
    public String
    toString()
    {
        StringBuilder buf = new StringBuilder();
        for (int i = size() - 1; i >= 0; i--)
        {
            Element element = get(i);
            buf.append(element);
            if (i > 0)
            {
                buf.append(" <- ");
            }
        }
        return buf.toString();
    }
    
    public String
    toMultiLineString()
    {
        StringBuilder buf = new StringBuilder();
        for (int i = size() - 1; i >= 0; i--)
        {
            Element element = get(i);
            buf.append("  ");
            buf.append(element);
            if (i > 0)
            {
                buf.append("\n");
            }
        }
        return buf.toString();
    }
}
