/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.config;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import rasmus_torkel.config.exception.ConfigSyntaxException;
import rasmus_torkel.misc.Constants;
import rasmus_torkel.text.read.CursoredString;

/**
 * This class performs all sorts of functionality for configuration except actually reading it.
 * 
 * @author Torkel
 *
 */
public class ConfigProcessor
{
    protected    ArrayList<ConfigParam>      _paramAl = new ArrayList<ConfigParam>();
    protected    HashMap<String,ConfigParam> _nameToParam = new HashMap<String, ConfigParam>();
    public final PropertySourceDescription   _sourceDesc;
    
    protected static final Pattern _paramPattern = Pattern.compile("^([^=]+)\\=(.*)?$");
    
    private static enum VarType
    {
        CONFIG,
        SYSTEM_PROP,
    }
    
    public
    ConfigProcessor(PropertySourceDescription sourceDesc)
    {
        _sourceDesc = sourceDesc;
    }
    
    protected boolean
    tryParam(String line,
             File   file,
             int    lineNr) throws ConfigSyntaxException
    {
        Matcher paramMatcher = _paramPattern.matcher(line);
        if (!paramMatcher.matches())
        {
            //System.out.println("no match: " + line);
            return false;
        }
        String lhs = paramMatcher.group(1);
        String rhs = paramMatcher.group(2);
        String name = evaluateExpression(lhs, file, lineNr);
        //System.out.println("name = \"" + name + "\"");
        String value = evaluateExpression(rhs, file, lineNr);
        //System.out.println("value = \"" + value + "\"");
        ConfigParam param = new ConfigParam(name, value, file, lineNr);
        _nameToParam.put(name, param);
        _paramAl.add(param);
        return true;
    }
    
    /**
     * Evaluates and expression to an actual value.
     * 
     * @param expression The expression to evaluate
     * @param file      The file the expression came from (used only for exceptions)
     * @param lineNr    The line number of the expression within the file (used only for exceptions)
     * @return
     * @throws ConfigSyntaxException
     */
    public String
    evaluateExpression(String  expression,
                       File    file,
                       int     lineNr) throws ConfigSyntaxException
    {
        if (expression == null)
        {
            return "";
        }
        expression = expression.trim();
        int sLength = expression.length();
        if (sLength >= 2 && expression.charAt(0) == '"' && expression.charAt(sLength - 1) == '"')
        {
            expression = expression.substring(1, sLength -1);
        }
        StringBuilder allBuf = new StringBuilder();
        StringBuilder buf = allBuf;
        StringBuilder varNameBuf = null;
        VarType varType = null;
        boolean varIsBraced = false;
        int varBraceLevel = 0;
        CursoredString cExpr = new CursoredString(expression);
        while (!cExpr.finished())
        {
            boolean varStart = false;
            if (cExpr.startsWith("\\n", true))
            {
                buf.append('\n');
            }
            else if (cExpr.startsWith("\\r", true))
            {
                buf.append('\r');
            }
            else if (cExpr.startsWith("\\t", true))
            {
                buf.append('\t');
            }
            else if (cExpr.startsWith("\\\\", true))
            {
                buf.append('\\');
            }
            else if (cExpr.startsWith("\\$", true))
            {
                buf.append('$');
            }
            else if (cExpr.startsWith("\\", true))
            {
                if (cExpr.finished())
                {
                    throw new ConfigSyntaxException(
                            "\\ at end of expression is invalid", file, lineNr);
                }
                else
                {
                    char ch = cExpr.consume();
                    throw new ConfigSyntaxException(
                            "\\" + ch + " is invalid, escape char \\ does not go with " + ch,
                            file, lineNr);
                }
            }
            else if (cExpr.startsWith("$$$", true))
            {
                throw new ConfigSyntaxException("$$$ is invalid", file, lineNr);
            }
            else if (cExpr.startsWith("$$", true))
            {
                varStart = true;
                varType = VarType.SYSTEM_PROP;
            }
            else if (cExpr.startsWith("$", true))
            {
                varStart = true;
                varType = VarType.CONFIG;
            }
            else if (varNameBuf != null && cExpr.startsWith("{", true))
            {
                varBraceLevel++;
                varIsBraced = true;
                buf.append("{");
            }
            else if (varNameBuf != null && cExpr.startsWith("}", true))
            {
                if (varBraceLevel > 0)
                {
                    varBraceLevel--;
                    buf.append("}");
                }
                else if (varIsBraced)
                {
                    derefVar(varNameBuf, varType, allBuf, file, lineNr);
                    varNameBuf = null;
                    varIsBraced = false;
                    buf = allBuf;
                }
            }
            else
            {
                char ch = cExpr.consume();
                buf.append(ch);
            }
            if (varStart)
            {
                if (varNameBuf != null)
                {
                    throw new ConfigSyntaxException(
                            "$ in variabe name is invalid", file, lineNr);
                }
                varNameBuf = new StringBuilder();
                buf = varNameBuf;
                if (cExpr.startsWith("{", true))
                {
                    varIsBraced = true;
                }
                else
                {
                    varIsBraced = false;
                }
                varBraceLevel = 0;
            }
        }
        if (varNameBuf != null)
        {
            if (varIsBraced)
            {
                throw new ConfigSyntaxException(
                        "Braced variable opened but not closed", file, lineNr);
            }
            else
            {
                derefVar(varNameBuf, varType, allBuf, file, lineNr);
            }
        }
        return allBuf.toString();
    }
    
    private void
    derefVar(StringBuilder varNameBuf,
             VarType       varType,
             StringBuilder allBuf,
             File          file,
             int           lineNr) throws ConfigSyntaxException
    {
        String varName = varNameBuf.toString();
        String value;
        if (varType == VarType.CONFIG)
        {
            ConfigParam param = _nameToParam.get(varName);
            if (param == null)
            {
                throw new ConfigSyntaxException(
                        "Reference to non-existing parameter " + varName, file, lineNr);
            }
            value = param._value;
        }
        else if (varType == VarType.SYSTEM_PROP)
        {
            value = System.getProperty(varName);
            if (value == null)
            {
                throw new ConfigSyntaxException(
                        "Reference to non-existing system property " + varName, file, lineNr);
            }
        }
        else
        {
            throw new RuntimeException("Internal error, varType = " + varType);
        }
        allBuf.append(value);
    }
    
    /**
     * If you really want, you can retrieve the value directly from this class instead of
     * retrieving the properties first.
     * But this function was actually created for test purposes.
     * @param name
     * @return value of parameter. Null for non-existing parameter.
     */
    public String
    getValue(String name)
    {
        ConfigParam param = _nameToParam.get(name);
        if (param == null)
        {
            return null;
        }
        return param._value;
    }
    
    /**
     * Makes a properties object representing the configuration. However, unless you have
     * some interface to satisfy that requires a simple properties object, consider using
     * {@link #makePropertiesView()} to get a more useful {@link PropertiesView} object.
     * @return A simple properties object representing the configuration.
     */
    public Properties
    makeProperties()
    {
        Properties properties = new Properties();
        for (ConfigParam param : _paramAl)
        {
            properties.setProperty(param._name, param._value);
        }
        return properties;
    }
    
    /**
     * Gives a representation of the configuration in the author's preferred format for
     * actually extracting data out of it.
     * @return An object designed for easy extraction of the data.
     */
    public PropertiesView
    makePropertiesView()
    {
        Properties properties = makeProperties();
        return new PropertiesView(properties, _sourceDesc);
    }
    
    /**
     * Makes a configuration string that encapsulates the final state of all properties.
     * If one were to write this string to a file, one would end up with a flat, fully
     * evaluated configuration file, equivalent to the original configuration.
     *  
     * @return The string encapsulating the configuration.
     */
    public String
    makeConfigString()
    {
        StringBuilder buf = new StringBuilder();
        for (ConfigParam alParam : _paramAl)
        {
            ConfigParam mapParam = _nameToParam.get(alParam._name);
            if (alParam == mapParam) // else alParam has been replaced
            {
                buf.append(alParam._name).append("=")
                        .append(alParam.valueForDisplay()).append(Constants.LINE_END);
            }
        }
        return buf.toString();
    }
    
    /**
     * Returns a string that describes all assignments to the properties including superseded ones.
     * This function is intended for debugging.
     * @return The trace
     */
    public String
    makeTrace()
    {
        StringBuilder buf = new StringBuilder();
        for (ConfigParam alParam : _paramAl)
        {
            ConfigParam mapParam = _nameToParam.get(alParam._name);
            boolean isFinal = alParam == mapParam;
            if (isFinal)
            {
                buf.append("* ");
            }
            else
            {
                buf.append("  ");
            }
            String fileName;
            try
            {
                fileName = alParam._file.getCanonicalPath();
            }
            catch (IOException e)
            {
                fileName = alParam._file.getName();
            }
            buf.append(fileName).append(", ").append(alParam._lineNr).append(": ");
            buf.append(alParam._name).append("=").append(alParam._value).append(Constants.LINE_END);
        }
        return buf.toString();
    }
    
    /**
     * This method blends the functionality of {@link #makeConfigString()} and {@link #makeTrace()}.
     * If written to a file, the file would be a valid configuration file. Superseded assignments
     * are included as comments. It is however not as easy to read as {@link #makeConfigString()}
     * and {@link #makeTrace()}.
     * @return String version of properties
     */
    public String
    makeAnnotatedConfigString()
    {
        StringBuilder buf = new StringBuilder();
        for (ConfigParam alParam : _paramAl)
        {
            String fileName;
            try
            {
                fileName = alParam._file.getCanonicalPath();
            }
            catch (IOException e)
            {
                fileName = alParam._file.getName();
            }
            buf.append("# ").append(fileName).append(", ").append(alParam._lineNr).append(Constants.LINE_END);
            ConfigParam mapParam = _nameToParam.get(alParam._name);
            boolean isFinal = alParam == mapParam;
            if (!isFinal)
            {
                buf.append("# ");
            }
            buf.append(alParam._name).append("=").append(alParam.valueForDisplay()).append(Constants.LINE_END);
        }
        return buf.toString();
    }
}
