/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.config;

import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.Properties;

import rasmus_torkel.config.exception.PropertyDoesNotMatchTypeException;
import rasmus_torkel.config.exception.PropertyInternalBugException;
import rasmus_torkel.config.exception.PropertyNotFoundException;
import rasmus_torkel.misc.Constants;
import rasmus_torkel.text.read.exception.SuppliedClassNotEnumException;
import rasmus_torkel.text.read.exception.SuppliedClassNotStringMappedException;

/**
 * This class is a wrapper for the java.util.Properties object and has a multitude of
 * convenience functions. As the methods that follow are all named along similar lines,
 * they are not documented individually. Basically the names consist of what is returned,
 * followed by an N, E or D. N means return null if the parameter does not exist, E means
 * throw a {@link PropertyNotFoundException} if property does not exist. D means return
 * defaultValue if property does not exist.
 * <p>
 * All functions (other than string<i>X</i> where this is obviously not applicable), will
 * throw a {@link PropertyDoesNotMatchTypeException} if there is a value but it can't be
 * converted to the type.
 * <p>
 * Two sets of methods, which both use reflection, deserve special mention.
 * <ul>
 *   <li>The enum<i>X</i> methods return enum values. This requires an enum class to be
 *       supplied from which the methods look up the allowable values.</li>
 *   <li>The stringMapped<i>X</i> methods return instances of the supplied class. The
 *       supplied class must have a constructor which takes only a string parameter.
 *       The stringMapped<i>X</i> will throw a {@link PropertyDoesNotMatchType} exception
 *       if that constructor throws any exception.</li>
 * </ul> 
 */
public class PropertiesView
{
    private final Properties                _properties;
    public  final PropertySourceDescription _sourceDesc;
    
    public
    PropertiesView(Properties                properties,
                   PropertySourceDescription sourceDesc)
    {
        _properties = properties;
        _sourceDesc = sourceDesc;
    }
    
    @Override
    public String
    toString()
    {
        StringBuilder buf = new StringBuilder();
        buf.append(getClass().getSimpleName());
        buf.append(" loaded from ");
        buf.append(_sourceDesc);
        buf.append(Constants.LINE_END);
        Enumeration<?> keyEnum = _properties.propertyNames();
        while (keyEnum.hasMoreElements())
        {
            Object keyO = keyEnum.nextElement();
            if (keyO instanceof String)
            {
                String key = (String)keyO;
                String value = _properties.getProperty(key);
                buf.append(key);
                buf.append(" = ");
                buf.append(value);
                buf.append(Constants.LINE_END);
            }
        }
        return buf.toString();
    }
    
    public int
    propertyQty()
    {
        return _properties.size();
    }
    
    public String
    stringN(String name)
    {
        return _properties.getProperty(name);
    }
    
    public String
    stringE(String name) throws PropertyNotFoundException
    {
        String value = _properties.getProperty(name);
        if (value == null)
        {
            throw new PropertyNotFoundException(name, _sourceDesc);
        }
        return value;
    }
    
    public String
    stringD(String name,
            String defaultValue)
    {
        String value = _properties.getProperty(name);
        if (value == null)
        {
            return defaultValue;
        }
        return value;
    }
    
    public Boolean
    boolN(String name) throws PropertyDoesNotMatchTypeException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return null;
        }
        return stringToBool(name, valueS);
    }
    
    public boolean
    boolE(String name) throws PropertyNotFoundException, PropertyDoesNotMatchTypeException
    {
        String valueS = stringE(name);
        return stringToBool(name, valueS);
    }
    
    public boolean
    boolD(String  name,
          boolean defaultValue) throws PropertyDoesNotMatchTypeException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return stringToBool(name, valueS);
    }
    
    private boolean
    stringToBool(String name,
                 String valueS) throws PropertyDoesNotMatchTypeException
    {
        if (valueS.equalsIgnoreCase("true") || valueS.equalsIgnoreCase("yes"))
        {
            return true;
        }
        else if (valueS.equalsIgnoreCase("false") || valueS.equalsIgnoreCase("no"))
        {
            return false;
        }
        else
        {
            throw new PropertyDoesNotMatchTypeException(name, valueS, "boolean", _sourceDesc);
        }
    }
    
    public Integer
    intN(String name) throws PropertyDoesNotMatchTypeException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return null;
        }
        return stringToInt(name, valueS);
    }
    
    public int
    intE(String name) throws PropertyNotFoundException, PropertyDoesNotMatchTypeException
    {
        String valueS = stringE(name);
        return stringToInt(name, valueS);
    }
    
    public int
    intD(String name,
         int    defaultValue) throws PropertyDoesNotMatchTypeException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return stringToInt(name, valueS);
    }
    
    private int
    stringToInt(String name,
                String valueS) throws PropertyDoesNotMatchTypeException
    {
        try
        {
            return Integer.parseInt(valueS);
        }
        catch (Exception e)
        {
            throw new PropertyDoesNotMatchTypeException(name, valueS, "int", _sourceDesc);
        }
    }
    
    public Long
    longN(String name) throws PropertyDoesNotMatchTypeException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return null;
        }
        return stringToLong(name, valueS);
    }
    
    public long
    longE(String name) throws PropertyNotFoundException, PropertyDoesNotMatchTypeException
    {
        String valueS = stringE(name);
        return stringToLong(name, valueS);
    }
    
    public long
    longD(String name,
          long   defaultValue) throws PropertyDoesNotMatchTypeException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return stringToLong(name, valueS);
    }
    
    private long
    stringToLong(String name,
                 String valueS) throws PropertyDoesNotMatchTypeException
    {
        try
        {
            return Long.parseLong(valueS);
        }
        catch (Exception e)
        {
            throw new PropertyDoesNotMatchTypeException(name, valueS, "long", _sourceDesc);
        }
    }
    
    public Float
    floatN(String name) throws PropertyDoesNotMatchTypeException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return null;
        }
        return stringToFloat(name, valueS);
    }
    
    public float
    floatE(String name) throws PropertyNotFoundException, PropertyDoesNotMatchTypeException
    {
        String valueS = stringE(name);
        return stringToFloat(name, valueS);
    }
    
    public float
    floatD(String name,
           float  defaultValue) throws PropertyDoesNotMatchTypeException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return stringToFloat(name, valueS);
    }
    
    private float
    stringToFloat(String name,
                  String valueS) throws PropertyDoesNotMatchTypeException
    {
        try
        {
            return Float.parseFloat(valueS);
        }
        catch (Exception e)
        {
            throw new PropertyDoesNotMatchTypeException(name, valueS, "float", _sourceDesc);
        }
    }
    
    public Double
    doubleN(String name) throws PropertyDoesNotMatchTypeException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return null;
        }
        return stringToDouble(name, valueS);
    }
    
    public double
    doubleE(String name) throws PropertyNotFoundException, PropertyDoesNotMatchTypeException
    {
        String valueS = stringE(name);
        return stringToDouble(name, valueS);
    }
    
    public double
    doubleD(String name,
            double  defaultValue) throws PropertyDoesNotMatchTypeException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return stringToDouble(name, valueS);
    }
    
    private double
    stringToDouble(String name,
                   String valueS) throws PropertyDoesNotMatchTypeException
    {
        try
        {
            return Double.parseDouble(valueS);
        }
        catch (Exception e)
        {
            throw new PropertyDoesNotMatchTypeException(name, valueS, "double", _sourceDesc);
        }
    }
    
    public File
    fileN(String name) throws PropertyDoesNotMatchTypeException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return null;
        }
        return stringToFile(name, valueS);
    }
    
    public File
    fileE(String name) throws PropertyNotFoundException, PropertyDoesNotMatchTypeException
    {
        String valueS = stringE(name);
        return stringToFile(name, valueS);
    }
    
    public File
    fileD(String name,
          File   defaultValue) throws PropertyDoesNotMatchTypeException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return stringToFile(name, valueS);
    }
    
    private File
    stringToFile(String name,
                 String valueS) throws PropertyDoesNotMatchTypeException
    {
        try
        {
            return new File(valueS);
        }
        catch (Exception e)
        {
            throw new PropertyDoesNotMatchTypeException(name, valueS, File.class, _sourceDesc, e);
        }
    }
    
    public Enum<?>
    enumN(String   name,
          Class<?> enumClass) throws PropertyDoesNotMatchTypeException, SuppliedClassNotEnumException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return null;
        }
        return stringToEnum(name, valueS, enumClass);
    }
    
    public Enum<?>
    enumE(String   name,
          Class<?> enumClass) throws PropertyNotFoundException, PropertyDoesNotMatchTypeException, SuppliedClassNotEnumException
    {
        String valueS = stringE(name);
        return stringToEnum(name, valueS, enumClass);
    }
    
    public Enum<?>
    enumD(String   name,
          Enum<?>  defaultValue,
          Class<?> enumClass) throws PropertyDoesNotMatchTypeException, SuppliedClassNotEnumException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return stringToEnum(name, valueS, enumClass);
    }
    
    // With error handling
    private Enum<?>
    stringToEnum(String   name,
                 String   valueS,
                 Class<?> enumClass) throws PropertyDoesNotMatchTypeException, SuppliedClassNotEnumException
    {

        if (!enumClass.isEnum())
        {
            throw new SuppliedClassNotEnumException(
                    "Attempting to map property " + name + " = " + valueS + " but supplied class " +
                    enumClass.getName() + " is not an Enum class");
        }
        Enum<?> enumValue = stringToEnum(valueS, enumClass);
        if (enumValue == null)
        {
            throw new PropertyDoesNotMatchTypeException(name, valueS, enumClass, _sourceDesc);
        }
        return enumValue;
    }
    
    // Without much error handling
    private Enum<?>
    stringToEnum(String   s,
                 Class<?> enumClass)
    {

        try
        {
            Method method = enumClass.getMethod("values");
            Object returnObj = method.invoke(null);
            Enum<?>[] values = (Enum<?>[])returnObj;
            for (Enum<?> value : values)
            {
                String valueString = value.toString();
                if (valueString.equals(s))
                {
                    return value;
                }
            }
        }
        catch (Exception e)
        {
            // If reflection is coded properly, there will be no reflection based exception.
            throw new PropertyInternalBugException("Exception indicates internal error", e);
        }
        return null;
    }
    
    public Object
    stringMappedN(String   name,
                  Class<?> stringMappedClass) throws PropertyDoesNotMatchTypeException, SuppliedClassNotStringMappedException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return null;
        }
        return stringToMapped(name, valueS, stringMappedClass);
    }
    
    public Object
    stringMappedE(String   name,
                  Class<?> stringMappedClass) throws PropertyDoesNotMatchTypeException, SuppliedClassNotStringMappedException, PropertyNotFoundException
    {
        String valueS = stringE(name);
        return stringToMapped(name, valueS, stringMappedClass);
    }
    
    public Object
    stringMappedD(String   name,
                  Object   defaultValue,
                  Class<?> stringMappedClass) throws PropertyDoesNotMatchTypeException, SuppliedClassNotStringMappedException
    {
        String valueS = stringN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return stringToMapped(name, valueS, stringMappedClass);
    }
    
    private Object
    stringToMapped(String   name,
                   String   valueS,
                   Class<?> stringMappedClass) throws PropertyDoesNotMatchTypeException, SuppliedClassNotStringMappedException
    {
        Constructor<?> constructor;
        try
        {
            constructor = stringMappedClass.getConstructor(String.class);
        }
        catch (SecurityException e)
        {
            throw new SuppliedClassNotStringMappedException(
                    "Attempting to make string-mapped instance from " + name + " = " + valueS +
                    " but the String constructor of supplied class " +
                    stringMappedClass.getName() +
                    " is not accessible, reflection exception is: " + e);
        }
        catch (NoSuchMethodException e)
        {
            throw new SuppliedClassNotStringMappedException(
                    "Attempting to make string-mapped instance from " + name + " = " + valueS +
                    " but supplied class " + stringMappedClass.getName() +
                    " has no String constructor or possibly an invisible constructor, reflection exception is: " +
                    e);
        }
        Object stringMappedInstance;
        try
        {
            stringMappedInstance = constructor.newInstance(valueS);
        }
        catch (InvocationTargetException e)
        {
            throw new PropertyDoesNotMatchTypeException(name, valueS, stringMappedClass, _sourceDesc, e);
        }
        catch (Exception e)
        {
            throw new PropertyInternalBugException("Exception indicates internal error", e);
        }
        return stringMappedInstance;
    }
}
