/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.config.tool;

import java.io.File;
import java.io.FileNotFoundException;

import rasmus_torkel.config.ConfigFileReader;
import rasmus_torkel.config.ConfigProcessor;
import rasmus_torkel.config.exception.ConfigArgumentException;
import rasmus_torkel.misc.FileUtil;
import rasmus_torkel.text.read.exception.ReadException;

/**
 * Program which generates a single flat file from all the input files and whatever files they reference.
 * See {@link #main(String[])} for more details.
 * 
 * @author Rasmus Torkel
 */
public class ConfigFileSimplifier
{
    protected final File[] _inputFiles;
    protected final File   _outputFile;
    
    private
    ConfigFileSimplifier(String[] args) throws ReadException
    {
        if (args.length < 2)
        {
            throw new ConfigArgumentException(
                    "There must be at least two arguments of which all except the last are the names of input configuration files and the last one is the name of the file for simplified configuration");
        }
        _inputFiles = new File[args.length - 1];
        for (int i = 0; i < _inputFiles.length; i++)
        {
            _inputFiles[i] = new File(args[i]);
        }
        _outputFile = new File(args[_inputFiles.length]);
    }
    
    private String
    makeConfigString(ConfigProcessor processor)
    {
        return processor.makeConfigString();
    }
    
    private void
    simplify() throws ReadException, FileNotFoundException
    {
        ConfigFileReader reader = new ConfigFileReader(_inputFiles);
        String configString = makeConfigString(reader);
        FileUtil.stringToFile(configString, _outputFile);
    }
    
    private static void
    processThrowable(Throwable e)
    {
        if (e instanceof ReadException)
        {
            System.err.println(e.getMessage());
        }
        else if (e instanceof RuntimeException || e instanceof Error)
        {
            e.printStackTrace(System.err);
        }
        else
        {
            System.err.println(e.toString());
        }
    }
    
    /**
     * The program starts here. This is not a test method.
     * 
     * @param args At least two arguments are required.
     * All except the last are the names of the configuration files to be read.
     * The last is the name of the equivalent flat file to be generated.
     * The flat file to be generated should not already exist.
     */
    public static void
    main(String[] args)
    {
        try
        {
            ConfigFileSimplifier simplifier = new ConfigFileSimplifier(args);
            if (simplifier._outputFile.exists())
            {
                throw new ConfigArgumentException(
                        "Output file " + simplifier._outputFile +
                        ", already exists, to protect you from overwriting files by accident, this tool does not overwrite any existing file.");
            }
            simplifier.simplify();
        }
        catch (Throwable e)
        {
            processThrowable(e);
        }
    }
}
