/*
 * Copyright Rasmus Torkel 2015
 * Author    Rasmus Torkel 2015
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.misc;

import java.util.List;

/**
 * This class is for creating arrays generically. In Java Generics, the type is not
 * known at runtime, creating some complications when we need to create objects
 * and arrays generically. The way this is solved in the Rasmus Torkel system is
 * that generic classes that create arrays take a generic ArrayMaker. Every class
 * that needs an ArrayMaker needs to have its own ArrayMaker subclass. An every
 * subclass has a function for actually creating the array which is not generic.
 * <p>
 * It is strongly recommended that where possible, you declare each array maker
 * in a public static final field called ARRAY_MAKER in the class for which it
 * creates arrays. An anonymous inner class is a good way to do it.
 * <p>
 * There are separate subclasses for higher dimensions. If you require an array
 * maker that also creates two dimensional arrays, than your ArrayMaker should
 * be a subclass of ArrayMaker2D but still satisfy the conventions described
 * in the previous paragraph.
 * 
 * @author Rasmus Torkel
 *
 * @param <T>
 */
public abstract class ArrayMaker<T>
{
    public abstract T[]
    newArray(int size);
    
    public int
    supportedDimensionQty()
    {
        return 1;
    }
    
    public T[]
    fromList(List<T> list)
    {
        int size = list.size();
        T[] array = newArray(size);
        for (int i = 0; i < size; i++)
        {
            T element = list.get(i);
            array[i] = element;
        }
        return array;
    }
}
