/*
 * Copyright Rasmus Torkel 2015, 2016
 * Author    Rasmus Torkel 2015, 2016
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.misc;

public class CharUtil
{
    /**
     * It may sometimes be necessary to identify a character to the user.
     * However, it is not always sensible to put the character in single
     * quotes and append it to the message. That's because the character
     * may not always be represented by a glyph (symbol). For example,
     * control characters don't have glyphs associated with them. Not
     * only that, but incorporating a control character may make a mess of
     * the message. Also, even if a glyph is defined for a character, if it
     * is not an ASCII character, there is no guarantee that in your
     * environment the glyph can actually be drawn.
     * <p>
     * So this method will always display a hex code but for all ASCII
     * characters and many non-ASCII characters it will include the character
     * if it has a glyph (or is the plain space) and a word description
     * otherwise.
     * <p>
     * Examples of the kinds of strings that are returned:
     * <ul>
     *   <li>Record Separator/x1E</li>
     *   <li>' '/x20</li>
     *   <li>'A'/x41</li>
     *   <li>x300</li>
     *   <li>Ideographic Space/x3000</li>
     * </ul>
     * 
     * @param ch
     * @return
     */
    public static String
    unicodeCharToStringForMessage(char ch)
    {
        int chi = (int)ch;
        StringBuilder buf = new StringBuilder();
        String words = unicodeCharToWords(ch);
        if (words != null)
        {
            buf.append(words);
        }
        boolean isPrintable = unicodeCharHasGlyph(ch) || ch == ' ';
        if (isPrintable)
        {
            if (words != null)
            {
                buf.append('/');
            }
            buf.append('\'');
            buf.append(ch);
            buf.append('\'');
        }
        if (words != null || isPrintable)
        {
            buf.append('/');
        }
        buf.append('x');
        String chi16Str = Integer.toString(chi, 16).toUpperCase();
        buf.append(chi16Str);
        return buf.toString();
    }
    
    public static boolean
    isPlainPrintable(char ch)
    {
        return ch == ' ' || unicodeCharHasGlyph(ch);
    }
    
    /**
     * This function attempt to determine if a unicode character has a glyph (symbol associated with character).
     * @param ch The character for which this determination is made.
     * @return true means character has a glyph. false means the character may or may not have
     *         a glyph. At the time of writing, not all logic has been implemented and if in doubt
     *         we return false.
     */
    public static boolean
    unicodeCharHasGlyph(char ch)
    {
        int chi = (int)ch;
        if (chi <= 0x20)
        {
            return false;
        }
        else if (chi <= 0x7E)
        {
            return true;
        }
        else if (chi <= 0xA0)
        {
            return false;
        }
        else if (chi <= 0x24F)
        {
            return true;
        }
        return false;
    }
    
    // Returns description in words if author deemed it appropriate and got around
    // to writing the code for it.
    private static String
    unicodeCharToWords(char ch)
    {
        int chi = (int)ch;
        if (chi < 0x20) // C0 controls
        {
            switch (chi)
            {
            case 0: return "Null";
            case 1: return "Start of Heading";
            case 2: return "Start of Text";
            case 3: return "End of Text";
            case 4: return "End of Transmission";
            case 5: return "Enquiry";
            case 6: return "Acknowledge";
            case 7: return "Bell";
            case 8: return "Backspace";
            case 9: return "Horizontal Tab";
            case 0xA: return "Line Feed";
            case 0xB: return "Vertical Tab";
            case 0xC: return "Form Feed";
            case 0xD: return "Carriage Return";
            case 0xE: return "Shift Out";
            case 0xF: return "Shift In";
            case 0x10: return "Data Link Escape";
            case 0x11: return "Device Control 1";
            case 0x12: return "Device Control 2";
            case 0x13: return "Device Control 3";
            case 0x14: return "Device Control 4";
            case 0x15: return "Negative Acknowledge";
            case 0x16: return "Synchronous Idle";
            case 0x17: return "End of Transmission Block";
            case 0x18: return "Cancel Character";
            case 0x19: return "End of Medium";
            case 0x1A: return "Substitute";
            case 0x1B: return "Escape";
            case 0x1C: return "File Separator";
            case 0x1D: return "Group Separator";
            case 0x1E: return "Record Separator";
            case 0x1F: return "Unit Separator";
            }
        }
        else if (chi < 127) // plain space or ASCII glyph
        {
        }
        else if (chi < 160) // Delete and C1 controls
        {
            switch (chi)
            {
            case 0x7F: return "Delete";
            case 0x80: return "Padding";
            case 0x81: return "High Octet Preset";
            case 0x82: return "Break Permitted Here";
            case 0x83: return "No Break Here";
            case 0x84: return "Index";
            case 0x85: return "Next Line";
            case 0x86: return "Start of Selected Area";
            case 0x87: return "End of Selected Area";
            case 0x88: return "Character Tabulation Set";
            case 0x89: return "Character Tabulation with Justification";
            case 0x8A: return "Line Tabulation Set";
            case 0x8B: return "Partial Line Forward";
            case 0x8C: return "Partial Line Backwards";
            case 0x8D: return "Reverse Line Feed";
            case 0x8E: return "Single Shift Two";
            case 0x8F: return "Single Shift Three";
            case 0x90: return "Device Control String";
            case 0x91: return "Private Use 1";
            case 0x92: return "Private Use 2";
            case 0x93: return "Set Transit State";
            case 0x94: return "Cancel Character";
            case 0x95: return "Message Waiting";
            case 0x96: return "Start of Protected Area";
            case 0x97: return "End of Protected Area";
            case 0x98: return "Start of String";
            case 0x99: return "Single Graphic Character Introducer";
            case 0x9A: return "Single Character Introducer";
            case 0x9B: return "Control Sequence Introducer";
            case 0x9C: return "String Terminator";
            case 0x9D: return "Operation System Command";
            case 0x9E: return "Private Message";
            case 0x9F: return "Application Program Command";
            }
        }
        else
        {
            switch (chi)
            {
            case 0xA0: return "No-Break Space";
            case 0x1680: return "Ogham Space Mark";
            case 0x180E: return "Mongolian Vowel Separator";
            case 0x2000: return "EN Quad";
            case 0x2001: return "EM Quad";
            case 0x2002: return "EN Space";
            case 0x2003: return "EM Space";
            case 0x2004: return "Three-per-EM Space";
            case 0x2005: return "Four-per-EM Space";
            case 0x2006: return "Six-per-EM Space";
            case 0x2007: return "Figure Space";
            case 0x2008: return "Punctuation Space";
            case 0x2009: return "Thin Space";
            case 0x200A: return "Hair Space";
            case 0x200B: return "Zero Width Space";
            case 0x200C: return "Zero Width Non-Joiner";
            case 0x200D: return "Zero Width Joiner";
            case 0x2028: return "Line Separator";
            case 0x2029: return "Paragraph Separator";
            case 0x202F: return "Narrow No-Break Space";
            case 0x205F: return "Medium Mathematical Space";
            case 0x2060: return "Word Joiner";
            case 0x3000: return "Ideographic Space";
            case 0xFEFF: return "Zero Width Non-Breaking Space";
            }
        }
        return null;
    }
    
    public static String
    join(char[]        joinees,
         String        sep)
    {
        StringBuilder buf = new StringBuilder();
        join(buf, joinees, sep);
        return buf.toString();
    }
    
    public static void
    join(StringBuilder   buf,
         char[]          joinees,
         JoinPunctuation punctuation)
    {
        join(buf, joinees, punctuation._start, punctuation._separator, punctuation._end);
    }
    
    public static void
    join(StringBuilder buf,
         char[]        joinees,
         String        start,
         String        sep,
         String        end)
    {
        if (start != null)
        {
            buf.append(start);
        }
        join(buf, joinees, sep);
        if (end != null)
        {
            buf.append(end);
        }
    }
    
    public static void
    join(StringBuilder buf,
         char[]        joinees,
         String        sep)
    {
        for (int i = 0; i < joinees.length; i++)
        {
            if (i != 0)
            {
                buf.append(sep);
            }
            buf.append(joinees[i]);
        }
    }
}
