package rasmus_torkel.misc;

import java.io.File;
import java.io.PrintStream;

/**
 * Traverses a directory tree and invokes callback handler for performing whatever actions
 * need to be performed for each file.
 * 
 * @author Torkel
 */
public class DirectoryTraverser
{
    /**
     * Callback handler.
     * 
     * @author Torkel
     */
    public interface CallbackHandlerInterface
    {
        /**
         * Called for every file encountered, including the directory at the root of the traversed tree.
         * @param file  The file which was encountered
         * @param level From traversal root. Root itself is 0.
         */
        public void
        handleFile(File file,
                   int  level);
        
        public void
        handleEnd();
    }
    
    /**
     * Often there is nothing to do at the end, so for convenience we provided an abstract class with
     * empty function for handleEnd.
     */
    public static abstract class CallbackHandler implements CallbackHandlerInterface
    {
        public void
        handleEnd()
        {
        }
    }
    
    /**
     * Implementation which just prints file paths.
     */
    public static class FilePrinter extends CallbackHandler
    {
        private PrintStream _printStream;
        
        public
        FilePrinter(PrintStream printStream)
        {
            _printStream = printStream;
        }
        
        @Override
        public void
        handleFile(File file,
                   int  level)
        {
            _printStream.println(file.toString());
        }
    }
    
    /**
     * Implementation which prints file paths to standard output.
     */
    public static final FilePrinter STDOUT_PRINTER = new FilePrinter(System.out);
    
    public static void
    traverse(File                     root,
             CallbackHandlerInterface handler)
    {
        traverse(root, 0, handler);
        handler.handleEnd();
    }
    
    private static void
    traverse(File                     file,
             int                      level,
             CallbackHandlerInterface handler)
    {
        handler.handleFile(file, level);
        if (file.isDirectory())
        {
            File[] children = file.listFiles();
            for (File child : children)
            {
                traverse(child, level + 1, handler);
            }
        }
    }
    
    /**
     * Main function which prints file paths to standard output.
     * 
     * @param args Only one argument, the name/path of the root for the traversal.
     */
    public static void
    main(String[] args)
    {
        if (args.length != 1)
        {
            System.err.println("Need exacty one argument, the root for traversal");
        }
        File file = new File(args[0]);
        DirectoryTraverser.traverse(file, STDOUT_PRINTER);
    }
}
