/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.misc;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;
import java.util.StringTokenizer;

public class FileUtil
{
    public static final ArrayMaker<File> FILE_ARRAY_MAKER = new ArrayMaker<File>()
    {
        @Override
        public File[]
        newArray(int size)
        {
            return new File[size];
        }
    };
    
    public static void
    stringToFile(String s,
                 File   file) throws FileNotFoundException
    {
        PrintStream pStream = new PrintStream(file);
        pStream.print(s);
        pStream.close();
    }
    
    /**
     * Tries to find the absolute file corresponding to a file.
     * <>
     * For example, file ../x/x.txt with context directory /opt/y leads to absolute file /opt/x/x.txt.
     * @param givenFile  The file for which we want the absolute file.
     * @param contextDir If givenFile is relative, this is the directory towards which it is
     *                   relative.
     * @return The absolute file
     * @throws IOException 
     */
    public static File
    findFile(File givenFile,
             File contextDir) throws IOException
    {
        if (!contextDir.exists())
        {
            throw new FileNotFoundException("context directory " + contextDir + " does not exist");
        }
        if (!contextDir.isDirectory())
        {
            throw new IOException("context directory " + contextDir + " is not a directory");
        }
        if (givenFile.isAbsolute())
        {
            return givenFile;
        }
        File currentFile = contextDir;
        String givenFileName = givenFile.getName();
        StringTokenizer tokeniser = new StringTokenizer(givenFileName, "/\\");
        while (tokeniser.hasMoreTokens())
        {
            String namePiece = tokeniser.nextToken();
            if (namePiece.equals("."))
            {
            }
            else if (namePiece.equals(".."))
            {
                currentFile = currentFile.getParentFile();
                if (currentFile == null)
                {
                    throw new IOException(
                            "file " + givenFile + " goes back to far for contextDir " + contextDir);
                }
            }
            else
            {
                currentFile = new File(currentFile, namePiece);
            }
        }
        return currentFile;
    }
    
    /* Not yet tested
    public static boolean
    compareContents(File file1,
                    File file2) throws FileNotFoundException
    {
        long fileSize1 = file1.length();
        long fileSize2 = file2.length();
        if (fileSize1 != fileSize2)
        {
            System.out.println(file1 != file2);
            return false;
        }
        BufferedReader bReader1 = null;
        BufferedReader bReader2 = null;
        try
        {
            bReader1 = new BufferedReader(new FileReader(file1));
            bReader2 = new BufferedReader(new FileReader(file2));
            while (true)
            {
                int chi1;
                try
                {
                    chi1 = bReader1.read();
                }
                catch (Exception e)
                {
                    System.out.println("file " + file1 + " can't be read: " + e);
                    return false;
                }
                int chi2;
                try
                {
                    chi2 = bReader2.read();
                }
                catch (Exception e)
                {
                    System.out.println("file " + file2 + " can't be read: " + e);
                    return false;
                }
                if (chi1 != chi2)
                {
                    System.out.println(file1 != file2);
                    return false;
                }
                if (chi1 < 0)
                {
                    return true;
                }
            }
        }
        finally
        {
            if (bReader1 != null)
            {
                try
                {
                    bReader1.close();
                }
                catch (Exception e)
                {
                    System.out.println("Failed to close " + file1 + ": " + e);
                }
            }
            if (bReader2 != null)
            {
                try
                {
                    bReader2.close();
                }
                catch (Exception e)
                {
                    System.out.println("Failed to close " + file2 + ": " + e);
                }
            }
        }
    }*/
    
    /**
     * Makes a FileWriter. Two differences to calling new FileWriter(File):
     * Error messages are better. Exceptions are RuntimeExceptions.
     * @param file
     * @return
     */
    public static FileWriter
    makeFileWriter(File file)
    {
        verifyForCreation(file);
        try
        {
            return new FileWriter(file);
        }
        catch (IOException e)
        {
            throw new RuntimeException(e.getMessage(), e);
        }
    }
    
    /**
     * Performs some checks to establish that file can be created.
     * The checks are not guaranteed to be exhaustive.
     * The point is that the error messages are somewhat better than
     * the standard ones.
     * 
     * @param file The file to be checked
     * @throws RuntimeException if the checks fail.
     */
    public static void
    verifyForCreation(File file)
    {
        File canonicalFile;
        try
        {
            canonicalFile = file.getCanonicalFile();
        }
        catch (IOException e)
        {
            throw new RuntimeException("Could not get canonical path for " + file);
        }
        File parentDir = canonicalFile.getParentFile();
        if (!parentDir.exists())
        {
            String fileInMessageString = makeFileStringForErrorMessage(file, canonicalFile);
            File furthestNonExistentAncestor = furthestNonExistentAncestorForNonExistentCanonicalFile(parentDir);
            throw new RuntimeException("Can not create " + fileInMessageString + " because " + furthestNonExistentAncestor + " does not exist");
        }
        if (!parentDir.isDirectory())
        {
            String fileInMessageString = makeFileStringForErrorMessage(file, canonicalFile);
            throw new RuntimeException("Can not create " + fileInMessageString + " because " + parentDir + " is a file other than a directory");
        }
    }
    
    private static String
    makeFileStringForErrorMessage(File file,
                                  File canonicalFile)
    {
        String fileString = file.toString();
        String fileCanonicalFileString = canonicalFile.toString();
        String fileInMessageString;
        if (fileString.equals(fileCanonicalFileString))
        {
            fileInMessageString = "File " + fileString;
        }
        else
        {
            fileInMessageString = "File " + fileString + " with canonical path " + fileCanonicalFileString;
        }
        return fileInMessageString;
    }                          

    /*
    // work in progress
    private static void
    verifyDirectory(File dir)
    {
        if (!dir.exists())
        {
            File nonExistentAncestor = furthestNonExistentAncestorForNonExistentFile(dir);
            throw new RuntimeException("Directory " + dir + " does not exist because " + nonExistentAncestor + " does not exist");
        }
        if (!dir.isDirectory())
        {
            throw new RuntimeException(dir + " is not a directory");
        }
    }*/

    /*
    // work in progress
    private static File
    furthestNonExistentAncestorForNonExistentFile(File file)
    {
        File canFile;
        try
        {
            canFile = file.getCanonicalFile();
        }
        catch (IOException e)
        {
            throw new RuntimeException("Could not get canonical path for " + file);
        }
        return furthestNonExistentAncestorForNonExistentCanonicalFile(canFile);
    }*/
    
    // work in progress
    private static File
    furthestNonExistentAncestorForNonExistentCanonicalFile(File file)
    {
        File parent = file.getParentFile();
        if (parent == null)
        {
            return file;
        }
        if (parent.exists())
        {
            return file;
        }
        return furthestNonExistentAncestorForNonExistentCanonicalFile(parent);
    }
}
