/*
 * Copyright Rasmus Torkel 2016
 * Author    Rasmus Torkel 2016
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.misc;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

import rasmus_torkel.text.read.exception.ReadIoException;

public class Sha1Util
{
    private static final String SHA1 = "SHA-1";
    
    public static void
    main(String[] args)
    {
        System.out.println("Number of files: " + args.length);
        for (String fileName : args)
        {
            File file = new File(fileName);
            String sha1 = fileToSha1String(file);
            System.out.println();
            System.out.println((sha1.length() / 2) + " bytes SHA-1 checksum for file " + file + ":");
            System.out.println(sha1);
        }
    }
    
    public static String
    fileToSha1String(File file)
    {
        byte[] sha1Bytes = fileToSha1Bytes(file);
        return bytesToHexString(sha1Bytes);
    }
    
    public static byte[]
    fileToSha1Bytes(File file)
    {
        MessageDigest md;
        try
        {
            md = MessageDigest.getInstance(SHA1);
        }
        catch (NoSuchAlgorithmException e)
        {
            throw new RuntimeException("Can't get " + SHA1 + " MessageDigest", e);
        }
        FileInputStream input;
        try
        {
            input = new FileInputStream(file);
        }
        catch (FileNotFoundException e)
        {
            throw new ReadIoException("Could not open file " + file, e);
        }
        return toBytes(md, input);
    }
    
    private static byte[]
    toBytes(MessageDigest md,
            InputStream   stream)
    {
        try
        {
            int  length;
            byte buffer[] = new byte[8192];
            do
            {
                length = stream.read(buffer);
                if (length != -1)
                {
                    md.update(buffer, 0, length);
                }
            }
            while (length != -1);
            return md.digest();
        }
        catch (IOException e)
        {
            throw new ReadIoException("IOException while calculating SHA-1", e);
        }
        finally
        {
            try
            {
                stream.close();
            }
            catch (IOException e)
            {
                // Not much we can do
            }
        }
    }
    
    private static String
    bytesToHexString(byte[] bytes)
    {
        StringBuilder buf = new StringBuilder(bytes.length * 2);
        for (byte b : bytes)
        {
            int bint = (int)b;
            if (bint < 0)
            {
                bint += 256;
            }
            int high = bint >> 4;
            int low = bint & 0xF;
            char highCh = toHexChar(high);
            char lowCh = toHexChar(low);
            buf.append(highCh);
            buf.append(lowCh);
        }
        return buf.toString();
    }
    
    private static char
    toHexChar(int digit)
    {
        if (digit < 10)
        {
            return (char)(digit + (int)'0');
        }
        else
        {
            return (char)(digit - 10 + (int)'A');
        }
    }
}
