/*
 * Copyright Rasmus Torkel 2012
 * Author    Rasmus Torkel 2012
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.set.chars;

import rasmus_torkel.misc.CharUtil;
import rasmus_torkel.set.strings.JustifyingStringSet;

/**
 * Defines a character set. Technically, all characters have 16 bits, but when we talk about
 * n bit sets (where n is smaller than 16), we are implying that for all characters in the set
 * all bits other than the least significant n are 0.
 * 
 * @author Rasmus Torkel
 *
 */
public abstract class CharSet implements JustifyingStringSet
{
    /**
     * The highest character that the set contains. It is not a huge problem
     * if this field is higher than the actual maximum, but it must not be
     * smaller.
     */
    public final char   _max;
    
    /**
     * Number of bits needed to represent the character. For example, if _max is
     * at least 128 but less than 256, 8 bits are required.
     */
    public final int    _bitQty;
    
    /**
     * Free format character string which describes the set in one word or short phrase.
     * Also used for toString function.
     */
    public final String _label;
    
    public
    CharSet(char   max,
            String label)
    {
        _max = max;
        _label = label;
        int maxi = (int)max;
        int bitQty = 1;
        int limitExcl = 2;
        while (limitExcl <= maxi)
        {
            bitQty++;
            limitExcl *= 2;
        }
        _bitQty = bitQty;
    }
    
    public
    CharSet(int    bitQty,
            String label)
    {
        if (bitQty > 16)
        {
            throw new RuntimeException(
                    "Attempt to create CharSet for " + bitQty +
                    " characters, character can only be up to 16 bits");
        }
        if (bitQty < 1)
        {
            throw new RuntimeException(
                    "Attempt to create CharSet for " + bitQty +
                    " characters, characters must have at least one bit");
        }
        _bitQty = bitQty;
        _label = label;
        int maxi = (1 << bitQty) - 1;
        _max = (char)maxi;
    }
    
    /**
     * Determines whether a character is in the set. In this base class we treat all
     * characters from (char)0 to _max as being in the set. However, subclasses may
     * override this.
     * @param ch  The character being tested for set membership
     * @return    True if and only if ch is in the set
     */
    public abstract boolean
    isIn(char ch);

    /**
     * Determines whether all character of a string are in the set.
     * @param s  The string for which the characters are being tested for set membership
     * @return   True if and only if all characters of s are in the set
     */
    public boolean
    isIn(String s)
    {
        int sLength = s.length();
        for (int i = 0; i < sLength; i++)
        {
            char ch = s.charAt(i);
            if (!isIn(ch))
            {
                return false;
            }
        }
        return true;
    }
    
    @Override
    public String
    outReason(String s)
    {
        int sLength = s.length();
        for (int i = 0; i < sLength; i++)
        {
            char ch = s.charAt(i);
            if (!isIn(ch))
            {
                return "has character " + CharUtil.unicodeCharToStringForMessage(ch) +
                       " in position " + (i + 1) + " which is not in " + this;
            }
        }
        return null;
    }
    
    /**
     * Determines the position of the first character which is not in the set.
     * It is a precondition to this function that it has already been established that
     * the string has a character which is not in the set.
     * @param s
     * @return position of first character not in the set.
     */
    public int
    firstOutCharPosition(String s)
    {
        int sLength = s.length();
        for (int i = 0; i < sLength; i++)
        {
            char ch = s.charAt(i);
            if (!isIn(ch))
            {
                return i;
            }
        }
        throw new RuntimeException("Called firstOutCharPosition even though all chars are in");
    }
    
    @Override
    public String
    toString()
    {
        return _label;
    }
}
