/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2012/2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.text.encode;

import rasmus_torkel.set.chars.CharSet;
import rasmus_torkel.set.chars.DefinedCharSets;

/**
 * These options describe options for encoding characters during output.
 * The options are supposed to be format independent, so while this is suitable for
 * XML, there may be other supported formats.
 * <p>
 * Characters may fall into the following, potentially overlapping, categories.
 * <ul>
 *   <li>The set of all characters for which outputting is supported. Any attempt to
 *       write a character outside the set should trigger an exception.
 *   </li>
 *   <li>What might be considered the core set. This may vary somewhat but the default
 *       is letters (not including special non-English letters like umlauts), digits
 *       and other symbols on a standard English-language keyboards, as well as a simple
 *       space. Each of those characters will fit into 7 bits
 *       (characters which take up more than 7 bits typically don't fit into
 *       a byte, even 8 bit character often don't as the 8th bit in a byte is often a flag
 *       to indicate whether the character is multi-byte). The core set should be a subset of the
 *       set of all supported characters.</li>
 *   </li>
 *   <li>Supported characters outside the core set. Such characters would probably be
 *       be represented as a unicode, whether the text format is XML or not.
 *   </li>
 *   <li>White space character which are not the simple space character. If such characters
 *       are in the core set then depending on the escapeNonSimpleWhiteSpace options, they should
 *       also be escaped, though not necessarily (depending on the data format) the same
 *       way as non-core characters. For example, and this does NOT apply to XML, line feed
 *       might be \n while non-core characters are may be represented as unicodes.
 *   </li>
 *   <li>
 *       Characters with special meaning in the data format, for example the less than and
 *       ampersand characters in XML. As the same instance of this class is supposed to
 *       be usable for different data formats, such characters are NOT represented here at
 *       all.
 *   </li>
 * </ul>
 * 
 * So a typical algorithm for encoding characters might go like this:
 * <ul>
 *   <li>If the character is unsupported, throw exception</li>
 *   <li>Else if the character has special meaning for the data format (eg ampersand in XML),
 *       which is determined without the assistance of this class, encode suitably.
 *   </li>
 *   <li>Else if the character is a non-core character, apply non-core character treatment.
 *   </li>
 *   <li>Else if the character is a space character other than simple space and the option
 *       requires it, apply suitable treatment, which, depending on the data format, may
 *       differ from the non-core encoding.
 *   </li>
 *   <li>Else write the character the regular way</li>
 * 
 * @author Rasmus Torkel
 *
 */
public class CharEncodeOptions
{
    public final CharSet           _coreChars;
    public final CharSet           _allChars;
    public final SpaceEscapeOption _spaceEscapeOption;
    
    public static final CharSet           DEFAULT_CORE_CHARS = DefinedCharSets.ENGLISH_KEYBOARD_NON_CONTROL;
    public static final CharSet           DEFAULT_ALL_CHARS = DefinedCharSets.BITS_16_NON_NULL;
    public static final SpaceEscapeOption DEFAULT_SPACE_ESCAPE_OPTION = SpaceEscapeOption.ESCAPE_NON_SIMPLE;
    
    public static final CharEncodeOptions DEFAULT = new CharEncodeOptions(DEFAULT_CORE_CHARS, DEFAULT_ALL_CHARS, DEFAULT_SPACE_ESCAPE_OPTION);
    public static final CharEncodeOptions DONT_ESCAPE_SPACE = new CharEncodeOptions(DefinedCharSets.ENGLISH_KEYBOARD_GENERAL_TEXT, DEFAULT_ALL_CHARS, SpaceEscapeOption.DONT_ESCAPE);
    public static final CharEncodeOptions ENGLISH_KEYBOARD_GENERAL_TEXT = DEFAULT.diffAllChars(DefinedCharSets.ENGLISH_KEYBOARD_GENERAL_TEXT);
    public static final CharEncodeOptions ENGLISH_KEYBOARD_GENERAL_TEXT_DONT_ESCAPE = new CharEncodeOptions(DefinedCharSets.ENGLISH_KEYBOARD_GENERAL_TEXT, DefinedCharSets.ENGLISH_KEYBOARD_GENERAL_TEXT, SpaceEscapeOption.DONT_ESCAPE);
    
    public
    CharEncodeOptions(CharSet coreChars,
                      CharSet allChars)
    {
        this(coreChars, allChars, SpaceEscapeOption.DONT_ESCAPE);
    }
    
    public
    CharEncodeOptions(CharSet           coreChars,
                      CharSet           allChars,
                      SpaceEscapeOption spaceEscapeOption)
    {
        _coreChars = coreChars;
        _allChars = allChars;
        _spaceEscapeOption = spaceEscapeOption;
    }
    
    /**
     * Determines whether the character is supported at all.
     * @param ch The character to be assessed.
     * @return true if ch is supported and false otherwise.
     */
    public boolean
    isIn(char ch)
    {
        return _allChars.isIn(ch);
    }
    /*
    public boolean
    isIn(String s)
    {
        return _allChars.isIn(s);
    }*/
    
    /**
     * Determines whether the character is a non-core character. It is probably advisable
     * to check that the character is supported.
     * @param The character to be assessed.
     * @return false if ch is a core character and true otherwise (ie for unsupported and
     * for supported non-core characters).
     */
    public boolean
    isNonCore(char ch)
    {
        return !_coreChars.isIn(ch);
    }
    /*
    public boolean
    isNonCore(String s)
    {
        return !_coreChars.isIn(s);
    }*/
    
    /**
     * Determines whether the character needs special character.
     * This depends on both character and options.
     * @param ch The character to be assessed.
     * @return true if and only if special treatment is required.
     */
    public boolean
    needsSpecialSpaceTreatement(char ch)
    {
        if (_spaceEscapeOption == SpaceEscapeOption.DONT_ESCAPE)
        {
            return false;
        }
        else if (_spaceEscapeOption == SpaceEscapeOption.ESCAPE_NON_SIMPLE)
        {
            if (   Character.isWhitespace(ch)
                && ch != ' ')
            {
                return true;
            }
            else
            {
                return false;
            }
        }
        else
        {
            throw new RuntimeException(
                    "Internal error, unhandled SpaceEscapeOption " + _spaceEscapeOption);
        }
    }
    
    @Override
    public String
    toString()
    {
        return "{ " + _coreChars + ", " + _allChars + ", " + _spaceEscapeOption + " }";
    }
    
    /**
     * Returns an instance which has different core characters and is otherwise identical.
     * @param diffCoreChars The new set of core characters.
     * @return The new instance.
     */
    public CharEncodeOptions
    diffCoreChars(CharSet diffCoreChars)
    {
        return new CharEncodeOptions(diffCoreChars, _allChars, _spaceEscapeOption);
    }
    
    /**
     * Returns an instance which supports different characters and is otherwise identical.
     * @param diffAllChars The new set of all supported characters.
     * @return The new instance.
     */
    public CharEncodeOptions
    diffAllChars(CharSet diffAllChars)
    {
        return new CharEncodeOptions(_coreChars, diffAllChars, _spaceEscapeOption);
    }
    
    /**
     * Returns an instance which supports a different space escape option and is otherwise
     * identical.
     * @param diffSpaceEscapeOption
     * @return
     */
    public CharEncodeOptions
    diffSpaceEscapeOption(SpaceEscapeOption diffSpaceEscapeOption)
    {
        return new CharEncodeOptions(_coreChars, _allChars, diffSpaceEscapeOption);
    }
}
