/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2012/2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.text.encode;

/**
 * Encodes strings, taking care to properly encode special characters in the process.
 * 
 * @author Rasmus Torkel
 */
public abstract class TextEncoder
{
    protected final CharEncodeOptions _encodeOptions;
    protected final CharMaplet[]      _maplets;
    
    protected static final int[] _hexCoefficients = { 0x1000, 0x100, 0x10, 0x1 };
    
    /**
     * Constructor.
     * @param encodeOptions Populates {@link #_encodeOptions}
     * @param maplets       Populates {@link #_maplets}
     */
    protected
    TextEncoder(CharEncodeOptions encodeOptions,
                CharMaplet[]      maplets)
    {
        _encodeOptions = encodeOptions;
        _maplets = maplets;
    }
    
    /**
     * Does the obvious.
     * @param buf Where the characters representing ch go.
     * @param ch  The non-core character to encode.
     */
    protected abstract void
    encodeUniCodeChar(StringBuilder buf,
                      char          ch);
    
    /**
     * Does the obvious.
     * @param buf Where the characters representing ch go.
     * @param ch  The white space character to encode.
     */
    protected abstract void
    encodeWhiteSpaceChar(StringBuilder buf,
                         char          ch);

    /**
     * Encodes a string.
     * @param inText String to be encoded.
     * @return Result of the encoding.
     * @throws UnsupportedCharacterEncodeException if the character is not supported at all.
     */
    public String
    encode(String inText) throws UnsupportedCharacterEncodeException
    {
        int inTextLength = inText.length();
        StringBuilder buf = new StringBuilder(inTextLength + inTextLength / 4 + 10);
        encode(buf, inText);
        return buf.toString();
    }
    
    /**
     * Encodes a string.
     * @param buf Where the output of the encoding goes.
     * @param inText The string to be encoded.
     * @throws UnsupportedCharacterEncodeException if the character is not supported at all.
     */
    public void
    encode(StringBuilder buf,
           String        inText) throws UnsupportedCharacterEncodeException
    {
        int inTextLength = inText.length();
        for (int i = 0; i < inTextLength; i++)
        {
            char ch = inText.charAt(i);
            encode(buf, ch, i, inText);
        }
    }
    
    private void
    encode(StringBuilder buf,
           char          ch,
           int           positionInString,
           String        inText) throws UnsupportedCharacterEncodeException
    {
        if (!_encodeOptions.isIn(ch))
        {
            throw new UnsupportedCharacterEncodeException(
                    ch, positionInString, "Unidentified text piece", inText, _encodeOptions._allChars);
        }
        CharMaplet maplet = maplet(ch);
        if (maplet != null)
        {
            buf.append(maplet._replacement);
        }
        else if (_encodeOptions.isNonCore(ch))
        {
            encodeUniCodeChar(buf, ch);
        }
        else if (_encodeOptions.needsSpecialSpaceTreatement(ch))
        {
            encodeWhiteSpaceChar(buf, ch);
        }
        else
        {
            buf.append(ch);
        }
    }
    
    private CharMaplet
    maplet(char ch)
    {
        for (CharMaplet maplet : _maplets)
        {
            if (maplet._ch == ch)
            {
                return maplet;
            }
        }
        return null;
    }
    
    /**
     * Generates a hexadecimal representation of a character.
     * @param buf Where the output goes.
     * @param ch  The character to be represented as hexadecimal.
     */
    protected final void
    charToHex(StringBuilder buf,
              char          ch)
    {
        int chi = ch;
        int remainder = chi;
        boolean foundNonZeroCoefficient = false;
        for (int coefIndex = 0; coefIndex < _hexCoefficients.length; coefIndex++)
        {
            int coef = _hexCoefficients[coefIndex];
            int digit = remainder / coef;
            if (digit > 0 || coef == 1 || foundNonZeroCoefficient)
            {
                if (digit > 0)
                {
                    foundNonZeroCoefficient = true;
                }
                remainder = remainder % coef;
                int digitChi;
                if (digit < 10)
                {
                    digitChi = '0' + digit;
                }
                else
                {
                    digitChi = 'A' + digit - 10;
                }
                char digitCh = (char)digitChi;
                buf.append(digitCh);
            }
        }
    }
}
