/*
 * Copyright Rasmus Torkel 2014
 * Author    Rasmus Torkel 2014
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.text.read;

/**
 * A string being parsed.
 * <p>
 * This class has a notion of consuming and remainder. Characters are
 * consumed from the beginning of the string towards the end and whatever has not been consumed
 * is the remainder.
 * 
 * @author Torkel
 */
public class CursoredString
{
    public  final String _s;
    private final int    _sLength;
    private       int    _cursor;
    
    public
    CursoredString(String s)
    {
        _s = s;
        _sLength = _s.length();
        _cursor = 0;
    }
    
    /**
     * Determines whether the remainder of the string starts with a particular prefix.
     * @param prefix  The prefix to check
     * @param consume Whether to consume the prefix
     * @return 
     */
    public boolean
    startsWith(String  prefix,
               boolean consume)
    {
        int pLength = prefix.length();
        if (_cursor + pLength > _sLength)
        {
            return false;
        }
        for (int pCursor = 0; pCursor < pLength; pCursor++)
        {
            int tmpCursor = _cursor + pCursor;
            char sCh = _s.charAt(tmpCursor);
            char pCh = prefix.charAt(pCursor);
            if (sCh != pCh)
            {
                return false;
            }
        }
        if (consume)
        {
            _cursor += pLength;
        }
        return true;
    }
    
    public boolean
    startsWith(char    prefix,
               boolean consume)
    {
        if (_cursor >= _sLength)
        {
            return false;
        }
        char sCh = _s.charAt(_cursor);
        if (prefix != sCh)
        {
            return false;
        }
        if (consume)
        {
            _cursor++;
        }
        return true;
    }
    
    public boolean
    startsWith(char... prefixes)
    {
        for (char prefix : prefixes)
        {
            char ch = _s.charAt(_cursor);
            if (prefix == ch)
            {
                return true;
            }
        }
        return false;
    }
    
    public char
    consume()
    {
        if (finished())
        {
            throw new RuntimeException("Attempted to consume past end of string");
        }
        char ch = _s.charAt(_cursor);
        _cursor++;
        return ch;
    }
    
    public boolean
    finished()
    {
        return _cursor >= _sLength;
    }
}
