/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2012/2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.text.read;

import java.io.IOException;
import java.io.Reader;

import rasmus_torkel.text.read.exception.AlreadyClosedReadException;

/**
 * A wrapper class for reading characters, providing lots of convenient features, including
 * the ability to look ahead without consuming the characters.
 * <p>
 * Apart from the convenience features, this class differs from other readers
 * in one important philosophical aspect. The orthodox approach for character read functions
 * is to return integers instead of characters, where -1 means end of file. Here the approach
 * is different. We give the consumer the opportunity to check first that a character is available,
 * but the read functions return characters. So the typical sequence is check then read instead of
 * read then check then cast. This saves one step. Attempting operations that wrongly presume the
 * existence of an unconsumed character will result in a RuntimeException.
 * 
 * @author Rasmus Torkel
 *
 */
public class LookAheadReader
{
    private Reader  _reader;
    
    /**
     * The next four members together make up a circular buffer
     */
    private final char[]  _buffer;
    public  final int     _bufferSize;
    private int           _bufferCursor;        // index into buffer for next character.
    private int           _bufferUnconsumedQty; // number of characters in buffer but not yet consumed
    
    private final RWTextPosition      _textPos;
    
    /**
     * To allow the client access to text positions
     */
    public  final RWTextPosition.View _textPosView;
    
    private boolean _sawEnd;      // last char was read by this class but not client
    private boolean _consumedAll; // if we know that last char was read by this class and client.
    
    public
    LookAheadReader(Reader reader,
                    int    bufferSize)
    {
        if (bufferSize < 1)
        {
            throw new RuntimeException("Empty buffer specified");
        }
        _reader = reader;
        _bufferSize = bufferSize;
        _buffer = new char[bufferSize];
        _textPos = new RWTextPosition(1, 1);
        _textPosView = _textPos.view();
    }
    
    /**
     * Closes internal reader.
     * @throws IOException
     */
    public void
    close() throws IOException
    {
        if (_reader != null)
        {
            _reader.close();
            _reader = null;
        }
    }
    
    public boolean
    isOpen()
    {
        return _reader != null;
    }
    
    /**
     * Indicates whether there is another character that can be read by client.
     * @return true if and only if another character is available for the client.
     * @throws IOException
     */
    public boolean
    haveChar() throws IOException
    {
        tryNChars(1);
        return _bufferUnconsumedQty >= 1;
    }
    
    /**
     * Indicates whether there are another n characters that can be read by the client.
     * @param n The number of characters to check. Must not be bigger than the size
     *          of the buffer this class instance was constructed with.
     * @return true if and only if n characters are available
     * @throws IOException
     */
    public boolean
    haveNChars(int n) throws IOException
    {
        if (n > _bufferSize)
        {
            throw new RuntimeException(
                    "haveNChars(" + n + ") invoked but look ahead is only " + _bufferSize);
        }
        tryNChars(n);
        return _bufferUnconsumedQty >= n;
    }
    
    /**
     * Reads next unconsumed character, thereby consuming it. If you don't want RuntimeExceptions,
     * you should ensure beforehand that there is actually a character available.
     * Calling peep and advance is equivalent to calling read.
     */
    public char
    read() throws IOException
    {
        tryNChars(1);
        if (_consumedAll)
        {
            throw new RuntimeException("Attempted to consume past end of input");
        }
        char ch = _buffer[_bufferCursor];
        advanceBufferCursor();
        return ch;
    }
    
    /**
     * Looks at next unconsumed character, without consuming it. If you don't want RuntimeExceptions,
     * you should ensure beforehand that there is actually a character available.
     * Calling peep and advance is equivalent to calling read.
     */
    public char
    peep() throws IOException
    {
        tryNChars(1);
        if (_consumedAll)
        {
            throw new RuntimeException("Attempted to peep past end of input");
        }
        return _buffer[_bufferCursor];
    }
    
    /**
     * Advances read position. If you don't want RuntimeExceptions,
     * you should ensure beforehand that there is actually a character available to advance past.
     * Calling peep and advance is equivalent to calling read.
     */
    public void
    advance() throws IOException
    {
        tryNChars(1);
        if (_consumedAll)
        {
            throw new RuntimeException("Attempted to advance past end of input");
        }
        advanceBufferCursor();
    }
    
    /**
     * Advances read position. If you don't want RuntimeExceptions,
     * you should ensure beforehand that there is actually a character available to advance past.
     */
    public void
    advanceN(int n) throws IOException
    {
        if (!haveNChars(n))
        {
            throw new RuntimeException("Attempted to advance past end of input");
        }
        advanceBufferCursor(n);
    }
    
    /**
     * Determines whether the client has read or advances past all the clients. It is the
     * inverse of haveChar.
     * @return true if and only if there is no more character to read.
     * @throws IOException
     */
    public boolean
    consumedAll() throws IOException
    {
        tryNChars(1);
        return _consumedAll;
    }
    
    /**
     * Determines whether the remaining text to be read starts with string s.
     * s should not be longer than the buffer size with which this class instance was constructed.
     * @param s       The string for which we are determining whether the text starts with it
     * @param consume If true and the remaining text starts with s, then we advance past s
     * @return True if and only if remaining text starts with s
     * @throws IOException
     */
    public boolean
    startsWith(String  s,
               boolean consume) throws IOException
    {
        boolean result = startsWith(s);
        if (result && consume)
        {
            advanceBufferCursor(s.length());
        }
        return result;
    }

    /**
     * Determines whether the remaining text to be read starts with string s.
     * s should not be longer than the buffer size with which this class instance was constructed.
     * @param s       The string for which we are determining whether the text starts with it
     * @return True if and only if remaining text starts with s
     * @throws IOException
     */
    public boolean
    startsWith(String s) throws IOException
    {
        if (s == null)
        {
            throw new RuntimeException("startsWith called with null string");
        }
        int sLength = s.length();
        if (sLength > _bufferSize)
        {
            throw new RuntimeException(
                    "startsWith(" + s + "), string length is " + sLength +
                    " but look ahead is only " + _bufferSize);
        }
        tryNChars(sLength);
        if (_bufferUnconsumedQty < sLength)
        {
            return false;
        }
        for (int i = 0; i < sLength; i++)
        {
            char chS = s.charAt(i);
            char chB = bufferCharAt(i);
            if (chS != chB)
            {
                return false;
            }
        }
        return true;
    }
    
    /**
     * Determines whether the remaining text starts with character ch.
     * @param ch The character for which we are determining whether the rest of the
     *           text starts with it.
     * @return True if and only the remaining text starts with ch
     * @throws IOException
     */
    public boolean
    startsWith(char ch) throws IOException
    {
        tryNChars(1);
        if (_bufferUnconsumedQty < 1)
        {
            return false;
        }
        char chB = bufferCharAt(0);
        if (ch != chB)
        {
            return false;
        }
        return true;
    }
    
    /**
     * Advances text to before or just after any of the specified end characters or end of text.
     * @param endChars The character
     * @param buf      If this is non-null, the characters before the end character or
     *                 end of text are written in there. Does not place the end character in there.
     * @param consumeEndToken If true, advances past the end character, if any (but doesn't put it
     *                        into buf).
     * @return true if one of the specified end characters is reached, false if we reach end of
     *         text first.
     * @throws IOException
     */
    public boolean
    consumeUntil(char[]        endChars,
                 StringBuilder buf,
                 boolean       consumeEndToken) throws IOException
    {
        if (endChars == null)
        {
            throw new RuntimeException("consumeUntil called with null end token");
        }
        if (endChars.length == 0)
        {
            throw new RuntimeException("consumeUntil called with empty end token");
        }
        while (true)
        {
            for (char endChar : endChars)
            {
                if (startsWith(endChar))
                {
                    if (consumeEndToken)
                    {
                        advanceBufferCursor();
                    }
                    return true;
                }
            }
            if (_consumedAll)
            {
                return false;
            }
            else
            {
                char ch = read();
                if (buf != null)
                {
                    buf.append(ch);
                }
            }
        }
    }
    
    /**
     * Advances until just before or just after the end token occurs in the remaining
     * text or until end of text is reached.
     * @param endToken What we are looking for in the text.
     * @param buf if non-null, writes all characters consumed before end token (if it exists)
     *            or end of text into it.
     * @param consumeEndToken If true, advances past the end token as well (but never puts
     *                        the end token into buf). 
     * @return
     * @throws IOException
     */
    public boolean
    consumeUntil(String        endToken,
                 StringBuilder buf,
                 boolean       consumeEndToken) throws IOException
    {
        if (endToken == null)
        {
            throw new RuntimeException("consumeUntil called with null end token");
        }
        int endTokenLength = endToken.length();
        if (endTokenLength == 0)
        {
            throw new RuntimeException("consumeUntil called with empty end token");
        }
        while (true)
        {
            if (startsWith(endToken))
            {
                if (consumeEndToken)
                {
                    advanceBufferCursor(endTokenLength);
                }
                return true;
            }
            else if (_consumedAll)
            {
                return false;
            }
            else
            {
                char ch = read();
                if (buf != null)
                {
                    buf.append(ch);
                }
            }
        }
    }
    
    // We would not call this unless we had at least n unconsumed characters in the buffer.
    private void
    advanceBufferCursor(int n)
    {
        for (int i = 0; i < n; i++)
        {
            advanceBufferCursor();
        }
    }
    
    // We would not call this unless we had an unconsumed character in the buffer.
    private void
    advanceBufferCursor()
    {
        if (_bufferUnconsumedQty == 0)
        {
            throw new RuntimeException("Internal error");
        }
        if (_buffer[_bufferCursor] == '\n')
        {
            _textPos._lineNr++;
            _textPos._charNr = 1;
        }
        else
        {
            _textPos._charNr++;
        }
        _bufferCursor = (_bufferCursor + 1) % _bufferSize;
        _bufferUnconsumedQty -= 1;
        if (_sawEnd && _bufferUnconsumedQty == 0)
        {
            _consumedAll = true;
        }
    }
    
    // unconsumed character at specified logical position in the buffer.
    private char
    bufferCharAt(int index)
    {
        int internalIndex = (_bufferCursor + index) % _bufferSize;
        return _buffer[internalIndex];
    }
    
    /**
     * Attempts to ensure that the buffer contains n characters. If there isn't that
     * much remaining text, we put however many characters we can.
     * @param n The number of characters we aim to have in the buffer.
     * @throws IOException
     */
    public void
    tryNChars(int n) throws IOException
    {
        for (int i = _bufferUnconsumedQty; i < n && !_sawEnd; i++)
        {
            readInternally();
        }
    }
    
    /**
     * Advances past all whitespace.
     * @return the number of white space characters encountered before non-whitespace
     *         character or end of text if there is no more non-whitespace character.
     * @throws IOException
     */
    public int
    advancePastWhitespace() throws IOException
    {
        int wsQty = 0;
        while (haveChar())
        {
            char ch = peep();
            if (!Character.isWhitespace(ch))
            {
                return wsQty;
            }
            else
            {
                advance();
            }
        }
        return wsQty;
    }
    
    /**
     * Places a single character into the buffer unless we encounter end of text.
     * @throws IOException
     */
    private void
    readInternally() throws IOException
    {
        if (_reader == null)
        {
            throw new AlreadyClosedReadException("Attempt made to read after closing");
        }
        int chi = _reader.read();
        if (chi == -1)
        {
            _sawEnd = true;
            if (_bufferUnconsumedQty == 0)
            {
                _consumedAll = true;
            }
            close();
        }
        else
        {
            char ch = (char)chi;
            int charIndex = (_bufferCursor + _bufferUnconsumedQty) % _bufferSize;
            _buffer[charIndex] = ch;
            _bufferUnconsumedQty++;
        }
    }
}
