/*
 * Copyright Rasmus Torkel 2016
 * Author    Rasmus Torkel 2016
 * 
 * This version is from 2016-01-26
 * 
 * I, Rasmus Torkel, make this file available for all to use and share,
 * including for commercial purposes. You may change package name if you
 * wish. The following restrictions apply to this file and any copy,
 * whether the copy contains modifications or not:
 * -- You must not add to, modify, or delete the part of the header above the
 *    line consisting mostly of hyphen characters.
 * -- If you add any text to the header, it must go after the line consisting
 *    mostly of hyphens.
 * -- If you make any modification to the file, you must add a statement to
 *    that effect below the line consisting mostly of hyphens.
 * For the purposes of the above text, any file format conversion, for example
 * line ending change from \r\n to \n, does not count as a modification.
 * -------------------------------------------------------------------------
 *
 * Add text here if you like or if you modify the file.
 */
package rasmus_torkel.text.read;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This class is a tool for reading file, standard input and other text input.
 * It is modelled on how perl processes such input. The idea is to make it
 * possible to code java for file processing with a similar number of lines.
 * <p>
 * This class uses standard java.util.regex classes for regular expression
 * processing.
 * <p>
 * 
 * Below is some sample code:
 * <pre>
PerlStyleInput input = new PerlStyleInput("data/rasmus_torkel/test/misc/NameAndYear.txt");
System.out.println("Test with " + input);
while (input.read())
{
    if (input.matchC("^\\s*([A-Za-z ]*[A-Za-z])\\s+(\\d+)\\s*$"))
    {
        int groupCount = input.groupCount();
        String name = input.group(1);
        String year = input.group(2);
        System.out.println(groupCount + " groups, name is " + name + ", year is " + year);
    }
    else
    {
        System.out.println(input.messageForLine("Is not valid name and year"));
    }
}
input.close();
   </pre>
 * 
 * @author Rasmus Torkel
 */
public class PerlStyleInput
{
    /**
     * This class reads lines into this field. This field corresponds to $_ in perl.
     * You can modify it if you want, just like in perl.
     */
    public        String                  __;
    
    /**
     * Describes the nature of this input. See File constructor and parameter-less
     * constructor.
     */
    public  final String                  _label;
    
    private final BufferedReader          _bReader;
    private       int                     _lineNr = 0;
    private       Matcher                 _matcher;
    private final HashMap<String,Pattern> _patternCache = new HashMap<String,Pattern>();
    
    /**
     * Most basic constructor.
     * @param bReader For reading lines of input.
     * @param label   See {@link _label}
     */
    public
    PerlStyleInput(BufferedReader bReader,
                   String         label)
    {
        _bReader = bReader;
        _label = label;
    }
    
    /**
     * Reader constructor.
     * @param reader For input.
     * @param label  See {@link _label}
     */
    public
    PerlStyleInput(Reader reader,
                   String label)
    {
        this(new BufferedReader(reader), label);
    }
    
    /**
     * File constructor.
     * @param file Where input comes from.
     */
    public
    PerlStyleInput(File file)
    {
        this(makeFileReader(file), "file " + file);
    }
    
    // For converting IOException to RuntimeException
    private static FileReader
    makeFileReader(File file)
    {
        try
        {
            return new FileReader(file);
        }
        catch (FileNotFoundException e)
        {
            throw new RuntimeException("Could not make a FileReader for File " + file, e);
        }
    }
    
    /**
     * File name constructor.
     * @param fileName of the file from where input comes.
     */
    public
    PerlStyleInput(String fileName)
    {
        this(new File(fileName));
    }
    
    /**
     * Input stream constructor.
     * @param inputStream For the input
     * @param label       See {@link _label}
     */
    public
    PerlStyleInput(InputStream inputStream,
                   String      label)
    {
        this(new InputStreamReader(inputStream), label);
    }
    
    /**
     * This constructor takes input from standard input.
     */
    public
    PerlStyleInput()
    {
        this(System.in, "STDIN");
    }
    
    @Override
    public String
    toString()
    {
        return this.getClass().getSimpleName() + " for " + _label;
    }
    
    /**
     * Closes the buffered reader.
     */
    public void
    close()
    {
        try
        {
            _bReader.close();
        }
        catch (IOException e)
        {
            throw new RuntimeException("Failed to close " + _label, e);
        }
    }
    
    /**
     * Reads the next line. Small difference to Perl: line is returned without line terminators.
     * @return True if and only if the line was successfully read.
     */
    public boolean
    read()
    {
        try
        {
            __ = _bReader.readLine();
        }
        catch (IOException e)
        {
            throw new RuntimeException("IOException while reading line from " + _label, e);
        }
        if (__ == null)
        {
            return false;
        }
        _lineNr++;
        return true;
    }
    
    /**
     * Gets line number.
     * @return Quantity of successfully read lines. Also number of current line
     *         with first line being 1. 
     */
    public int
    lineNr()
    {
        return _lineNr;
    }
    
    /**
     * Matches the pattern and keeps the generated Matcher in case the invoker
     * wants to retrieve groups
     * @param pattern The pattern to match
     * @return true if and only the match succeeds
     */
    public boolean
    match(Pattern pattern)
    {
        _matcher = pattern.matcher(__);
        return _matcher.matches();
    }
    
    /**
     * Compiles regular expression into Pattern, matches and keeps the generated
     * Matcher in case the invoker wants to retrieve groups.
     * This method, unlike {@link matchC()}, does not cache the compiled pattern.
     * The F in the method name stands for "Forget pattern", in other words
     * "Do not cache pattern".
     * @param regularExpression For the pattern
     * @return true if and only the match succeeds
     */
    public boolean
    matchF(String regularExpression)
    {
        Pattern pattern = Pattern.compile(regularExpression);
        return match(pattern);
    }

    /**
     * Obtains Pattern, either from cache or by compiling, matches and keeps
     * Matcher in case the invoker wants to retrieve groups.
     * The C in the method name stands for "Cache pattern".
     * @param regularExpression For the pattern
     * @return true if and only the match succeeds
     */
    public boolean
    matchC(String regularExpression)
    {
        Pattern pattern = _patternCache.get(regularExpression);
        if (pattern == null)
        {
            pattern = Pattern.compile(regularExpression);
            _patternCache.put(regularExpression, pattern);
        }
        return match(pattern);
    }
    
    /**
     * Calls java.util.regex.Matcher.group on the last generated Matcher.
     * @param groupNumber
     * @return
     */
    public String
    group(int groupNumber)
    {
        return _matcher.group(groupNumber);
    }

    /**
     * Calls java.util.regex.Matcher.groupCount on the last generated Matcher
     * @param groupNumber
     * @return
     */
    public int
    groupCount()
    {
        return _matcher.groupCount();
    }
    
    /**
     * Adds input label and line number to message.
     * @param message Without input label and line number
     * @return Message with input label and line number
     */
    public String
    messageForLine(String message)
    {
        return _label + ", line " + _lineNr + ": " + message;
    }
}
