/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.text.type.strings;

public class Ip4Address implements Comparable<Ip4Address>
{
    public final String _string;
    public final int    _byte0;
    public final int    _byte1;
    public final int    _byte2;
    public final int    _byte3;
    public final long   _number;
    
    public static class InitialisationException extends RuntimeException
    {
        public
        InitialisationException(String message)
        {
            super(message);
        }
    }
    
    public
    Ip4Address(String string) throws InitialisationException
    {
        _string = string;
        int cursor = putCursorPastDigits(0, false);
        _byte0 = getComponent(0, cursor);
        int start = cursor + 1;
        cursor = putCursorPastDigits(start, false);
        _byte1 = getComponent(start, cursor);
        start = cursor + 1;
        cursor = putCursorPastDigits(start, false);
        _byte2 = getComponent(start, cursor);
        start = cursor + 1;
        cursor = putCursorPastDigits(start, true);
        _byte3 = getComponent(start, cursor);
        _number = makeNumber();
    }
    
    public
    Ip4Address(int byte0,
               int byte1,
               int byte2,
               int byte3) throws InitialisationException
    {
        _byte0 = byte0;
        _byte1 = byte1;
        _byte2 = byte2;
        _byte3 = byte3;
        checkByte(byte0);
        checkByte(byte1);
        checkByte(byte2);
        checkByte(byte3);
        _number = makeNumber();
        _string = makeString();
    }
    
    public
    Ip4Address(long number) throws InitialisationException
    {
        _number = number;
        if (_number < 0)
        {
            throw rejectNumber("number is negative");
        }
        if (_number > 0xffffffffL)
        {
            throw rejectNumber("number does not fit into 32 bits");
        }
        _byte0 = (int)((number & 0xff000000L) >> 24);
        _byte1 = (int)((number & 0xff0000L) >> 16);
        _byte2 = (int)((number & 0xff00L) >> 8);
        _byte3 = (int)(number & 0xffL);
        _string = makeString();
    }
    
    private int
    putCursorPastDigits(int     cursor,
                        boolean expectedEnd)
    {
        int sLength = _string.length();
        if (cursor >= sLength)
        {
            throw rejectString("string end where component was expected");
        }
        char ch = _string.charAt(cursor);
        if (!Character.isDigit(ch))
        {
            throw rejectString("unexpected non-digit character in position " + cursor);
        }
        while (true)
        {
            if (cursor >= sLength)
            {
                if (!expectedEnd)
                {
                    throw new InitialisationException("not enough components");
                }
                return cursor;
            }
            ch = _string.charAt(cursor);
            if (ch >= '0' && ch <= '9')
            {
                cursor++;
            }
            else
            {
                if (expectedEnd)
                {
                    throw rejectString("not finished after 4th component");
                }
                if (ch != '.')
                {
                    throw rejectString(
                            "contains invalid character '" + ch + "', numerical value of character is " + (int)ch);
                }
                return cursor;
            }
        }
    }
    
    private int
    getComponent(int startIndex,
                 int endCursor)
    {
        int digitQty = endCursor - startIndex;
        if (digitQty > 3)
        {
            throw rejectBigComponent(startIndex, endCursor);
        }
        int n = 0;
        for (int i = 0; i < digitQty; i++)
        {
            n *= 10;
            char ch = _string.charAt(startIndex + i);
            int digitValue = (int)ch - (int)'0';
            n += digitValue;
        }
        if (n > 255)
        {
            throw rejectBigComponent(startIndex, endCursor);
        }
        return n;
    }
    
    private InitialisationException
    rejectBigComponent(int startIndex,
                       int endCursor)
    {
        String component = _string.substring(startIndex, endCursor);
        return rejectString("component " + component + " is too big");
    }
    
    private InitialisationException
    rejectString(String reason)
    {
        return new InitialisationException(
                "\"" + _string + "\" is not a valid IPv4 address, " + reason);
    }
    
    private InitialisationException
    rejectByte(String reason)
    {
        return new InitialisationException(
                _byte0 + "." + _byte1 + "." + _byte2 + "." + _byte3 +
                " is not a valid IPv4 address, " + reason);
    }
    
    private InitialisationException
    rejectNumber(String reason)
    {
        return new InitialisationException(_number + " is not a valid IPv4 address, " + reason);
    }
    
    private void
    checkByte(int byte_)
    {
        if (byte_ < 0)
        {
            throw rejectByte("component " + byte_ + " is negative");
        }
        if (byte_ > 255)
        {
            throw rejectByte("component " + byte_ + " is greater than 255");
        }
    }
    
    private long
    makeNumber()
    {
        return   (long)_byte0 * 0x1000000L
               + (long)_byte1 * 0x10000L
               + (long)_byte2 * 0x100L
               + (long)_byte3;
    }
    
    /**
     * Puts all fields into a string. This function is intended for test purposes.
     * @return String
     */
    public String
    allFieldsToString()
    {
        return _string + "/" + _byte0 + " " + _byte1 + " " + _byte2 + " " + _byte3 + "/" + _number;
    }
    
    private String
    makeString()
    {
        return _byte0 + "." + _byte1 + "." + _byte2 + "." + _byte3;
    }
    
    @Override
    public String
    toString()
    {
        return _string;
    }
    
    @Override
    public boolean
    equals(Object other)
    {
        if (other == null)
        {
            return false;
        }
        else if (!(other instanceof Ip4Address))
        {
            return false;
        }
        Ip4Address address2 = (Ip4Address)other;
        {
            return _number == address2._number;
        }
    }

    @Override
    public int
    compareTo(Ip4Address address2)
    {
        if (_number < address2._number)
        {
            return -1;
        }
        else if (_number > address2._number)
        {
            return 1;
        }
        else
        {
            return 0;
        }
    }
    
    @Override
    public int
    hashCode()
    {
        return _string.hashCode();
    }
}
