/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.text.write;

/**
 * <p>
 * Represents a single indent for formatting functionality which requires indentation.
 * The main functionality that this class provides over and above a simple String is
 * validation. Then there is the obvious benefit of strict typing.
 * </p>
 * <p>
 * Validation is very important because without it the integrity of parsed output
 * can be compromised. In particular, if non-white-space characters are supplied, this
 * is likely to violate whatever syntax the output was meant to conform to. Having
 * said this, by requiring that indents must be empty or a single horizontal tab or
 * one or more simple spaces, this class is probably somewhat stricter than it needs
 * to be.
 * </p>
 * <p>
 * I was very tempted to make this an enumeration. If one assumes that there is no
 * need for indents bigger that eight spaces, then ten different values would cover it.
 * But then I thought that at some point I might want to relax the validation rules.
 * </p>
 * 
 * @author Rasmus Torkel
 *
 */
public class Indent
{
    public final String _indent;
    public final String _label;
    
    public static final Indent TAB =          new Indent("\t");
    public static final Indent TWO_SPACES =   new Indent("  ");
    public static final Indent THREE_SPACES = new Indent("   ");
    public static final Indent FOUR_SPACES =  new Indent("    ");
    public static final Indent DEFAULT = TWO_SPACES;
    
    /**
     * Constructor.
     * 
     * @param indent Empty string or single tab or one or more simple spaces.
     */
    public
    Indent(String indent)
    {
        _indent = indent;
        if (_indent.equals(""))
        {
            _label = "empty";
        }
        else if (_indent.equals("\t"))
        {
            _label = "tab";
        }
        else
        {
            int sLength = _indent.length();
            for (int i = 0; i < sLength; i++)
            {
                char ch = _indent.charAt(i);
                if (ch != ' ')
                {
                    throw new RuntimeException(
                            "Invalid indent \"" + _indent +
                            "\" because indent must be either a single horizontal tab or one or more simple spaces");
                }
            }
            if (sLength == 1)
            {
                _label = "single space";
            }
            else
            {
                _label = sLength + " spaces";
            }
        }
    }
    
    public String
    toString()
    {
        return _indent;
    }
}
