/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2012/2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.text.write;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;

public class IndentingLineSink
{
    private final Writer _writer; 
    public  final Indent _indent;
    public  final String _lineEnd;
    
    private boolean _atLineBoundary = true;
    private int     _level = 0;
    
    private int     _lineNr = 1;
    
    public
    IndentingLineSink()
    {
        this(IndentingLineSinkOptions.DEFAULT);
    }
    
    public
    IndentingLineSink(IndentingLineSinkOptions options)
    {
        this(new StringWriter(), options);
    }
    
    public
    IndentingLineSink(File file)
    {
        this(file, IndentingLineSinkOptions.DEFAULT);
    }
    
    public
    IndentingLineSink(File                     file,
                      IndentingLineSinkOptions options)
    {
        this(fileToWriter(file), options);
    }
    
    public
    IndentingLineSink(Writer writer)
    {
        this(writer, IndentingLineSinkOptions.DEFAULT);
    }
    
    public
    IndentingLineSink(Writer                   writer,
                      IndentingLineSinkOptions options)
    {
        _writer = writer;
        _indent = options._indent;
        _lineEnd = options._lineEndOption._lineEndStr;
    }
    
    private static FileWriter
    fileToWriter(File file)
    {
        try
        {
            return new FileWriter(file);
        }
        catch (IOException e)
        {
            throw new RuntimeException("Could not make FileWriter from file " + file, e);
        }
    }
    
    public void
    incrementLevel()
    {
        _level++;
    }
    
    public void
    decrementLevel()
    {
        if (_level == 0)
        {
            throw new RuntimeException("Attempt to decrement level when level is 0");
        }
        _level--;
    }
    
    public int
    level()
    {
        return _level;
    }
    
    /**
     * Writes a line with indentation
     * @param coreLine For correct indentation, it is assumed that this contains no new lines.
     */
    public void
    writeLine(String coreLine)
    {
        startLine();
        sinkPiece(coreLine);
        endLine();
    }
    
    public void
    sinkPiece(String piece)
    {
        if (_atLineBoundary)
        {
            startLine();
        }
        write(piece);
    }
    
    public void
    endLine()
    {
        write(_lineEnd);
        _atLineBoundary = true;
        _lineNr++;
    }
    
    private void
    startLine()
    {
        for (int i = 0; i < _level; i++)
        {
            write(_indent._indent);
        }
        _atLineBoundary = false;
    }
    
    private void
    write(String s)
    {
        try
        {
            _writer.write(s);
        }
        catch (IOException e)
        {
            throw new RuntimeException("Could not write to Writer", e);
        }
    }
    
    public void
    close()
    {
        try
        {
            _writer.close();
        }
        catch (IOException e)
        {
            throw new RuntimeException("Could not close Writer", e);
        }
    }
    
    /**
     * Returns the nominal line number.
     * Line number may not be accurate if there are line ends hidden in text.
     * @return
     */
    public int
    lineNr()
    {
        return _lineNr;
    }
    
    @Override
    public String
    toString()
    {
        return _writer.toString();
    }
}
