/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.text.write;

import java.util.Properties;

/**
 * <p>
 * This class encapsulates the options for the character or characters which constitute the end of
 * a line. There are the explicit options such as UNIX as well as the option to take the characters
 * from the system property (line.separator).
 * </p>
 * <p>
 * The explicit options of this enumeration and the values in the LineEndType enumeration
 * correspond one-to-one, as the line-endings one should be able to choose explicitly are the same
 * ones that are supported. But also, and that is the point of having this class in addition to
 * LineEndType, one might reasonably want to specify how the line ending to be used is to be
 * determined. There is only one scheme for this so far which is to get it from the appropriate
 * system property. That's the same scheme that is used in the java.io.PrintStream.println function,
 * although LineEndOption performs validation while PrintStream.println appeared not to do any, at
 * least not in my experiments.
 * </p>
 * 
 * @author Rasmus Torkel
 *
 */
public enum LineEndOption
{
    UNIX(LineEndType.UNIX),
    WINDOWS(LineEndType.WINDOWS),
    MAC(LineEndType.MAC),
    
    /**
     * <p>
     * This will end with the same characters as one of the other options. If the system
     * property @see LINE_END_SYSTEM_PROPERTY_NAME does not have characters matching any
     * of the other options, or if the property is missing, the line ending will be
     * defaulted.
     * </p>
     * <p>
     * As this type is immutable, once it is loaded, any subsequent changes to the
     * system property will have no effect.
     * </p>
     * <p>
     * The thought occurred to me to have a different system properties option for each
     * LineEndType to take that LineEndType as a default, but then I decided it was overkill
     * because the system property is always going to be present and correct unless somebody
     * deliberately makes it not so (as I did in my tests).
     * </p>
     */
    SYSTEM_PROPERTY();
    
    public final String      _label;
    
    /**
     * Same as _lineEndType._lineEndStr
     */
    public final String      _lineEndStr;
    
    public final LineEndType _lineEndType;
    
    /**
     * _lineEndType + " (" + explanation_how_lineEndType_was_assigned + ")"
     */
    public final String      _fullDescription;
    
    /**
     * The system property used for the line ending for the SYSTEM_PROPERTY option.
     */
    private static final String LINE_END_SYSTEM_PROPERTY_NAME = "line.separator";
    
    @Override
    public String
    toString()
    {
        // Returns _label and not _lineEndStr because this class is more about the
        // choice of line end rather than the actual line end.
        return _label;
    }
    
    // Constructor for explicit choices
    private
    LineEndOption(LineEndType lineEndType)
    {
        _label = lineEndType._label;
        _lineEndStr = lineEndType._lineEndStr;
        _lineEndType = lineEndType;
        _fullDescription = _label + " (explicit)";
    }
    
    // Constructor for the system property choice
    private
    LineEndOption()
    {
        LineEndType defaultType = LineEndType.UNIX;
        LineEndType propType = null;
        String explanation = null;
        Properties systemProps = System.getProperties();
        String lineEndStr = (String)systemProps.getProperty(LINE_END_SYSTEM_PROPERTY_NAME);
        if (lineEndStr == null)
        {
            explanation = "defaulted because system property " + LINE_END_SYSTEM_PROPERTY_NAME +
                          " was not defined";
        }
        else
        {
            LineEndType[] lineEndTypes = LineEndType.values();
            for (LineEndType lineEndType : lineEndTypes)
            {
                if (lineEndType._lineEndStr.equals(lineEndStr))
                {
                    propType = lineEndType;
                }
            }
            if (propType == null)
            {
                explanation = "defaulted because system property " + LINE_END_SYSTEM_PROPERTY_NAME +
                              " had a value which did not match any acceptable line end";
            }
        }
        _label = "System";
        if (propType != null)
        {
            _lineEndStr = propType._lineEndStr;
            _lineEndType = propType;
            explanation = "from system property " + LINE_END_SYSTEM_PROPERTY_NAME;
        }
        else
        {
            _lineEndStr = defaultType._lineEndStr;
            _lineEndType = defaultType;
        }
        _fullDescription = _lineEndType._label + " (" + explanation + ")";
    }
}
