package rasmus_torkel.xml_basic.read;
/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2012/2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;

import rasmus_torkel.set.doubles.AllDoubles;
import rasmus_torkel.set.doubles.DoubleSet;
import rasmus_torkel.set.floats.AllFloats;
import rasmus_torkel.set.floats.FloatSet;
import rasmus_torkel.set.ints.AllInts;
import rasmus_torkel.set.ints.IntSet;
import rasmus_torkel.set.longs.AllLongs;
import rasmus_torkel.set.longs.LongSet;
import rasmus_torkel.set.strings.StringSet;
import rasmus_torkel.text.read.ROTextPosition;
import rasmus_torkel.text.read.exception.InternalBugException;
import rasmus_torkel.text.read.exception.MissingNodeException;
import rasmus_torkel.text.read.exception.SuppliedClassNotEnumException;
import rasmus_torkel.text.read.exception.SuppliedClassNotStringMappedException;
import rasmus_torkel.text.read.exception.WrongNodeException;
import rasmus_torkel.xml_basic.read.exception.XmlAttributeDoesNotMatchTypeException;
import rasmus_torkel.xml_basic.read.exception.XmlAttributeNotInSetException;
import rasmus_torkel.xml_basic.read.exception.XmlMissingAttributeException;
import rasmus_torkel.xml_basic.read.exception.XmlNotOnlyTextException;
import rasmus_torkel.xml_basic.read.exception.XmlTextFieldDoesNotMatchTypeException;
import rasmus_torkel.xml_basic.read.exception.XmlTextFieldNotInSetException;
import rasmus_torkel.xml_basic.rw.TagNodeId;
import rasmus_torkel.xml_basic.rw.XmlAttribute;
import rasmus_torkel.xml_basic.rw.XmlNameSpace;

/**
 * This class represents an XML node that has a tag. It contains everything except the actual
 * parsing which is left to the implementation.
 * <p>
 * Because of the multitude of functions available for retrieval of data, we explain here how to
 * derive the functionality of a class from its signature.
 * <p>
 * Firstly, there are three possible courses of actions if a desired node or field is not present.
 * <ul>
 *   <li>Some functions simply return null. This is indicated by an "N" at the end of the
 *       function.</li>
 *   <li>Some functions throw an exception. This is indicated by an "E" at the end of the
 *       function.</li>
 *   <li>Some functions return a default value. This is indicated by a "D" at the end of the
 *       function and by the presence of a default value
 * </ul>
 * <p>
 * Then there are different possibilities with regard to the name spaces the node is in.
 * <ul>
 *   <li>Some functions are for node names not in any name space.</li>
 *   <li>Some functions allow for a relative name and a name space parameter.</li>
 *   <li>Some functions allow for relative name and name space parameter to supplied in
 *       a single object of type TagNodeId.</li>
 * </ul>
 * <p>
 * Then there are different possibilities with regard to the order of retrieval.
 * <ul>
 *   <li>Functions whose name starts with "anyOrder" will look for a corresponding child node
 *       anywhere within the node. However these functions may not be used if ordered retrieval
 *       is specified in the read options.</li>
 *   <li>Functions whose name starts with "next" are for ordered retrieval with some sort of
 *       cursor which advances.</li>
 *   <li>Functions whose name starts with "peep" are like "next" except that the cursor does
 *       not advance.</li> 
 * </ul>
 * <p>
 * Then there is retrieval of text field with automatic conversion to various types.
 * The following retrievals are supported with automatic conversion as necessary:
 * <ul>
 *   <li>Plain text</li>
 *   <li>Constants</li>
 *   <li>Primitives: int, long, float, double, boolean</li>
 *   <li>Enum classes. Enum instance retrieval requires reflection and cast.</li>
 *   <li>String Mapped classes. These are defined by contract. They must have a
 *       constructor with one parameter which is a String and their toString function
 *       produces an identical or at least equivalent result to the String it was
 *       constructed with.</li>
 * </ul>
 * <p>
 * For some numerical types and for plain text, it is also possible to specify a set that
 * the retrieved value must belong to (not the java.util set, but sets like for example
 * an integer range for integers). An exception will result if the value does not
 * belong to the set.
 * <p>
 * On closing: An underlying java.io.Reader object is used and it is good practice to
 * close this after use to avoid resource leaks. When you read all the way to the root
 * node, the reader is closed automatically. However, it is not good enough to just
 * read the last child node of the root node. It is necessary to look for a child
 * node after the last one and find that it is not there. XmlSyntaxExceptions also
 * will lead to automatic closing. And you can also close explicitly my invoking
 * {@link closeReader} on any node. SemanticExceptions do not lead to automatic closing.
 * 
 * @author Rasmus Torkel
 *
 */
public abstract class TagNode extends SupportedNode
{
    public final TagNodeId  _id;
    public final TagNode    _parent;
    
    /**
     * This may be something about where the XML comes from, for example the name of the file
     */
    public final String     _xmlContext;
    
    protected
    TagNode(ROTextPosition textPos,
            String         xmlContext,
            TagNodeId      id,
            TagNode        parent)
    {
        super(textPos);
        _id = id;
        _parent = parent;
        _xmlContext = xmlContext;
    }
    
    /**
     * This method closes the underlying reader. If the reader calls another method that requires
     * reading more XML on any node created using this reader, this will be an error condition.
     * It's probably safest not too call any method on the nodes after closing the reader.
     * However, closing multiple times is permitted.
     */
    public abstract void
    closeReader();
    
    /**
     * Checks if reader is open.
     * @return indication of whether underlying reader is open or not
     */
    public abstract boolean
    isReaderOpen();
    
    /**
     * By setting this, the user guarantees that subnodes of this node will be retrieved in order.
     * this means that subnode references can be dropped once we scan past the subnodes, allowing them to
     * be picked up by garbage collection.
     * <p>
     * This option is only for this node. See read options if you want to specify this for the entire input.
     */
    public abstract void
    setOrderedRetrieval();
    
    public abstract int
    attributeQty();
    
    public abstract XmlAttribute
    attribute(int index);
    
    public abstract String
    attributeValueN(String       relativeName,
                    XmlNameSpace nameSpace);
    
    public abstract TextNode
    nextTextNodeN();
    
    public abstract TagNode
    nextChildN();
    
    public abstract TagNode
    peepChildN();
    
    /**
     * After reading all the child nodes that are expected, call this
     * to verify that there are no additional child nodes.
     * <p>
     * If this function succeeds on the root node, it will close
     * the reader.
     */
    public abstract void
    verifyNoMoreChildren();
    
    public abstract void
    verifyNoMoreText();
    
    /**
     * Reads the rest of the node.
     * If invoked on the root node, this method will close the reader.
     */
    public abstract void
    readRest();
    
    public abstract TagNode
    anyOrderChildN(String       relativeName,
                   XmlNameSpace nameSpace);
    
    public abstract void
    verifyNoOtherChildren(TagNodeId... expectedIds);
    
    /**
     * Return the text position of the end of the node. This must not be invoked unless one
     * can be sure that the end of the node has actually been reached.
     * @return TextPosition of the end of the node
     */
    public abstract ROTextPosition
    endPosition();
    
    public boolean
    idMatches(String       relativeName,
              XmlNameSpace nameSpace)
    {
        if (nameSpace == XmlNameSpace.ANY)
        {
            return relativeName.equals(_id._relativeName);
        }
        else
        {
            return _id.isSame(relativeName, nameSpace);
        }
    }
    
    public boolean
    idMatches(TagNodeId id)
    {
        if (id._nameSpace == XmlNameSpace.ANY)
        {
            return id._relativeName.equals(_id._relativeName);
        }
        else
        {
            return _id.isSame(id._relativeName, id._nameSpace);
        }
    }
    
    public TagNode
    nextChildE()
    {
        TagNode child = nextChildN();
        if (child == null)
        {
            ROTextPosition endPos = endPosition();
            throw new MissingNodeException(
                    _xmlContext,
                    this + " ended at " + endPos + " before nextChildE was called");
        }
        return child;
    }
    
    public TagNode
    nextChildN(String simpleNodeName)
    {
        return nextChildN(simpleNodeName, null);
    }
    
    public TagNode
    nextChildE(String simpleNodeName)
    {
        return nextChildE(simpleNodeName, null);
    }
    /*
    public TagNode
    nextChildSameNsN(String simpleNodeName)
    {
        return nextChildN(simpleNodeName, _id._nameSpace);
    }
    
    public TagNode
    nextChildSameNsE(String simpleNodeName)
    {
        return nextChildE(simpleNodeName, _id._nameSpace);
    }
    
    public TagNode
    nextChildAnyNsE(String simpleNodeName)
    {
        TagNode child = nextChildE();
        if (!simpleNodeName.equals(child._id._relativeName))
        {
            throw new XmlWrongNodeException(_xmlContext,
                    "Attempted to read " + simpleNodeName + " under " + this +
                    " but got " + child + " with different relative name");
        }
        return child;
    }
    */
    public TagNode
    nextChildN(String       relativeName,
               XmlNameSpace nameSpace)
    {
        TagNode child = peepChildN();
        if (child == null)
        {
            return null;
        }
        if (!child._id.isSame(relativeName, nameSpace))
        {
            return null;
        }
        nextChildN();
        return child;
    }
    public TagNode
    nextChildN(TagNodeId id)
    {
        return nextChildN(id._relativeName, id._nameSpace);
    }
    
    public TagNode
    nextChildE(String       relativeName,
               XmlNameSpace nameSpace)
    {
        TagNode child = peepChildN();
        if (child == null)
        {
            String comparisonName = TagNodeId.makeComparisonName(relativeName, nameSpace);
            ROTextPosition endPos = endPosition();
            throw new MissingNodeException(_xmlContext,
                    "Attempted to read " + comparisonName + " under " + this +
                    " but reached node end at " + endPos);
        }
        if (!child.idMatches(relativeName, nameSpace))
        {
            String comparisonName = TagNodeId.makeComparisonName(relativeName, nameSpace);
            throw new WrongNodeException(_xmlContext,
                    "Attempted to read " + comparisonName + " under " + this +
                    " but got " + child._id.comparisonName() + " at " + child._textPos);
        }
        nextChildN();
        return child;
    }
    
    private String
    checkConstField(TagNode childNode,
                    String  expectedValue)
    {
        String value = childNode.onlyText();
        if (!value.equals(expectedValue))
        {
            throw new XmlTextFieldDoesNotMatchTypeException(_xmlContext,
                    childNode + " contains text field " + value +
                    " and not expected value " + expectedValue);
        }
        return value;
    }
    
    public String
    nextConstFieldN(String       relativeName,
                    XmlNameSpace nameSpace,
                    String       expectedValue)
    {
        TagNode childNode = nextChildN(relativeName, nameSpace);
        if (childNode == null)
        {
            return null;
        }
        return checkConstField(childNode, expectedValue);
    }
    
    public String
    nextConstFieldE(String       relativeName,
                    XmlNameSpace nameSpace,
                    String       expectedValue)
    {
        TagNode childNode = nextChildE(relativeName, nameSpace);
        return checkConstField(childNode, expectedValue);
    }
    
    public String
    nextConstFieldD(String       relativeName,
                    XmlNameSpace nameSpace,
                    String       expectedValue)
    {
        TagNode childNode = nextChildN(relativeName, nameSpace);
        if (childNode == null)
        {
            return expectedValue;
        }
        return checkConstField(childNode, expectedValue);
    }
    
    public String
    nextConstFieldN(String       name,
                    String       expectedValue)
    {
        return nextConstFieldN(name, null, expectedValue);
    }
    
    public String
    nextConstFieldE(String       name,
                    String       expectedValue)
    {
        return nextConstFieldE(name, null, expectedValue);
    }
    
    public String
    nextConstFieldD(String       name,
                    String       expectedValue)
    {
        return nextConstFieldD(name, null, expectedValue);
    }
    
    public String
    nextConstFieldN(TagNodeId id,
                    String    expectedValue)
    {
        return nextConstFieldN(id._relativeName, id._nameSpace, expectedValue);
    }
    
    public String
    nextConstFieldE(TagNodeId id,
                    String    expectedValue)
    {
        return nextConstFieldE(id._relativeName, id._nameSpace, expectedValue);
    }
    
    public String
    nextConstFieldD(TagNodeId id,
                    String    expectedValue)
    {
        return nextConstFieldD(id._relativeName, id._nameSpace, expectedValue);
    }
    
    public TagNode
    nextChildE(TagNodeId id)
    {
        return nextChildE(id._relativeName, id._nameSpace);
    }
    
    public String
    nextTextFieldN(String name)
    {
        return nextTextFieldN(name, (XmlNameSpace)null);
    }
    
    public String
    nextTextFieldE(String name)
    {
        return nextTextFieldE(name, (XmlNameSpace)null);
    }
    
    public String
    nextTextFieldD(String name,
                   String defaultValue)
    {
        return nextTextFieldD(name, null, defaultValue);
    }
    
    public String
    nextTextFieldN(String       relativeName,
                   XmlNameSpace nameSpace)
    {
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyText();
    }
    
    public String
    nextTextFieldE(String       relativeName,
                   XmlNameSpace nameSpace)
    {
        TagNode child = nextChildE(relativeName, nameSpace);
        return child.onlyText();
    }
    
    public String
    nextTextFieldD(String       relativeName,
                   XmlNameSpace nameSpace,
                   String       defaultValue)
    {
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyText();
    }
    
    public String
    nextTextFieldN(TagNodeId id)
    {
        return nextTextFieldN(id._relativeName, id._nameSpace);
    }
    
    public String
    nextTextFieldE(TagNodeId id)
    {
        return nextTextFieldE(id._relativeName, id._nameSpace);
    }
    
    public String
    nextTextFieldD(TagNodeId id,
                   String    defaultValue)
    {
        return nextTextFieldD(id._relativeName, id._nameSpace, defaultValue);
    }
    
    public String
    nextTextFieldN(String    name,
                   StringSet validSet)
    {
        return nextTextFieldN(name, null, validSet);
    }
    
    public String
    nextTextFieldE(String    name,
                   StringSet validSet)
    {
        return nextTextFieldE(name, null, validSet);
    }
    
    public String
    nextTextFieldD(String    name,
                   String    defaultValue,
                   StringSet validSet)
    {
        return nextTextFieldD(name, null, defaultValue, validSet);
    }
    
    public String
    nextTextFieldN(String       relativeName,
                   XmlNameSpace nameSpace,
                   StringSet    validSet)
    {
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyText(validSet);
    }
    
    public String
    nextTextFieldE(String       relativeName,
                   XmlNameSpace nameSpace,
                   StringSet    validSet)
    {
        TagNode child = nextChildE(relativeName, nameSpace);
        return child.onlyText(validSet);
    }
    
    public String
    nextTextFieldD(String       relativeName,
                   XmlNameSpace nameSpace,
                   String       defaultValue,
                   StringSet    validSet)
    {
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyText(validSet);
    }
    
    public String
    nextTextFieldN(TagNodeId id,
                   StringSet validSet)
    {
        return nextTextFieldN(id._relativeName, id._nameSpace, validSet);
    }
    
    public String
    nextTextFieldE(TagNodeId id,
                   StringSet validSet)
    {
        return nextTextFieldE(id._relativeName, id._nameSpace, validSet);
    }
    
    public String
    nextTextFieldD(TagNodeId id,
                   String    defaultValue,
                   StringSet validSet)
    {
        return nextTextFieldD(id._relativeName, id._nameSpace, defaultValue, validSet);
    }
    
    public String[]
    nextStringArrayN(String       arrayRelativeName,
                     XmlNameSpace arrayNameSpace,
                     String       elementRelativeName,
                     XmlNameSpace elementNameSpace)
    {
        TagNode arrayNode = nextChildN(arrayRelativeName, arrayNameSpace);
        if (arrayNode == null)
        {
            return null;
        }
        ArrayList<String> elementAl = new ArrayList<String>();
        boolean finished = false;
        while (!finished)
        {
            String element = arrayNode.nextTextFieldN(elementRelativeName, elementNameSpace);
            if (element == null)
            {
                finished = true;
            }
            else
            {
                elementAl.add(element);
            }
        }
        arrayNode.verifyNoMoreChildren();
        String[] elements = new String[elementAl.size()];
        elementAl.toArray(elements);
        return elements;
    }
    
    public String[]
    nextStringArrayD(String       arrayRelativeName,
                     XmlNameSpace arrayNameSpace,
                     String       elementRelativeName,
                     XmlNameSpace elementNameSpace,
                     String[]     defaultValue)
    {
        String[] value =
            nextStringArrayN(arrayRelativeName, arrayNameSpace,
                             elementRelativeName, elementNameSpace);
        if (value == null)
        {
            return defaultValue;
        }
        return value;
    }
    
    public String[]
    nextStringArrayD(String       arraySimpleName,
                     String       elementSimpleName,
                     String[]     defaultValue)
    {
        String[] value = nextStringArrayN(arraySimpleName, null, elementSimpleName, null);
        if (value == null)
        {
            return defaultValue;
        }
        return value;
    }
    
    public XmlAttribute[]
    sortedAttributes()
    {
        XmlAttribute[] attributes = attributes();
        Arrays.sort(attributes);
        return attributes;
    }
    
    /**
     * Returns all the attributes of the tag node. The array is a copy, so can be modified without
     * affecting TagNode.
     * 
     * @return attributes of the node
     */
    public XmlAttribute[]
    attributes()
    {
        XmlAttribute[] attributes = new XmlAttribute[attributeQty()];
        for (int i = 0; i < attributes.length; i++)
        {
            attributes[i] = attribute(i);
        }
        return attributes;
    }
    
    public String
    attributeValueN(String name)
    {
        return attributeValueN(name, (XmlNameSpace)null);
    }
    
    public String
    attributeValueE(String name)
    {
        String value = attributeValueN(name);
        if (value == null)
        {
            throw new XmlMissingAttributeException(_xmlContext, this + " is missing attribute " + name);
        }
        return value;
    }
    
    public String
    attributeValueD(String name,
                    String defaultValue)
    {
        String value = attributeValueN(name);
        if (value == null)
        {
            return defaultValue;
        }
        return value;
    }
    
    public String
    attributeValueN(String    name,
                    StringSet validSet)
    {
        String value = attributeValueN(name);
        if (value == null)
        {
            return null;
        }
        verifyAttributeStringValue(name, value, validSet);
        return value;
    }
    
    public String
    attributeValueE(String    name,
                    StringSet validSet)
    {
        String value = attributeValueE(name);
        verifyAttributeStringValue(name, value, validSet);
        return value;
    }
    
    public String
    attributeValueD(String    name,
                    String    defaultValue,
                    StringSet validSet)
    {
        String value = attributeValueN(name);
        if (value == null)
        {
            return defaultValue;
        }
        verifyAttributeStringValue(name, value, validSet);
        return value;
    }
    
    public void
    verifyAttributeStringValue(String    name,
                               String    value,
                               StringSet validSet)
    {
        if (!validSet.isIn(value))
        {
            throw new XmlAttributeNotInSetException(_xmlContext, 
                    "Attribute " + name + " of " + this + " has value " + value +
                    " which is not in StringSet " + validSet);
        }
    }
    
    public String
    nextTextN()
    {
        TextNode node = nextTextNodeN();
        if (node == null)
        {
            return null;
        }
        return node._text;
    }
    
    public String
    nextTextE()
    {
        String text = nextTextN();
        if (text == null)
        {
            throw new MissingNodeException(
                    _xmlContext, "Attempted to retrieve text beyond end of node " + this);
        }
        return text;
    }
    
    public String
    nextTextD(String defaultValue)
    {
        String text = nextTextN();
        if (text == null)
        {
            return defaultValue;
        }
        return text;
    }

    /**
     * Obtains the text contained by a node while validating that there is no child node.
     * <p>
     * According to the XML standard, and start tag and end tag with nothing in between
     * and an self-closing tag with the same name and attributes are equivalent. For
     * example &lt;x>&lt;/x> is equivalent to &lt;x/>. When this method is called for such
     * nodes, it returns empty string "" in both cases.
     * @param validSet method validates string against set.
     * @return The contained text if present and in set, empty if text is not present.
     * @throws XmlTextFieldNotInSetException
     */
    public String
    onlyText(StringSet validSet)
    {
        String text = onlyText();
        if (!validSet.isIn(text))
        {
            throw new XmlTextFieldNotInSetException(
                    _xmlContext,
                    "Node " + this + " contains text " + text +
                    " which is not in StringSet " + validSet);
        }
        return text;
    }
    
    /**
     * Obtains the text contained by a node while validating that there is no child node.
     * <p>
     * According to the XML standard, and start tag and end tag with nothing in between
     * and an self-closing tag with the same name and attributes are equivalent. For
     * example &lt;x>&lt;/x> is equivalent to &lt;x/>. When this method is called for such
     * nodes, it returns empty string "" in both cases.
     * @return The contained text if present, empty string otherwise.
     */
    public String
    onlyText()
    {
        String text = nextTextD("");
        TagNode child = nextChildN();
        if (child != null)
        {
            throw new XmlNotOnlyTextException(_xmlContext,
                    "Node " + this +
                    " is expected to contain only text but also contains child node " + child);
        }
        return text;
    }
    
    public TagNode
    anyOrderChildE(String       relativeName,
                   XmlNameSpace nameSpace)
    {
        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            String comparisonName = TagNodeId.makeComparisonName(relativeName, nameSpace);
            throw new MissingNodeException(_xmlContext, this + " does not contain " + comparisonName);
        }
        return child;
    }
    
    public TagNode
    anyOrderChildN(String name)
    {
        return anyOrderChildN(name, null);
    }
    
    public TagNode
    anyOrderChildE(String name)
    {
        return anyOrderChildE(name, null);
    }
    
    public TagNode
    anyOrderChildN(TagNodeId id)
    {
        return anyOrderChildN(id._relativeName, id._nameSpace);
    }
    
    public TagNode
    anyOrderChildE(TagNodeId id)
    {
        return anyOrderChildE(id._relativeName, id._nameSpace);
    }
    
    public String
    anyOrderTextFieldN(String       name,
                       XmlNameSpace nameSpace)
    {
        TagNode child = anyOrderChildN(name, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyText();
    }
    
    public String
    anyOrderTextFieldE(String       name,
                       XmlNameSpace nameSpace)
    {
        TagNode child = anyOrderChildE(name, nameSpace);
        return child.onlyText();
    }
    
    public String
    anyOrderTextFieldD(String       name,
                       XmlNameSpace nameSpace,
                       String       defaultValue)
    {
        TagNode child = anyOrderChildN(name, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyText();
    }
    
    public String
    anyOrderTextFieldN(String name)
    {
        TagNode child = anyOrderChildN(name, null);
        if (child == null)
        {
            return null;
        }
        return child.onlyText();
    }
    
    public String
    anyOrderTextFieldE(String name)
    {
        TagNode child = anyOrderChildE(name, null);
        return child.onlyText();
    }
    
    public String
    anyOrderTextFieldD(String name,
                       String defaultValue)
    {
        TagNode child = anyOrderChildN(name, null);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyText();
    }
    
    public String
    anyOrderTextFieldN(TagNodeId id)
    {
        TagNode child = anyOrderChildN(id._relativeName, id._nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyText();
    }
    
    public String
    anyOrderTextFieldE(TagNodeId id)
    {
        TagNode child = anyOrderChildE(id._relativeName, id._nameSpace);
        return child.onlyText();
    }
    
    public String
    anyOrderTextFieldD(TagNodeId id,
                       String    defaultValue)
    {
        TagNode child = anyOrderChildN(id._relativeName, id._nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyText();
    }
    
    //--
    
    public String
    anyOrderTextFieldN(String       name,
                       XmlNameSpace nameSpace,
                       StringSet    validSet)
    {
        TagNode child = anyOrderChildN(name, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyText(validSet);
    }
    
    public String
    anyOrderTextFieldE(String       name,
                       XmlNameSpace nameSpace,
                       StringSet    validSet)
    {
        TagNode child = anyOrderChildE(name, nameSpace);
        return child.onlyText(validSet);
    }
    
    public String
    anyOrderTextFieldD(String       name,
                       XmlNameSpace nameSpace,
                       String       defaultValue,
                       StringSet    validSet)
    {
        TagNode child = anyOrderChildN(name, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyText(validSet);
    }
    
    public String
    anyOrderTextFieldN(String    name,
                       StringSet validSet)
    {
        TagNode child = anyOrderChildN(name, null);
        if (child == null)
        {
            return null;
        }
        return child.onlyText(validSet);
    }
    
    public String
    anyOrderTextFieldE(String    name,
                       StringSet validSet)
    {
        TagNode child = anyOrderChildE(name, null);
        return child.onlyText(validSet);
    }
    
    public String
    anyOrderTextFieldD(String    name,
                       String    defaultValue,
                       StringSet validSet)
    {
        TagNode child = anyOrderChildN(name, null);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyText(validSet);
    }
    
    public String
    anyOrderTextFieldN(TagNodeId id,
                       StringSet validSet)
    {
        TagNode child = anyOrderChildN(id._relativeName, id._nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyText(validSet);
    }
    
    public String
    anyOrderTextFieldE(TagNodeId id,
                       StringSet validSet)
    {
        TagNode child = anyOrderChildE(id._relativeName, id._nameSpace);
        return child.onlyText(validSet);
    }
    
    public String
    anyOrderTextFieldD(TagNodeId id,
                       String    defaultValue,
                       StringSet validSet)
    {
        TagNode child = anyOrderChildN(id._relativeName, id._nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyText(validSet);
    }
    
    public void
    verifyNoOtherChildren(XmlNameSpace nameSpace,
                          String...    expectedNames)
    {
        TagNodeId[] expectedIds = new TagNodeId[expectedNames.length];
        for (int i = 0; i < expectedNames.length; i++)
        {
            expectedIds[i] = new TagNodeId(expectedNames[i], nameSpace);
        }
        verifyNoOtherChildren(expectedIds);
    }
    
    public void
    verifyNoOtherChildren(String... expectedNames)
    {
        TagNodeId[] expectedIds = new TagNodeId[expectedNames.length];
        for (int i = 0; i < expectedNames.length; i++)
        {
            expectedIds[i] = new TagNodeId(expectedNames[i], null);
        }
        verifyNoOtherChildren(expectedIds);
    }
    
    /**
     * Obtains an int from the node where text representing the int is the only thing between start and end tag.
     * @return int contained in the node.
     */
    public int
    onlyInt()
    {
        String textField = onlyText();
        try
        {
            int intField = Integer.parseInt(textField);
            return intField;
        }
        catch (NumberFormatException e)
        {
            throw new XmlTextFieldDoesNotMatchTypeException(
                    _xmlContext, this + " contains text " + textField +
                    " which can not be converted to int");
        }
    }
    
    public int
    onlyInt(IntSet  validSet)
    {
        int intField = onlyInt();
        if (!validSet.isIn(intField))
        {
            throw new XmlTextFieldNotInSetException(
                    _xmlContext,
                    this + " contains int " + intField + " which is not in IntSet " + validSet);
        }
        return intField;
    }

    /**
     * Obtains a long from the node where text representing the long is the only thing between start and end tag.
     * @return long contained in the node.
     */
    public long
    onlyLong()
    {
        String textField = onlyText();
        try
        {
            long longField = Long.parseLong(textField);
            return longField;
        }
        catch (NumberFormatException e)
        {
            throw new XmlTextFieldDoesNotMatchTypeException(
                    _xmlContext, this + " contains text " + textField +
                    " which can not be converted to long");
        }
    }
    
    public long
    onlyLong(LongSet validSet)
    {
        long longField = onlyLong();
        if (!validSet.isIn(longField))
        {
            throw new XmlTextFieldNotInSetException(
                    _xmlContext,
                    this + " contains long " + longField + " which is not in LongSet " + validSet);
        }
        return longField;
    }

    /**
     * Obtains a boolean from the node where text representing the boolean is the only thing between start and end tag.
     * @return boolean contained in the node.
     */
    public boolean
    onlyBool()
    {
        String textField = this.onlyText();
        if (textField.equalsIgnoreCase("true"))
        {
            return true;
        }
        else if (textField.equalsIgnoreCase("false"))
        {
            return false;
        }
        else
        {
            throw new XmlTextFieldDoesNotMatchTypeException(
                    _xmlContext, this + " contains text " + textField +
                    " which can not be converted to boolean");
        }
    }
    
    /**
     * Obtains an float from the node where text representing the float is the only thing between start and end tag.
     * @return float contained in the node.
     */
    public float
    onlyFloat()
    {
        String textField = onlyText();
        try
        {
            float floatField = Float.parseFloat(textField);
            return floatField;
        }
        catch (NumberFormatException e)
        {
            throw new XmlTextFieldDoesNotMatchTypeException(
                    _xmlContext, this + " contains text " + textField +
                    " which can not be converted to float");
        }
    }
    
    public float
    onlyFloat(FloatSet validSet)
    {
        float floatField = onlyFloat();
        if (!validSet.isIn(floatField))
        {
            throw new XmlTextFieldNotInSetException(
                 _xmlContext,
                 this + " contains float " + floatField + " which is not in FloatSet " + validSet);
        }
        return floatField;
    }
    
    /**
     * Obtains a double from the node where text representing the double is the only thing between start and end tag.
     * @return double contained in the node.
     */
    public double
    onlyDouble()
    {
        String textField = onlyText();
        try
        {
            double doubleField = Double.parseDouble(textField);
            return doubleField;
        }
        catch (NumberFormatException e)
        {
            throw new XmlTextFieldDoesNotMatchTypeException(
                    _xmlContext, this + " contains text " + textField +
                    " which can not be converted to double");
        }
    }
    
    public double
    onlyDouble(DoubleSet validSet)
    {
        double doubleField = this.onlyDouble();
        if (!validSet.isIn(doubleField))
        {
            throw new XmlTextFieldNotInSetException(
                    _xmlContext,
                    this + " contains double " + doubleField +
                    " which is not in DoubleSet " + validSet);
        }
        return doubleField;
    }

    /**
     * Obtains an enum from the node where text representing the enum is the only thing between start and end tag.
     * @param enumClass The class of the enum to be obtained. Text will be matched against the string returned by
     *                  the toString function of the various enum values in the class.
     * @return enum contained in the node.
     */
    public Enum<?>
    onlyEnum(Class<?> enumClass)
    {
        if (!enumClass.isEnum())
        {
            throw new SuppliedClassNotEnumException(
                    _xmlContext, "Attempting to map " + this + " to enum but supplied class " +
                    enumClass.getName() + " is not an Enum class");
        }
        String textField = onlyText();
        try
        {
            Method method = enumClass.getMethod("values");
            Object returnObj = method.invoke(null);
            Enum<?>[] values = (Enum<?>[])returnObj;
            for (Enum<?> value : values)
            {
                String valueString = value.toString();
                if (valueString.equals(textField))
                {
                    return value;
                }
            }
        }
        catch (Exception e)
        {
            // If reflection is coded properly, there will be no reflection based exception.
            throw new InternalBugException(_xmlContext, e.toString());
        }
        throw new XmlTextFieldDoesNotMatchTypeException(_xmlContext,
                this + " contains text " + textField +
                " which can not be converted to " + enumClass.getName());
    }

    /**
     * Obtains a object from a class with string constructor from the node where text
     * representing the object is the only thing between start and end tag.
     * @param stringMappedClass Class of the object to be created. Must have a string constructor.
     *                          That string constructor will be invoked with the node text.
     * @return object contained in the node.
     */
    protected Object
    onlyStringMapped(Class<?> stringMappedClass)
    {
        Constructor<?> constructor;
        try
        {
            constructor = stringMappedClass.getConstructor(String.class);
        }
        catch (SecurityException e)
        {
            throw new SuppliedClassNotStringMappedException(
                    _xmlContext,
                    "Attempting to make string-mapped instance from " + this +
                    " but the String constructor of supplied class " +
                    stringMappedClass.getName() +
                    " is not accessible, reflection exception is: " + e);
        }
        catch (NoSuchMethodException e)
        {
            throw new SuppliedClassNotStringMappedException(
                    _xmlContext,
                    "Attempting to make string-mapped instance from " + this +
                    " but supplied class " + stringMappedClass.getName() +
                    " has no String constructor or possibly an invisible constructor, reflection exception is: " + e);
        }
        String textField = onlyText();
        Object stringMappedInstance;
        try
        {
            stringMappedInstance = constructor.newInstance(textField);
        }
        catch (InvocationTargetException e)
        {
            throw new XmlTextFieldDoesNotMatchTypeException(_xmlContext,
                    this + " contains text " + textField +
                    " which can not be converted to " + stringMappedClass.getName() +
                    ", constructor throws " + e.getCause());
        }
        catch (Exception e)
        {
            throw new InternalBugException(_xmlContext, e.toString());
        }
        return stringMappedInstance;
    }
    
    public Integer
    nextIntFieldN(String       relativeName,
                  XmlNameSpace nameSpace)
    {/*
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return extractIntFromChild(child);*/
        return nextIntFieldN(relativeName, nameSpace, AllInts.INSTANCE);
    }
    
    public int
    nextIntFieldE(String       relativeName,
                  XmlNameSpace nameSpace)
    {
        return nextIntFieldE(relativeName, nameSpace, AllInts.INSTANCE);
    }
    
    public int
    nextIntFieldD(String       relativeName,
                  XmlNameSpace nameSpace,
                  int          defaultValue)
    {
        return nextIntFieldD(relativeName, nameSpace, defaultValue, AllInts.INSTANCE);
    }
    
    public Integer
    nextIntFieldN(String name)
    {
        return nextIntFieldN(name, (XmlNameSpace)null);
    }
    
    public int
    nextIntFieldE(String name)
    {
        return nextIntFieldE(name, (XmlNameSpace)null);
    }
    
    public int
    nextIntFieldD(String name,
                  int    defaultValue)
    {
        return nextIntFieldD(name, (XmlNameSpace)null, defaultValue);
    }
    
    public Integer
    nextIntFieldN(TagNodeId id)
    {
        return nextIntFieldN(id._relativeName, id._nameSpace);
    }
    
    public int
    nextIntFieldE(TagNodeId id)
    {
        return nextIntFieldE(id._relativeName, id._nameSpace);
    }
    
    public int
    nextIntFieldD(TagNodeId id,
                  int       defaultValue)
    {
        return nextIntFieldD(id._relativeName, id._nameSpace, defaultValue);
    }
    
    public Integer
    nextIntFieldN(String       relativeName,
                  XmlNameSpace nameSpace,
                  IntSet       validSet)
    {
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyInt(validSet);
    }
    
    public int
    nextIntFieldE(String       relativeName,
                  XmlNameSpace nameSpace,
                  IntSet       validSet)
    {

        TagNode child = nextChildE(relativeName, nameSpace);
        return child.onlyInt(validSet);
    }
    
    public int
    nextIntFieldD(String       relativeName,
                  XmlNameSpace nameSpace,
                  int          defaultValue,
                  IntSet       validSet)
    {
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyInt(validSet);
    }
    
    public Integer
    nextIntFieldN(String name,
                  IntSet validSet)
    {
        return nextIntFieldN(name, null, validSet);
    }
    
    public int
    nextIntFieldE(String name,
                  IntSet validSet)
    {

        return nextIntFieldE(name, null, validSet);
    }
    
    public int
    nextIntFieldD(String name,
                  int    defaultValue,
                  IntSet validSet)
    {
        return nextIntFieldD(name, null, defaultValue, validSet);
    }
    
    public Integer
    nextIntFieldN(TagNodeId id,
                  IntSet    validSet)
    {
        return nextIntFieldN(id._relativeName, id._nameSpace, validSet);
    }
    
    public int
    nextIntFieldE(TagNodeId id,
                  IntSet    validSet)
    {

        return nextIntFieldE(id._relativeName, id._nameSpace, validSet);
    }
    
    public int
    nextIntFieldD(TagNodeId id,
                  int       defaultValue,
                  IntSet    validSet)
    {
        return nextIntFieldD(id._relativeName, id._nameSpace, defaultValue, validSet);
    }
    
    public Long
    nextLongFieldN(String       relativeName,
                   XmlNameSpace nameSpace)
    {
        /*TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return extractLongFromChild(child);*/
        return nextLongFieldN(relativeName, nameSpace, AllLongs.INSTANCE);
    }
    
    public long
    nextLongFieldE(String       relativeName,
                   XmlNameSpace nameSpace)
    {
        /*TagNode child = nextChildE(relativeName, nameSpace);
        return extractLongFromChild(child);*/
        return nextLongFieldE(relativeName, nameSpace, AllLongs.INSTANCE);
    }
    
    public long
    nextLongFieldD(String       relativeName,
                   XmlNameSpace nameSpace,
                   long         defaultValue)
    {
        /*
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return extractLongFromChild(child);*/
        return nextLongFieldD(relativeName, nameSpace, defaultValue, AllLongs.INSTANCE);
    }
    
    public Long
    nextLongFieldN(String       relativeName,
                   XmlNameSpace nameSpace,
                   LongSet      set)
    {
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyLong(set);
    }
    
    public long
    nextLongFieldE(String       relativeName,
                   XmlNameSpace nameSpace,
                   LongSet      set)
    {

        TagNode child = nextChildE(relativeName, nameSpace);
        return child.onlyLong(set);
    }
    
    public long
    nextLongFieldD(String       relativeName,
                   XmlNameSpace nameSpace,
                   long         defaultValue,
                   LongSet      set)
    {

        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyLong(set);
    }
    
    public Long
    nextLongFieldN(String name)
    {
        return nextLongFieldN(name, (XmlNameSpace)null);
    }
    
    public long
    nextLongFieldE(String name)
    {

        return nextLongFieldE(name, (XmlNameSpace)null);
    }
    
    public long
    nextLongFieldD(String name,
                   long   defaultValue)
    {
        return nextLongFieldD(name, (XmlNameSpace)null, defaultValue);
    }
    
    public Long
    nextLongFieldN(TagNodeId id)
    {
        return nextLongFieldN(id._relativeName, id._nameSpace);
    }
    
    public long
    nextLongFieldE(TagNodeId id)
    {

        return nextLongFieldE(id._relativeName, id._nameSpace);
    }
    
    public long
    nextLongFieldD(TagNodeId id, long defaultValue)
    {
        return nextLongFieldD(id._relativeName, id._nameSpace, defaultValue);
    }
    
    public Long
    nextLongFieldN(String  name,
                   LongSet set)
    {
        return nextLongFieldN(name, null, set);
    }
    
    public long
    nextLongFieldE(String  name,
                   LongSet set)
    {

        return nextLongFieldE(name, null, set);
    }
    
    public long
    nextLongFieldD(String  name,
                   long    defaultValue,
                   LongSet set)
    {
        return nextLongFieldD(name, null, defaultValue, set);
    }
    
    public Long
    nextLongFieldN(TagNodeId id,
                   LongSet   set)
    {
        return nextLongFieldN(id._relativeName, id._nameSpace, set);
    }
    
    public long
    nextLongFieldE(TagNodeId id,
                   LongSet   set)
    {

        return nextLongFieldE(id._relativeName, id._nameSpace, set);
    }
    
    public long
    nextLongFieldD(TagNodeId id,
                   long      defaultValue,
                   LongSet   set)
    {
        return nextLongFieldD(id._relativeName, id._nameSpace, defaultValue, set);
    }
    
    public Boolean
    nextBoolFieldN(String       relativeName,
                   XmlNameSpace nameSpace)
    {
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyBool();
    }
    
    public boolean
    nextBoolFieldE(String       relativeName,
                   XmlNameSpace nameSpace)
    {

        TagNode child = nextChildE(relativeName, nameSpace);
        return child.onlyBool();
    }
    
    public boolean
    nextBoolFieldD(String       relativeName,
                   XmlNameSpace nameSpace,
                   boolean      defaultValue)
    {

        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyBool();
    }
    
    public Boolean
    nextBoolFieldN(String name)
    {
        return nextBoolFieldN(name, null);
    }
    
    public boolean
    nextBoolFieldE(String name)
    {

        return nextBoolFieldE(name, null);
    }
    
    public boolean
    nextBoolFieldD(String  name,
                   boolean defaultValue)
    {
        return nextBoolFieldD(name, null, defaultValue);
    }
    
    public Boolean
    nextBoolFieldN(TagNodeId id)
    {
        return nextBoolFieldN(id._relativeName, id._nameSpace);
    }
    
    public boolean
    nextBoolFieldE(TagNodeId id)
    {

        return nextBoolFieldE(id._relativeName, id._nameSpace);
    }
    
    public boolean
    nextBoolFieldD(TagNodeId id, boolean defaultValue)
    {
        return nextBoolFieldD(id._relativeName, id._nameSpace, defaultValue);
    }
    
    public Float
    nextFloatFieldN(String       relativeName,
                    XmlNameSpace nameSpace)
    {
        /*TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return extractFloatFromChild(child);*/
        return nextFloatFieldN(relativeName, nameSpace, AllFloats.INSTANCE);
    }
    
    public Float
    nextFloatFieldN(String       relativeName,
                    XmlNameSpace nameSpace,
                    FloatSet     validSet)
    {
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyFloat(validSet);
    }
    
    public float
    nextFloatFieldE(String       relativeName,
                    XmlNameSpace nameSpace)
    {
        /*TagNode child = nextChildE(relativeName, nameSpace);
        return extractFloatFromChild(child);*/
        return nextFloatFieldE(relativeName, nameSpace, AllFloats.INSTANCE);
    }
    
    public float
    nextFloatFieldE(String       relativeName,
                    XmlNameSpace nameSpace,
                    FloatSet     validSet)
    {

        TagNode child = nextChildE(relativeName, nameSpace);
        return child.onlyFloat(validSet);
    }
    
    public float
    nextFloatFieldD(String       relativeName,
                    XmlNameSpace nameSpace,
                    float        defaultValue)
    {
        /*TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return extractFloatFromChild(child);*/
        return nextFloatFieldD(relativeName, nameSpace, defaultValue, AllFloats.INSTANCE);
    }
    
    public float
    nextFloatFieldD(String       relativeName,
                    XmlNameSpace nameSpace,
                    float        defaultValue,
                    FloatSet     validSet)
    {
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyFloat(validSet);
    }
    
    public Float
    nextFloatFieldN(String name)
    {
        return nextFloatFieldN(name, (XmlNameSpace)null);
    }
    
    public float
    nextFloatFieldE(String name)
    {

        return nextFloatFieldE(name, (XmlNameSpace)null);
    }
    
    public float
    nextFloatFieldD(String  name,
                    float   defaultValue)
    {
        return nextFloatFieldD(name, null, defaultValue);
    }
    
    public Float
    nextFloatFieldN(TagNodeId id)
    {
        return nextFloatFieldN(id._relativeName, id._nameSpace);
    }
    
    public float
    nextFloatFieldE(TagNodeId id)
    {

        return nextFloatFieldE(id._relativeName, id._nameSpace);
    }
    
    public float
    nextFloatFieldD(TagNodeId id, float defaultValue)
    {
        return nextFloatFieldD(id._relativeName, id._nameSpace, defaultValue);
    }
    
    public Float
    nextFloatFieldN(String   name,
                    FloatSet validSet)
    {
        return nextFloatFieldN(name, null, validSet);
    }
    
    public float
    nextFloatFieldE(String   name,
                    FloatSet validSet)
    {

        return nextFloatFieldE(name, null, validSet);
    }
    
    public float
    nextFloatFieldD(String   name,
                    float    defaultValue,
                    FloatSet validSet)
    {
        return nextFloatFieldD(name, null, defaultValue, validSet);
    }
    
    public Float
    nextFloatFieldN(TagNodeId id,
                    FloatSet  validSet)
    {
        return nextFloatFieldN(id._relativeName, id._nameSpace, validSet);
    }
    
    public float
    nextFloatFieldE(TagNodeId id,
                    FloatSet  validSet)
    {

        return nextFloatFieldE(id._relativeName, id._nameSpace, validSet);
    }
    
    public float
    nextFloatFieldD(TagNodeId id,
                    float     defaultValue,
                    FloatSet  validSet)
    {
        return nextFloatFieldD(id._relativeName, id._nameSpace, defaultValue, validSet);
    }
    
    public Double
    nextDoubleFieldN(String       relativeName,
                     XmlNameSpace nameSpace)
    {/*
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return extractDoubleFromChild(child);*/
        return nextDoubleFieldN(relativeName, nameSpace, AllDoubles.INSTANCE);
    }
    
    public Double
    nextDoubleFieldN(String       relativeName,
                     XmlNameSpace nameSpace,
                     DoubleSet    validSet)
    {
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyDouble(validSet);
    }
    
    public double
    nextDoubleFieldE(String       relativeName,
                     XmlNameSpace nameSpace)
    {
        /*TagNode child = nextChildE(relativeName, nameSpace);
        return extractDoubleFromChild(child);*/
        return nextDoubleFieldE(relativeName, nameSpace, AllDoubles.INSTANCE);
    }
    
    public double
    nextDoubleFieldE(String       relativeName,
                     XmlNameSpace nameSpace,
                     DoubleSet    validSet)
    {

        TagNode child = nextChildE(relativeName, nameSpace);
        return child.onlyDouble(validSet);
    }
    
    public double
    nextDoubleFieldD(String       relativeName,
                     XmlNameSpace nameSpace,
                     double       defaultValue)
    {

        /*TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return extractDoubleFromChild(child);*/
        return nextDoubleFieldD(relativeName, nameSpace, defaultValue, AllDoubles.INSTANCE);
    }
    
    public double
    nextDoubleFieldD(String       relativeName,
                     XmlNameSpace nameSpace,
                     double       defaultValue,
                     DoubleSet    validSet)
    {

        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyDouble(validSet);
    }
    
    public Double
    nextDoubleFieldN(String name)
    {
        return nextDoubleFieldN(name, (XmlNameSpace)null);
    }
    
    public double
    nextDoubleFieldE(String name)
    {

        return nextDoubleFieldE(name, (XmlNameSpace)null);
    }
    
    public double
    nextDoubleFieldD(String  name,
                     double  defaultValue)
    {
        return nextDoubleFieldD(name, null, defaultValue);
    }
    
    public Double
    nextDoubleFieldN(TagNodeId id)
    {
        return nextDoubleFieldN(id._relativeName, id._nameSpace);
    }
    
    public double
    nextDoubleFieldE(TagNodeId id)
    {
        return nextDoubleFieldE(id._relativeName, id._nameSpace);
    }
    
    public double
    nextDoubleFieldD(TagNodeId id, double defaultValue)
    {
        return nextDoubleFieldD(id._relativeName, id._nameSpace, defaultValue);
    }
    
    public Double
    nextDoubleFieldN(String    name,
                     DoubleSet validSet)
    {
        return nextDoubleFieldN(name, null, validSet);
    }
    
    public double
    nextDoubleFieldE(String    name,
                     DoubleSet validSet)
    {

        return nextDoubleFieldE(name, null, validSet);
    }
    
    public double
    nextDoubleFieldD(String    name,
                     double    defaultValue,
                     DoubleSet validSet)
    {
        return nextDoubleFieldD(name, null, defaultValue, validSet);
    }
    
    public Double
    nextDoubleFieldN(TagNodeId id,
                     DoubleSet validSet)
    {
        return nextDoubleFieldN(id._relativeName, id._nameSpace, validSet);
    }
    
    public double
    nextDoubleFieldE(TagNodeId id,
                     DoubleSet validSet)
    {
        return nextDoubleFieldE(id._relativeName, id._nameSpace, validSet);
    }
    
    public double
    nextDoubleFieldD(TagNodeId id,
                     double    defaultValue,
                     DoubleSet validSet)
    {
        return nextDoubleFieldD(id._relativeName, id._nameSpace, defaultValue, validSet);
    }
    
    public Enum<?>
    nextEnumFieldN(String       relativeName,
                   XmlNameSpace nameSpace,
                   Class<?>     enumClass)
    {
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyEnum(enumClass);
    }
    
    public Enum<?>
    nextEnumFieldE(String       relativeName,
                   XmlNameSpace nameSpace,
                   Class<?>     enumClass)
    {

        TagNode child = nextChildE(relativeName, nameSpace);
        return child.onlyEnum(enumClass);
    }
    
    public Enum<?>
    nextEnumFieldD(String       relativeName,
                   XmlNameSpace nameSpace,
                   Class<?>     enumClass,
                   Enum<?>      defaultValue)
    {

        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyEnum(enumClass);
    }
    
    public Enum<?>
    nextEnumFieldN(String   name,
                   Class<?> enumClass)
    {
        return nextEnumFieldN(name, null, enumClass);
    }
    
    public Enum<?>
    nextEnumFieldE(String   name,
                   Class<?> enumClass)
    {

        return nextEnumFieldE(name, null, enumClass);
    }
    
    public Enum<?>
    nextEnumFieldD(String   name,
                   Class<?> enumClass,
                   Enum<?>  defaultValue)
    {
        return nextEnumFieldD(name, null, enumClass, defaultValue);
    }
    
    public Enum<?>
    nextEnumFieldN(TagNodeId id,
                   Class<?>  enumClass)
    {
        return nextEnumFieldN(id._relativeName, id._nameSpace, enumClass);
    }
    
    public Enum<?>
    nextEnumFieldE(TagNodeId id,
                   Class<?>  enumClass)
    {

        return nextEnumFieldE(id._relativeName, id._nameSpace, enumClass);
    }
    
    public Enum<?>
    nextEnumFieldD(TagNodeId id,
                   Class<?>  enumClass,
                   Enum<?>   defaultValue)
    {
        return nextEnumFieldD(id._relativeName, id._nameSpace, enumClass, defaultValue);
    }
    
    public Object
    nextStringMappedFieldN(String       relativeName,
                           XmlNameSpace nameSpace,
                           Class<?>     stringMappedClass)
    {
        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyStringMapped(stringMappedClass);
    }
    
    public Object
    nextStringMappedFieldE(String       relativeName,
                           XmlNameSpace nameSpace,
                           Class<?>     stringMappedClass)
    {

        TagNode child = nextChildE(relativeName, nameSpace);
        return child.onlyStringMapped(stringMappedClass);
    }
    
    public Object
    nextStringMappedFieldD(String       relativeName,
                           XmlNameSpace nameSpace,
                           Class<?>     stringMappedClass,
                           Object      defaultValue)
    {

        TagNode child = nextChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyStringMapped(stringMappedClass);
    }
    
    public Object
    nextStringMappedFieldN(String   name,
                           Class<?> stringMappedClass)
    {
        return nextStringMappedFieldN(name, null, stringMappedClass);
    }
    
    public Object
    nextStringMappedFieldE(String   name,
                           Class<?> stringMappedClass)
    {

        return nextStringMappedFieldE(name, null, stringMappedClass);
    }
    
    public Object
    nextStringMappedFieldD(String   name,
                           Class<?> stringMappedClass,
                           Object   defaultValue)
    {
        return nextStringMappedFieldD(name, null, stringMappedClass, defaultValue);
    }
    
    public Object
    nextStringMappedFieldN(TagNodeId id,
                           Class<?>  stringMappedClass)
    {
        return nextStringMappedFieldN(id._relativeName, id._nameSpace, stringMappedClass);
    }
    
    public Object
    nextStringMappedFieldE(TagNodeId id,
                           Class<?>  stringMappedClass)
    {

        return nextStringMappedFieldE(id._relativeName, id._nameSpace, stringMappedClass);
    }
    
    public Object
    nextStringMappedFieldD(TagNodeId id,
                           Class<?>  stringMappedClass,
                           Object    defaultValue)
    {
        return nextStringMappedFieldD(id._relativeName, id._nameSpace, stringMappedClass, defaultValue);
    }
    
    //--
    
    public String
    anyOrderConstFieldN(String       relativeName,
                        XmlNameSpace nameSpace,
                        String       expectedValue)
    {
        TagNode childNode = anyOrderChildN(relativeName, nameSpace);
        if (childNode == null)
        {
            return null;
        }
        return checkConstField(childNode, expectedValue);
    }
    
    public String
    anyOrderConstFieldE(String       relativeName,
                        XmlNameSpace nameSpace,
                        String       expectedValue)
    {
        TagNode childNode = anyOrderChildE(relativeName, nameSpace);
        return checkConstField(childNode, expectedValue);
    }
    
    public String
    anyOrderConstFieldD(String       relativeName,
                        XmlNameSpace nameSpace,
                        String       expectedValue)
    {
        TagNode childNode = anyOrderChildN(relativeName, nameSpace);
        if (childNode == null)
        {
            return expectedValue;
        }
        return checkConstField(childNode, expectedValue);
    }
    
    public String
    anyOrderConstFieldN(String       name,
                        String       expectedValue)
    {
        return anyOrderConstFieldN(name, null, expectedValue);
    }
    
    public String
    anyOrderConstFieldE(String       name,
                        String       expectedValue)
    {
        return anyOrderConstFieldE(name, null, expectedValue);
    }
    
    public String
    anyOrderConstFieldD(String       name,
                        String       expectedValue)
    {
        return anyOrderConstFieldD(name, null, expectedValue);
    }
    
    public String
    anyOrderConstFieldN(TagNodeId id,
                        String    expectedValue)
    {
        return anyOrderConstFieldN(id._relativeName, id._nameSpace, expectedValue);
    }
    
    public String
    anyOrderConstFieldE(TagNodeId id,
                        String    expectedValue)
    {
        return anyOrderConstFieldE(id._relativeName, id._nameSpace, expectedValue);
    }
    
    public String
    anyOrderConstFieldD(TagNodeId id,
                        String    expectedValue)
    {
        return anyOrderConstFieldD(id._relativeName, id._nameSpace, expectedValue);
    }
    
    public Integer
    anyOrderIntFieldN(String       relativeName,
                      XmlNameSpace nameSpace)
    {
        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyInt();
    }
    
    public Integer
    anyOrderIntFieldE(String       relativeName,
                      XmlNameSpace nameSpace)
    {
        TagNode child = anyOrderChildE(relativeName, nameSpace);
        return child.onlyInt();
    }
    
    public int
    anyOrderIntFieldD(String       relativeName,
                      XmlNameSpace nameSpace,
                      int          defaultValue)
    {

        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyInt();
    }
    
    public Integer
    anyOrderIntFieldN(String name)
    {
        return anyOrderIntFieldN(name, (XmlNameSpace)null);
    }
    
    public int
    anyOrderIntFieldE(String name)
    {

        return anyOrderIntFieldE(name, (XmlNameSpace)null);
    }
    
    public int
    anyOrderIntFieldD(String name,
                      int    defaultValue)
    {
        return anyOrderIntFieldD(name, null, defaultValue);
    }
    
    public Integer
    anyOrderIntFieldN(TagNodeId id)
    {
        return anyOrderIntFieldN(id._relativeName, id._nameSpace);
    }
    
    public int
    anyOrderIntFieldE(TagNodeId id)
    {

        return anyOrderIntFieldE(id._relativeName, id._nameSpace);
    }
    
    public int
    anyOrderIntFieldD(TagNodeId id, int defaultValue)
    {
        return anyOrderIntFieldD(id._relativeName, id._nameSpace, defaultValue);
    }
    
    //------

    
    public Integer
    anyOrderIntFieldN(String       relativeName,
                      XmlNameSpace nameSpace,
                      IntSet       validSet)
    {
        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyInt(validSet);
    }
    
    public Integer
    anyOrderIntFieldE(String       relativeName,
                      XmlNameSpace nameSpace,
                      IntSet       validSet)
    {
        TagNode child = anyOrderChildE(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyInt(validSet);
    }
    
    public int
    anyOrderIntFieldD(String       relativeName,
                      XmlNameSpace nameSpace,
                      int          defaultValue,
                      IntSet       validSet)
    {

        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyInt(validSet);
    }
    
    public Integer
    anyOrderIntFieldN(String name,
                      IntSet validSet)
    {
        return anyOrderIntFieldN(name, null, validSet);
    }
    
    public int
    anyOrderIntFieldE(String name,
                      IntSet validSet)
    {

        return anyOrderIntFieldE(name, null, validSet);
    }
    
    public int
    anyOrderIntFieldD(String name,
                      int    defaultValue,
                      IntSet validSet)
    {
        return anyOrderIntFieldD(name, null, defaultValue, validSet);
    }
    
    public Integer
    anyOrderIntFieldN(TagNodeId id,
                      IntSet    validSet)
    {
        return anyOrderIntFieldN(id._relativeName, id._nameSpace, validSet);
    }
    
    public int
    anyOrderIntFieldE(TagNodeId id,
                      IntSet    validSet)
    {

        return anyOrderIntFieldE(id._relativeName, id._nameSpace, validSet);
    }
    
    public int
    anyOrderIntFieldD(TagNodeId id,
                      int       defaultValue,
                      IntSet    validSet)
    {
        return anyOrderIntFieldD(id._relativeName, id._nameSpace, defaultValue, validSet);
    }
    
    public Long
    anyOrderLongFieldN(String       relativeName,
                       XmlNameSpace nameSpace)
    {
        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyLong();
    }
    
    public long
    anyOrderLongFieldE(String       relativeName,
                       XmlNameSpace nameSpace)
    {

        TagNode child = anyOrderChildE(relativeName, nameSpace);
        return child.onlyLong();
    }
    
    public long
    anyOrderLongFieldD(String       relativeName,
                       XmlNameSpace nameSpace,
                       long         defaultValue)
    {

        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyLong();
    }
    
    public Long
    anyOrderLongFieldN(String name)
    {
        return anyOrderLongFieldN(name, (XmlNameSpace)null);
    }
    
    public long
    anyOrderLongFieldE(String name)
    {

        return anyOrderLongFieldE(name, (XmlNameSpace)null);
    }
    
    public long
    anyOrderLongFieldD(String name,
                       long   defaultValue)
    {
        return anyOrderLongFieldD(name, null, defaultValue);
    }
    
    public Long
    anyOrderLongFieldN(TagNodeId id)
    {
        return anyOrderLongFieldN(id._relativeName, id._nameSpace);
    }
    
    public long
    anyOrderLongFieldE(TagNodeId id)
    {

        return anyOrderLongFieldE(id._relativeName, id._nameSpace);
    }
    
    public long
    anyOrderLongFieldD(TagNodeId id,
                       long      defaultValue)
    {
        return anyOrderLongFieldD(id._relativeName, id._nameSpace, defaultValue);
    }
    
    // --
    
    public Long
    anyOrderLongFieldN(String       relativeName,
                       XmlNameSpace nameSpace,
                       LongSet      validSet)
    {
        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyLong(validSet);
    }
    
    public long
    anyOrderLongFieldE(String       relativeName,
                       XmlNameSpace nameSpace,
                       LongSet      validSet)
    {

        TagNode child = anyOrderChildE(relativeName, nameSpace);
        return child.onlyLong(validSet);
    }
    
    public long
    anyOrderLongFieldD(String       relativeName,
                       XmlNameSpace nameSpace,
                       long         defaultValue,
                       LongSet      validSet)
    {

        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyLong(validSet);
    }
    
    public Long
    anyOrderLongFieldN(String  name,
                       LongSet validSet)
    {
        return anyOrderLongFieldN(name, null, validSet);
    }
    
    public long
    anyOrderLongFieldE(String  name,
                       LongSet validSet)
    {

        return anyOrderLongFieldE(name, null, validSet);
    }
    
    public long
    anyOrderLongFieldD(String  name,
                       long    defaultValue,
                       LongSet validSet)
    {
        return anyOrderLongFieldD(name, null, defaultValue, validSet);
    }
    
    public Long
    anyOrderLongFieldN(TagNodeId id,
                       LongSet   validSet)
    {
        return anyOrderLongFieldN(id._relativeName, id._nameSpace, validSet);
    }
    
    public long
    anyOrderLongFieldE(TagNodeId id,
                       LongSet   validSet)
    {

        return anyOrderLongFieldE(id._relativeName, id._nameSpace, validSet);
    }
    
    public long
    anyOrderLongFieldD(TagNodeId id,
                       long      defaultValue,
                       LongSet   validSet)
    {
        return anyOrderLongFieldD(id._relativeName, id._nameSpace, defaultValue, validSet);
    }
    
    public Boolean
    anyOrderBoolFieldN(String       relativeName,
                       XmlNameSpace nameSpace)
    {
        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyBool();
    }
    
    public boolean
    anyOrderBoolFieldE(String       relativeName,
                       XmlNameSpace nameSpace)
    {

        TagNode child = anyOrderChildE(relativeName, nameSpace);
        return child.onlyBool();
    }
    
    public boolean
    anyOrderBoolFieldD(String       relativeName,
                       XmlNameSpace nameSpace,
                       boolean      defaultValue)
    {

        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyBool();
    }
    
    public Boolean
    anyOrderBoolFieldN(String name)
    {
        return anyOrderBoolFieldN(name, null);
    }
    
    public boolean
    anyOrderBoolFieldE(String name)
    {

        return anyOrderBoolFieldE(name, null);
    }
    
    public boolean
    anyOrderBoolFieldD(String  name,
                       boolean defaultValue)
    {
        return anyOrderBoolFieldD(name, null, defaultValue);
    }
    
    public Boolean
    anyOrderBoolFieldN(TagNodeId id)
    {
        return anyOrderBoolFieldN(id._relativeName, id._nameSpace);
    }
    
    public boolean
    anyOrderBoolFieldE(TagNodeId id)
    {

        return anyOrderBoolFieldE(id._relativeName, id._nameSpace);
    }
    
    public boolean
    anyOrderBoolFieldD(TagNodeId id, boolean defaultValue)
    {
        return anyOrderBoolFieldD(id._relativeName, id._nameSpace, defaultValue);
    }
    
    public Float
    anyOrderFloatFieldN(String       relativeName,
                        XmlNameSpace nameSpace)
    {
        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyFloat();
    }
    
    public float
    anyOrderFloatFieldE(String       relativeName,
                        XmlNameSpace nameSpace)
    {

        TagNode child = anyOrderChildE(relativeName, nameSpace);
        return child.onlyFloat();
    }
    
    public float
    anyOrderFloatFieldD(String       relativeName,
                        XmlNameSpace nameSpace,
                        float        defaultValue)
    {

        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyFloat();
    }
    
    public Float
    anyOrderFloatFieldN(String name)
    {
        return anyOrderFloatFieldN(name, (XmlNameSpace)null);
    }
    
    public float
    anyOrderFloatFieldE(String name)
    {

        return anyOrderFloatFieldE(name, (XmlNameSpace)null);
    }
    
    public float
    anyOrderFloatFieldD(String  name,
                    float   defaultValue)
    {
        return anyOrderFloatFieldD(name, null, defaultValue);
    }
    
    public Float
    anyOrderFloatFieldN(TagNodeId id)
    {
        return anyOrderFloatFieldN(id._relativeName, id._nameSpace);
    }
    
    public float
    anyOrderFloatFieldE(TagNodeId id)
    {

        return anyOrderFloatFieldE(id._relativeName, id._nameSpace);
    }
    
    public float
    anyOrderFloatFieldD(TagNodeId id,
                        float     defaultValue)
    {
        return anyOrderFloatFieldD(id._relativeName, id._nameSpace, defaultValue);
    }
    //--
    
    public Float
    anyOrderFloatFieldN(String       relativeName,
                        XmlNameSpace nameSpace,
                        FloatSet     validSet)
    {
        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyFloat(validSet);
    }
    
    public float
    anyOrderFloatFieldE(String       relativeName,
                        XmlNameSpace nameSpace,
                        FloatSet     validSet)
    {

        TagNode child = anyOrderChildE(relativeName, nameSpace);
        return child.onlyFloat(validSet);
    }
    
    public float
    anyOrderFloatFieldD(String       relativeName,
                        XmlNameSpace nameSpace,
                        float        defaultValue,
                        FloatSet     validSet)
    {

        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyFloat(validSet);
    }
    
    public Float
    anyOrderFloatFieldN(String   name,
                        FloatSet validSet)
    {
        return anyOrderFloatFieldN(name, null, validSet);
    }
    
    public float
    anyOrderFloatFieldE(String   name,
                        FloatSet validSet)
    {

        return anyOrderFloatFieldE(name, null, validSet);
    }
    
    public float
    anyOrderFloatFieldD(String   name,
                        float    defaultValue,
                        FloatSet validSet)
    {
        return anyOrderFloatFieldD(name, null, defaultValue, validSet);
    }
    
    public Float
    anyOrderFloatFieldN(TagNodeId id,
                        FloatSet  validSet)
    {
        return anyOrderFloatFieldN(id._relativeName, id._nameSpace, validSet);
    }
    
    public float
    anyOrderFloatFieldE(TagNodeId id,
                        FloatSet  validSet)
    {

        return anyOrderFloatFieldE(id._relativeName, id._nameSpace, validSet);
    }
    
    public float
    anyOrderFloatFieldD(TagNodeId id,
                        float     defaultValue,
                        FloatSet  validSet)
    {
        return anyOrderFloatFieldD(id._relativeName, id._nameSpace, defaultValue, validSet);
    }
    
    public Double
    anyOrderDoubleFieldN(String       relativeName,
                         XmlNameSpace nameSpace)
    {
        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyDouble();
    }
    
    public double
    anyOrderDoubleFieldE(String       relativeName,
                         XmlNameSpace nameSpace)
    {

        TagNode child = anyOrderChildE(relativeName, nameSpace);
        return child.onlyDouble();
    }
    
    public double
    anyOrderDoubleFieldD(String       relativeName,
                         XmlNameSpace nameSpace,
                         double       defaultValue)
    {

        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyDouble();
    }
    
    public Double
    anyOrderDoubleFieldN(String name)
    {
        return anyOrderDoubleFieldN(name, (XmlNameSpace)null);
    }
    
    public double
    anyOrderDoubleFieldE(String name)
    {

        return anyOrderDoubleFieldE(name, (XmlNameSpace)null);
    }
    
    public double
    anyOrderDoubleFieldD(String  name,
                         double  defaultValue)
    {
        return anyOrderDoubleFieldD(name, null, defaultValue);
    }
    
    public Double
    anyOrderDoubleFieldN(TagNodeId id)
    {
        return anyOrderDoubleFieldN(id._relativeName, id._nameSpace);
    }
    
    public double
    anyOrderDoubleFieldE(TagNodeId id)
    {
        return anyOrderDoubleFieldE(id._relativeName, id._nameSpace);
    }
    
    public double
    anyOrderDoubleFieldD(TagNodeId id,
                         double defaultValue)
    {
        return anyOrderDoubleFieldD(id._relativeName, id._nameSpace, defaultValue);
    }
    
    //--
    
    public Double
    anyOrderDoubleFieldN(String       relativeName,
                         XmlNameSpace nameSpace,
                         DoubleSet    validSet)
    {
        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyDouble(validSet);
    }
    
    public double
    anyOrderDoubleFieldE(String       relativeName,
                         XmlNameSpace nameSpace,
                         DoubleSet    validSet)
    {

        TagNode child = anyOrderChildE(relativeName, nameSpace);
        return child.onlyDouble(validSet);
    }
    
    public double
    anyOrderDoubleFieldD(String       relativeName,
                         XmlNameSpace nameSpace,
                         double       defaultValue,
                         DoubleSet    validSet)
    {

        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyDouble(validSet);
    }
    
    public Double
    anyOrderDoubleFieldN(String    name,
                         DoubleSet validSet)
    {
        return anyOrderDoubleFieldN(name, null, validSet);
    }
    
    public double
    anyOrderDoubleFieldE(String    name,
                         DoubleSet validSet)
    {

        return anyOrderDoubleFieldE(name, null, validSet);
    }
    
    public double
    anyOrderDoubleFieldD(String    name,
                         double    defaultValue,
                         DoubleSet validSet)
    {
        return anyOrderDoubleFieldD(name, null, defaultValue, validSet);
    }
    
    public Double
    anyOrderDoubleFieldN(TagNodeId id,
                         DoubleSet validSet)
    {
        return anyOrderDoubleFieldN(id._relativeName, id._nameSpace, validSet);
    }
    
    public double
    anyOrderDoubleFieldE(TagNodeId id,
                         DoubleSet validSet)
    {
        return anyOrderDoubleFieldE(id._relativeName, id._nameSpace, validSet);
    }
    
    public double
    anyOrderDoubleFieldD(TagNodeId id,
                         double    defaultValue,
                         DoubleSet validSet)
    {
        return anyOrderDoubleFieldD(id._relativeName, id._nameSpace, defaultValue, validSet);
    }
    
    public Enum<?>
    anyOrderEnumFieldN(String       relativeName,
                       XmlNameSpace nameSpace,
                       Class<?>     enumClass)
    {
        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyEnum(enumClass);
    }
    
    public Enum<?>
    anyOrderEnumFieldE(String       relativeName,
                       XmlNameSpace nameSpace,
                       Class<?>     enumClass)
    {

        TagNode child = anyOrderChildE(relativeName, nameSpace);
        return child.onlyEnum(enumClass);
    }
    
    public Enum<?>
    anyOrderEnumFieldD(String       relativeName,
                       XmlNameSpace nameSpace,
                       Class<?>     enumClass,
                       Enum<?>      defaultValue)
    {

        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyEnum(enumClass);
    }
    
    public Enum<?>
    anyOrderEnumFieldN(String   name,
                       Class<?> enumClass)
    {
        return anyOrderEnumFieldN(name, null, enumClass);
    }
    
    public Enum<?>
    anyOrderEnumFieldE(String   name,
                       Class<?> enumClass)
    {

        return anyOrderEnumFieldE(name, null, enumClass);
    }
    
    public Enum<?>
    anyOrderEnumFieldD(String   name,
                       Class<?> enumClass,
                       Enum<?>  defaultValue)
    {
        return anyOrderEnumFieldD(name, null, enumClass, defaultValue);
    }
    
    public Enum<?>
    anyOrderEnumFieldN(TagNodeId id,
                       Class<?>  enumClass)
    {
        return anyOrderEnumFieldN(id._relativeName, id._nameSpace, enumClass);
    }
    
    public Enum<?>
    anyOrderEnumFieldE(TagNodeId id,
                       Class<?>  enumClass)
    {

        return anyOrderEnumFieldE(id._relativeName, id._nameSpace, enumClass);
    }
    
    public Enum<?>
    anyOrderEnumFieldD(TagNodeId id,
                       Class<?>  enumClass,
                       Enum<?>   defaultValue)
    {
        return anyOrderEnumFieldD(id._relativeName, id._nameSpace, enumClass, defaultValue);
    }
    
    //---
    
    public Object
    anyOrderStringMappedFieldN(String       relativeName,
                               XmlNameSpace nameSpace,
                               Class<?>     stringMappedClass)
    {
        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return null;
        }
        return child.onlyStringMapped(stringMappedClass);
    }
    
    public Object
    anyOrderStringMappedFieldE(String       relativeName,
                               XmlNameSpace nameSpace,
                               Class<?>     stringMappedClass)
    {

        TagNode child = anyOrderChildE(relativeName, nameSpace);
        return child.onlyStringMapped(stringMappedClass);
    }
    
    public Object
    anyOrderStringMappedFieldD(String       relativeName,
                               XmlNameSpace nameSpace,
                               Class<?>     stringMappedClass,
                               Object      defaultValue)
    {
        TagNode child = anyOrderChildN(relativeName, nameSpace);
        if (child == null)
        {
            return defaultValue;
        }
        return child.onlyStringMapped(stringMappedClass);
    }
    
    public Object
    anyOrderStringMappedFieldN(String   name,
                               Class<?> stringMappedClass)
    {
        return anyOrderStringMappedFieldN(name, null, stringMappedClass);
    }
    
    public Object
    anyOrderStringMappedFieldE(String   name,
                               Class<?> stringMappedClass)
    {

        return anyOrderStringMappedFieldE(name, null, stringMappedClass);
    }
    
    public Object
    anyOrderStringMappedFieldD(String   name,
                               Class<?> stringMappedClass,
                               Object  defaultValue)
    {
        return anyOrderStringMappedFieldD(name, null, stringMappedClass, defaultValue);
    }
    
    public Object
    anyOrderStringMappedFieldN(TagNodeId id,
                               Class<?>  stringMappedClass)
    {
        return anyOrderStringMappedFieldN(id._relativeName, id._nameSpace, stringMappedClass);
    }
    
    public Object
    anyOrderStringMappedFieldE(TagNodeId id,
                               Class<?>  stringMappedClass)
    {

        return anyOrderStringMappedFieldE(id._relativeName, id._nameSpace, stringMappedClass);
    }
    
    public Object
    anyOrderStringMappedFieldD(TagNodeId id,
                               Class<?>  stringMappedClass,
                               Object   defaultValue)
    {
        return anyOrderStringMappedFieldD(id._relativeName, id._nameSpace, stringMappedClass, defaultValue);
    }
    
    public Integer
    attributeIntN(String name)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return null;
        }
        return extractIntFromAttrValue(name, valueS);
    }
    
    public int
    attributeIntE(String name)
    {
        String valueS = attributeValueE(name);
        return extractIntFromAttrValue(name, valueS);
    }
    
    public int
    attributeIntD(String name,
                  int    defaultValue)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return extractIntFromAttrValue(name, valueS);
    }
    
    public Integer
    attributeIntN(String name,
                  IntSet validSet)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return null;
        }
        return extractIntFromAttrValue(name, valueS, validSet);
    }
    
    public int
    attributeIntE(String name,
                  IntSet validSet)
    {
        String valueS = attributeValueE(name);
        return extractIntFromAttrValue(name, valueS, validSet);
    }
    
    public int
    attributeIntD(String name,
                  int    defaultValue,
                  IntSet validSet)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return extractIntFromAttrValue(name, valueS, validSet);
    }
    
    public Long
    attributeLongN(String name)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return null;
        }
        return extractLongFromAttrValue(name, valueS);
    }
    
    public long
    attributeLongE(String name)
    {
        String valueS = attributeValueE(name);
        return extractLongFromAttrValue(name, valueS);
    }
    
    public long
    attributeLongD(String name,
                   long   defaultValue)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return extractLongFromAttrValue(name, valueS);
    }
    
    public Long
    attributeLongN(String  name,
                   LongSet validSet)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return null;
        }
        return extractLongFromAttrValue(name, valueS, validSet);
    }
    
    public long
    attributeLongE(String  name,
                   LongSet validSet)
    {
        String valueS = attributeValueE(name);
        return extractLongFromAttrValue(name, valueS, validSet);
    }
    
    public long
    attributeLongD(String  name,
                   long    defaultValue,
                   LongSet validSet)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return extractLongFromAttrValue(name, valueS, validSet);
    }
    
    public Boolean
    attributeBoolN(String name)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return null;
        }
        return extractBoolFromAttrValue(name, valueS);
    }
    
    public boolean
    attributeBoolE(String name)
    {
        String valueS = attributeValueE(name);
        return extractBoolFromAttrValue(name, valueS);
    }
    
    public boolean
    attributeBoolD(String  name,
                   boolean defaultValue)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return extractBoolFromAttrValue(name, valueS);
    }
    
    public Float
    attributeFloatN(String name)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return null;
        }
        return extractFloatFromAttrValue(name, valueS);
    }
    
    public float
    attributeFloatE(String name)
    {
        String valueS = attributeValueE(name);
        return extractFloatFromAttrValue(name, valueS);
    }
    
    public float
    attributeFloatD(String name,
                    float  defaultValue)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return extractFloatFromAttrValue(name, valueS);
    }
    
    public Float
    attributeFloatN(String   name,
                    FloatSet validSet)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return null;
        }
        return extractFloatFromAttrValue(name, valueS, validSet);
    }
    
    public float
    attributeFloatE(String   name,
                    FloatSet validSet)
    {
        String valueS = attributeValueE(name);
        return extractFloatFromAttrValue(name, valueS, validSet);
    }
    
    public float
    attributeFloatD(String   name,
                    float    defaultValue,
                    FloatSet validSet)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return extractFloatFromAttrValue(name, valueS, validSet);
    }
    
    public Double
    attributeDoubleN(String name)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return null;
        }
        return extractDoubleFromAttrValue(name, valueS);
    }
    
    public double
    attributeDoubleE(String name)
    {
        String valueS = attributeValueE(name);
        return extractDoubleFromAttrValue(name, valueS);
    }
    
    public double
    attributeDoubleD(String name,
                     double defaultValue)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return extractDoubleFromAttrValue(name, valueS);
    }
    
    public Double
    attributeDoubleN(String    name,
                     DoubleSet validSet)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return null;
        }
        return extractDoubleFromAttrValue(name, valueS, validSet);
    }
    
    public double
    attributeDoubleE(String name,
                     DoubleSet validSet)
    {
        String valueS = attributeValueE(name);
        return extractDoubleFromAttrValue(name, valueS, validSet);
    }
    
    public double
    attributeDoubleD(String name,
                     double defaultValue,
                     DoubleSet validSet)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return extractDoubleFromAttrValue(name, valueS, validSet);
    }
    
    public Enum<?>
    attributeEnumN(String   name,
                   Class<?> enumClass)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return null;
        }
        return extractEnumFromAttrValue(name, valueS, enumClass);
    }
    
    public Enum<?>
    attributeEnumE(String   name,
                   Class<?> enumClass)
    {
        String valueS = attributeValueE(name);
        return extractEnumFromAttrValue(name, valueS, enumClass);
    }
    
    public Enum<?>
    attributeEnumD(String   name,
                   Class<?> enumClass,
                   Enum<?>  defaultValue)
    {
        String valueS = attributeValueN(name);
        if (valueS == null)
        {
            return defaultValue;
        }
        return extractEnumFromAttrValue(name, valueS, enumClass);
    }
    
    public Object
    attributeStringMappedN(String       attrName,
                           Class<?>     stringMappedClass)
    {
        String valueS = attributeValueN(attrName);
        if (valueS == null)
        {
            return null;
        }
        return extractStringMappedFromAttrValue(attrName, valueS, stringMappedClass);
    }
    
    public Object
    attributeStringMappedE(String       attrName,
                           Class<?>     stringMappedClass)
    {
        String valueS = attributeValueE(attrName);
        return extractStringMappedFromAttrValue(attrName, valueS, stringMappedClass);
    }
    
    public Object
    attributeStringMappedD(String       attrName,
                           Class<?>     stringMappedClass,
                           Object       defaultValue)
    {
        String valueS = attributeValueN(attrName);
        if (valueS == null)
        {
            return defaultValue;
        }
        return extractStringMappedFromAttrValue(attrName, valueS, stringMappedClass);
    }
    
    public String
    attributeConstN(String attrName,
                    String expectedValue)
    {
        String valueS = attributeValueN(attrName);
        if (valueS == null)
        {
            return null;
        }
        return checkConstAttribute(attrName, valueS, expectedValue);
    }
    
    public String
    attributeConstE(String attrName,
                    String expectedValue)
    {
        String valueS = attributeValueE(attrName);
        return checkConstAttribute(attrName, valueS, expectedValue);
    }
    
    public String
    attributeConstD(String attrName,
                    String expectedValue)
    {
        String valueS = attributeValueN(attrName);
        if (valueS == null)
        {
            return expectedValue;
        }
        return checkConstAttribute(attrName, valueS, expectedValue);
    }
    
    protected int
    extractIntFromAttrValue(String attrName,
                            String attrValue)
    {
        try
        {
            return Integer.parseInt(attrValue);
        }
        catch (Exception e)
        {
            throw makeAttributeDoesNotMatchTypeException(attrName, attrValue, "int");
        }
    }
    
    protected int
    extractIntFromAttrValue(String attrName,
                            String attrValue,
                            IntSet validSet)
    {
        int attrInt = extractIntFromAttrValue(attrName, attrValue);
        if (!validSet.isIn(attrInt))
        {
            throw new XmlAttributeNotInSetException(_xmlContext, 
                    "Attribute " + attrName + " of " + this + " has int value " + attrValue +
                    " which is not in IntSet " + validSet);
        }
        return attrInt;
    }
    
    protected long
    extractLongFromAttrValue(String attrName,
                             String attrValue)
    {
        try
        {
            return Long.parseLong(attrValue);
        }
        catch (Exception e)
        {
            throw makeAttributeDoesNotMatchTypeException(attrName, attrValue, "long");
        }
    }
    
    protected long
    extractLongFromAttrValue(String  attrName,
                             String  attrValue,
                             LongSet validSet)
    {
        long attrLong = extractLongFromAttrValue(attrName, attrValue);
        if (!validSet.isIn(attrLong))
        {
            throw new XmlAttributeNotInSetException(_xmlContext, 
                    "Attribute " + attrName + " of " + this + " has long value " + attrValue +
                    " which is not in LongSet " + validSet);
        }
        return attrLong;
    }
    
    protected boolean
    extractBoolFromAttrValue(String attrName,
                             String attrValue)
    {
        if (attrValue.equalsIgnoreCase("true"))
        {
            return true;
        }
        else if (attrValue.equalsIgnoreCase("false"))
        {
            return false;
        }
        else
        {
            throw makeAttributeDoesNotMatchTypeException(attrName, attrValue, "boolean");
        }
    }
    
    protected float
    extractFloatFromAttrValue(String attrName,
                              String attrValue)
    {
        try
        {
            return Float.parseFloat(attrValue);
        }
        catch (Exception e)
        {
            throw makeAttributeDoesNotMatchTypeException(attrName, attrValue, "float");
        }
    }
    
    protected float
    extractFloatFromAttrValue(String   attrName,
                              String   attrValue,
                              FloatSet validSet)
    {
        float attrFloat = extractFloatFromAttrValue(attrName, attrValue);
        if (!validSet.isIn(attrFloat))
        {
            throw new XmlAttributeNotInSetException(_xmlContext, 
                    "Attribute " + attrName + " of " + this + " has float value " + attrValue +
                    " which is not in FloatSet " + validSet);
        }
        return attrFloat;
    }
    
    protected double
    extractDoubleFromAttrValue(String attrName,
                               String attrValue)
    {
        try
        {
            return Double.parseDouble(attrValue);
        }
        catch (Exception e)
        {
            throw makeAttributeDoesNotMatchTypeException(attrName, attrValue, "double");
        }
    }
    
    protected double
    extractDoubleFromAttrValue(String    attrName,
                               String    attrValue,
                               DoubleSet validSet)
    {
        double attrDouble = extractDoubleFromAttrValue(attrName, attrValue);
        if (!validSet.isIn(attrDouble))
        {
            throw new XmlAttributeNotInSetException(_xmlContext, 
                    "Attribute " + attrName + " of " + this + " has double value " + attrValue +
                    " which is not in DoubleSet " + validSet);
        }
        return attrDouble;
    }
    
    protected Enum<?>
    extractEnumFromAttrValue(String   attrName,
                             String   attrValue,
                             Class<?> enumClass)
    {
        if (!enumClass.isEnum())
        {
            throw new SuppliedClassNotEnumException(
                    _xmlContext, "Attempting to map value " + attrValue +
                    " of attribute " + attrName + " of " + this + " to enum but supplied class " +
                    enumClass.getName() + " is not an Enum class");
        }
        Enum<?> enumValue = stringToEnum(attrValue, enumClass);
        if (enumValue == null)
        {
            throw makeAttributeDoesNotMatchTypeException(attrName, attrValue, enumClass.getName());
        }
        return enumValue;
    }
    
    protected Object
    extractStringMappedFromAttrValue(String   attrName,
                                     String   attrValue,
                                     Class<?> stringMappedClass)
    {
        Constructor<?> constructor;
        try
        {
            constructor = stringMappedClass.getConstructor(String.class);
        }
        catch (SecurityException e)
        {
            throw new SuppliedClassNotStringMappedException(
                    _xmlContext,
                    "Attempting to make string-mapped instance from value " +
                    attrValue + " of attribute " + attrName + " of " + this +
                    " but the String constructor of supplied class " +
                    stringMappedClass.getName() +
                    " is not accessible, reflection exception is: " + e);
        }
        catch (NoSuchMethodException e)
        {
            throw new SuppliedClassNotStringMappedException(
                    _xmlContext,
                    "Attempting to make string-mapped instance from value " +
                    attrValue + " of attribute " + attrName + " of " + this +
                    " but supplied class " + stringMappedClass.getName() +
                    " has no String constructor or possibly an invisible constructor, reflection exception is: " + e);
        }
        Object stringMappedInstance;
        try
        {
            stringMappedInstance = constructor.newInstance(attrValue);
        }
        catch (InvocationTargetException e)
        {
            throw new XmlAttributeDoesNotMatchTypeException(_xmlContext,
                    "Attribute " + attrName + " of " + this + " has value " + attrValue +
                    " which can not be converted to " + stringMappedClass.getName() +
                    ", constructor throws " + e.getCause());
        }
        catch (Exception e)
        {
            throw new InternalBugException(_xmlContext, e.toString());
        }
        return stringMappedInstance;
    }
    
    protected String
    checkConstAttribute(String attrName,
                        String attrValue,
                        String expectedValue)
    {
        if (!attrValue.equals(expectedValue))
        {
            throw new XmlAttributeDoesNotMatchTypeException(
                    _xmlContext,
                    "Attribute " + attrName + " of " + this + " has value " + attrValue +
                    " which is not expected value " + expectedValue);
        }
        return attrValue;
    }
    
    protected XmlAttributeDoesNotMatchTypeException
    makeAttributeDoesNotMatchTypeException(String attrName,
                                           String attrValue,
                                           String typeLabel)
    {
        return new XmlAttributeDoesNotMatchTypeException(
                _xmlContext,
                "Attribute " + attrName + " of " + this + " has value " + attrValue +
                " which can not be converted to " + typeLabel);
    }
    
    /**
     * Converts String to Enum.
     * @param s         The string to convert
     * @param enumClass The class for which were are trying to find an instance.
     *                  It is a precondition that we already verified that the
     *                  class is an enum class
     * @return          The enum value that s matches (by toString) or null
     *                  if s does not match any enum value.
     */
    protected Enum<?>
    stringToEnum(String   s,
                 Class<?> enumClass)
    {

        try
        {
            Method method = enumClass.getMethod("values");
            Object returnObj = method.invoke(null);
            Enum<?>[] values = (Enum<?>[])returnObj;
            for (Enum<?> value : values)
            {
                String valueString = value.toString();
                if (valueString.equals(s))
                {
                    return value;
                }
            }
        }
        catch (Exception e)
        {
            // If reflection is coded properly, there will be no reflection based exception.
            throw new InternalBugException(_xmlContext, e.toString());
        }
        return null;
    }
}
