/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2012/2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.xml_basic.read;

import rasmus_torkel.set.chars.CharSet;
import rasmus_torkel.set.chars.DefinedCharSets;
import rasmus_torkel.xml_basic.rw.XmlChars;

public class XmlReadOptions
{
    /**
     * A typical pattern for XML data is that only leaf nodes contain text other than white space.
     * Other nodes only contain child tag nodes and irrelevant white text in between. When this
     * option is specified, this is automatically enforced.
     */
    public final boolean _enforceChildrenXorText;
    
    /**
     * Specifying this amounts to the using classes promising to retrieve from nodes child nodes
     * and text in order.
     * This allows tag nodes to drop references to child nodes after the child nodes have been
     * read (peeping does not count).
     * <p>
     * With this option the using classes should avoid certain operations, such as retrieving child
     * nodes by index.
     * <p>
     * This facilitates garbage collection which means that we don't have to have all XML in memory
     * at the same time.
     */
    public final boolean _orderedRetrieval;
    
    /**
     * This is the set of characters which the parser will accept for tag and attribute names,
     * provided that they are also valid name characters or, in the case of the first character
     * in a name, valid first name characters. However, like for value characters, the set must
     * not include the zero character.
     * <p>
     * To put this another way, when reading a tag or attribute name, any characters encountered
     * must match this set as well as the XML rules for names. But apart from the zero character,
     * there is no harm to specify characters which are not valid name chars. That way, users are
     * spared the inconvenience of having to specify the name char rules.
     */
    public final CharSet _nameCharSet;
    
    /**
     * This is the set of characters which the parser will accept for attribute values and
     * for text in text nodes. This set must not include the zero character, the only character
     * forbidden in all variants of XML.
     */
    public final CharSet _valueCharSet;
    
    /**
     * Different XML standards allow different character sets. However there is one absolute:
     * The zero character is absolutely prohibited. Everything else is potentially allowed.
     */
    public static final CharSet MOST_LENIENT_CHAR_SET = DefinedCharSets.BITS_16_NON_NULL;
    
    public static final XmlReadOptions DEFAULT =
        new XmlReadOptions(true, true, XmlChars.VALUE_CHAR_SET);
    
    public static final XmlReadOptions ANY_ORDER_RETRIEVAL =
        new XmlReadOptions(DEFAULT._enforceChildrenXorText, false, DEFAULT._valueCharSet);
    
    public static final XmlReadOptions ALLOW_SUBNODES_AND_TEXT =
        new XmlReadOptions(false, DEFAULT._orderedRetrieval, DEFAULT._valueCharSet);
    
    /**
     * Constructor.
     * @param enforceChildrenOrText @see _enforceChildrenOrText
     * @param orderedRetrieval      @see _orderedRetrieval
     * @param charSet               @see _nameCharSet and @see _valueCharSet . This set must not
     *                              allow the character with integer value zero.
     */
    public
    XmlReadOptions(boolean enforceChildrenXorText,
                   boolean orderedRetrieval,
                   CharSet charSet)
    {
        this(enforceChildrenXorText, orderedRetrieval, charSet, charSet);
    }
    
    /**
     * Constructor.
     * @param enforceChildrenOrText @see _enforceChildrenOrText
     * @param orderedRetrieval      @see _orderedRetrieval
     * @param nameCharSet           @see _nameCharSet This set must not allow the character
     *                              with integer value zero. Apart from that, it makes no
     *                              difference if illegal name characters are specified.
     * @param valueCharSet          @see _valueCharSet This set must not allow the character
     *                              with integer value zero.
     */
    public
    XmlReadOptions(boolean enforceChildrenXorText,
                   boolean orderedRetrieval,
                   CharSet nameCharSet,
                   CharSet valueCharSet)
    {
        _enforceChildrenXorText = enforceChildrenXorText;
        _orderedRetrieval = orderedRetrieval;
        _nameCharSet = nameCharSet;
        _valueCharSet = valueCharSet;
    }
    
    @Override
    public String
    toString()
    {
        return   (_enforceChildrenXorText ?
                      "enforce children or text" : "allow children and text interleaving")
               + ", "
               + (_orderedRetrieval ?
                      "ordered retrieval" : "arbitrary retrieval")
               + ", names in " + _nameCharSet
               + ", values in " + _valueCharSet;
    }
    
    public XmlReadOptions
    diffNameCharSet(CharSet nameCharSet)
    {
        return new XmlReadOptions(_enforceChildrenXorText, _orderedRetrieval, nameCharSet, _valueCharSet);
    }
    
    public XmlReadOptions
    diffValueCharSet(CharSet valueCharSet)
    {
        return new XmlReadOptions(_enforceChildrenXorText, _orderedRetrieval, _nameCharSet, valueCharSet);
    }
    
    public XmlReadOptions
    diffCharSet(CharSet charSet)
    {
        return new XmlReadOptions(_enforceChildrenXorText, _orderedRetrieval, charSet, charSet);
    }
}
