package rasmus_torkel.xml_basic.read.factory;

import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.HashSet;

import rasmus_torkel.convention.ConventionConstants;
import rasmus_torkel.text.read.exception.WrongNodeException;
import rasmus_torkel.text.write.IndentingLineSink;
import rasmus_torkel.xml_basic.read.TagNode;
import rasmus_torkel.xml_basic.read.exception.XmlTypedNodeException;
import rasmus_torkel.xml_basic.rw.TagNodeId;
import rasmus_torkel.xml_basic.rw.XmlNameSpace;
import rasmus_torkel.xml_basic.write.XmlSink;
import rasmus_torkel.xml_basic.write.XmlSinkWritable;

public abstract class MainXmlObjectFactory<T> extends XmlObjectFactory<T> implements XmlSinkWritable
{
    protected RuntimeException
    makeExtractAbstractException()
    {
        return new RuntimeException(
                "XmlObjectFactory.extractFromNode was called for " + _supportedClass.getName() +
                " which is abstract");
    }
    
    public final TagNodeId _naturalId;
    
    private MainXmlObjectFactory<?>            _superClassFactory;
    private String                             _rootReason;
    private ArrayList<MainXmlObjectFactory<?>> _subclassFactoryAl;
    
    protected
    MainXmlObjectFactory(TagNodeId naturalId)
    {
        super();
        _naturalId = naturalId;
        checkNaturalId();
        joinToInheritanceHierarchy();
    }
    
    protected
    MainXmlObjectFactory(Class<?>  supportedClass,
                         TagNodeId naturalId)
    {
        super(supportedClass);
        _naturalId = naturalId;
        checkNaturalId();
        joinToInheritanceHierarchy();
    }
    
    protected Class<?>
    determineSupportedClass()
    {
        Thread currentThread = Thread.currentThread();
        StackTraceElement[] elements = currentThread.getStackTrace();
        Class<?> supportedClass = null;
        Class<?> factoryClass = MainXmlObjectFactory.class;
        String   factorySubclassName = elements[1].getClassName();
        
        // We don't have to worry about the first four, they are
        // the call to getStackTrace, this function call,
        // the base class constructor call, and at least one subclass
        // constructor call.
        for (int i = 4; i < elements.length && supportedClass == null; i++)
        {
            String className = elements[i].getClassName();
            if (!className.contains("$")) // class is not anonymous
            {
                Class<?> clas;
                try
                {
                    clas = Class.forName(className);
                }
                catch (Throwable e)
                {
                    throw new RuntimeException("Class.forName(" + className + ") failed on class we got from stack");
                }
                if (!factoryClass.isAssignableFrom(clas)) // Not factory subclass constructor
                {
                    try
                    {
                        clas.getField("FROM_XML_FACTORY");
                        supportedClass = clas;
                    }
                    catch (Exception e)
                    {
                    }
                }
            }
        }
        if (supportedClass == null)
        {
            throw new RuntimeException("Failed to ascertain runtime class to be supported by factory of class " + factorySubclassName);
        }
        return supportedClass;
    }
    
    private void
    checkNaturalId()
    {
        int modifiers = _supportedClass.getModifiers();
        boolean isAbstract = Modifier.isAbstract(modifiers);
        if (_naturalId == null && !isAbstract)
        {
            throw new RuntimeException("XmlObjectFactory for class " + _supportedClass.getName() + " has null natural id but is not abstract");
        }
        else if (_naturalId != null && isAbstract)
        {
            throw new RuntimeException("XmlObjectFactory for class " + _supportedClass.getName() + " has non-null natural id but is abstract");
        }
    }
    
    private void
    joinToInheritanceHierarchy()
    {
        String[] messageHolder = new String[1];
        _superClassFactory = MainXmlObjectFactoryCache.INSTANCE.getSuperClassFactory(_supportedClass, messageHolder);
        if (_superClassFactory != null)
        {
            _superClassFactory.registerSubclassFactory(this);
        }
        else
        {
            _rootReason = messageHolder[0];
        }
    }
    
    private void
    registerSubclassFactory(MainXmlObjectFactory<?> subclassFactory)
    {
        if (_subclassFactoryAl == null)
        {
            _subclassFactoryAl = new ArrayList<MainXmlObjectFactory<?>>();
        }
        _subclassFactoryAl.add(subclassFactory);
    }
    
    public void
    makeInheritanceBranchReport(IndentingLineSink lineSink)
    {
        StringBuilder buf = new StringBuilder();
        buf.append(_supportedClass.getName());
        if (_naturalId == null)
        {
            buf.append(", abstract");
        }
        else
        {
            buf.append(", relativeName = ");
            buf.append(_naturalId._relativeName);
            if (_naturalId._nameSpace != null)
            {
                buf.append(", nameSpace = ");
                buf.append(_naturalId._nameSpace);
            }
        }
        if (_superClassFactory == null)
        {
            buf.append(", root: ");
            buf.append(_rootReason);
        }
        String line = buf.toString();
        lineSink.writeLine(line);
        if (_subclassFactoryAl != null)
        {
            lineSink.incrementLevel();
            for (MainXmlObjectFactory<?> subclassFactory : _subclassFactoryAl)
            {
                subclassFactory.makeInheritanceBranchReport(lineSink);
            }
            lineSink.decrementLevel();
        }
    }
    
    public MainXmlObjectFactory<?>
    root()
    {
        if (_superClassFactory == null)
        {
            return this;
        }
        else
        {
            return _superClassFactory.root();
        }
    }
    
    public static String
    makeInheritanceReport(Class<?>[] classes)
    {
        forceLoad(classes);
        IndentingLineSink lineSink = new IndentingLineSink();
        lineSink.writeLine("Inheritance report for class array");
        lineSink.incrementLevel();
        makeInheritanceReport(classes, lineSink);
        lineSink.decrementLevel();
        return lineSink.toString();
    }
    
    public static void
    forceLoad(Class<?>[] classes)
    {
        for (Class<?> clas : classes)
        {
            forceLoad(clas);
        }
    }
    
    public static void
    forceLoad(Class<?> clas)
    {
        String[] messageHolder = new String[1];
        MainXmlObjectFactoryCache.INSTANCE.getClassFactory(clas, messageHolder);
    }
    
    public static void
    makeInheritanceReport(Class<?>[]        classes,
                          IndentingLineSink lineSink)
    {
        HashSet<String> rootClassNameHs = new HashSet<String>();
        for (Class<?> clas : classes)
        {
            String[] messageHolder = new String[1];
            MainXmlObjectFactory<?> factory = MainXmlObjectFactoryCache.INSTANCE.getClassFactory(clas, messageHolder);
            if (factory == null)
            {
                lineSink.writeLine(clas.getName() + ", no factory: " + messageHolder[0]);
            }
            else
            {
                MainXmlObjectFactory<?> rootFactory = factory.root();
                String rootClassName = rootFactory._supportedClass.getName();
                if (!rootClassNameHs.contains(rootClassName))
                {
                    rootFactory.makeInheritanceBranchReport(lineSink);
                    rootClassNameHs.add(rootClassName);
                }
            }
        }
    }
    
    public String
    makeInheritanceBranchReport()
    {
        IndentingLineSink lineSink = new IndentingLineSink();
        lineSink.writeLine("Inheritance branch report");
        lineSink.incrementLevel();
        makeInheritanceBranchReport(lineSink);
        lineSink.decrementLevel();
        return lineSink.toString();
    }
    
    private MainXmlObjectFactory<?>
    getFactoryByNaturalId(TagNodeId expectedNaturalId)
    {
        if (expectedNaturalId == null)
        {
            throw new RuntimeException("getFactoryByNaturalId called with null naturalId");
        }
        ArrayList<MainXmlObjectFactory<?>> factoryAl = new ArrayList<MainXmlObjectFactory<?>>();
        getFactoriesByNaturalId(expectedNaturalId, factoryAl);
        int factoryQty = factoryAl.size();
        if (factoryQty == 0)
        {
            return null;
        }
        else if (factoryQty == 1)
        {
            return factoryAl.get(0);
        }
        else
        {
            StringBuilder buf = new StringBuilder();
            for (int i = 0; i < factoryQty; i++)
            {
                if (i > 0)
                {
                    buf.append(", ");
                }
                MainXmlObjectFactory<?> factory = factoryAl.get(i);
                String className = factory._supportedClass.getName();
                buf.append(className);
            }
            throw new RuntimeException(
                    "Trying to find MainXmlObjectFactory in hierarchy for " + _supportedClass.getName() +
                    " where naturalId has relativeName " + expectedNaturalId._relativeName +
                    " and nameSpace " + expectedNaturalId._nameSpace +
                    " but there are multiple matching classes: " + buf.toString());
        }
    }
    
    private void
    getFactoriesByNaturalId(TagNodeId                      expectedNaturalId,
                            ArrayList<MainXmlObjectFactory<?>> factoryAl)
    {
        if (expectedNaturalId.equals(_naturalId))
        {
            factoryAl.add(this);
        }
        if (_subclassFactoryAl != null)
        {
            for (MainXmlObjectFactory<?> subclassFactory : _subclassFactoryAl)
            {
                subclassFactory.getFactoriesByNaturalId(expectedNaturalId, factoryAl);
            }
        }
    }
    
    @SuppressWarnings("unchecked")
    public T
    nextOnParentN(TagNode      parentNode,
                  String       relativeName,
                  XmlNameSpace nameSpace)
    {
        TagNode node = nextChildNodeN(parentNode, relativeName, nameSpace);
        if (node == null)
        {
            return null;
        }
        String className = node.attributeValueN(ConventionConstants.LABEL_FOR_CLASS_NAME_OF_OBJECT);
        MainXmlObjectFactory<?> extractionFactory;
        if (className == null)
        {
            if (relativeName != null)
            {
                extractionFactory = this;
            }
            else
            {
                extractionFactory = getFactoryByNaturalId(node._id);
                if (extractionFactory == null)
                {
                    throw new WrongNodeException(
                            node._xmlContext,
                            "tag node id of " + node +
                            " does not match natural tag node id for any loaded subclass of " +
                            _supportedClass.getName());
                }
            }
        }
        else
        {
            extractionFactory = MainXmlObjectFactoryCache.INSTANCE.getFactory(node);
            if (!_supportedClass.isAssignableFrom(extractionFactory._supportedClass))
            {
                throw new XmlTypedNodeException(
                        node, className,
                        "the class is not a subclass of the supported type for this factory which is " +
                        _supportedClass.getName());
            }
        }
        Object obj = extractionFactory.extractFromNode(node);
        if (obj == null)
        {
            throw new RuntimeException(
                    "extractFromNode returned null for factory of type " +
                    extractionFactory.getClass().getName());
        }
        node.verifyNoMoreChildren();
        return (T)obj;
    }
    
    public void
    mostToXml(XmlSink      xmlSink,
              String       relativeName,
              XmlNameSpace nameSpace)
    {
        super.mostToXml(xmlSink, relativeName, nameSpace);
        xmlSink.sinkSimpleNode("naturalId", _naturalId);
    }
}
