package rasmus_torkel.xml_basic.read.factory;

import java.lang.reflect.Field;
import java.util.HashMap;

import rasmus_torkel.convention.ConventionConstants;
import rasmus_torkel.text.read.exception.InternalBugException;
import rasmus_torkel.xml_basic.read.TagNode;
import rasmus_torkel.xml_basic.read.exception.XmlTypedNodeException;

public class MainXmlObjectFactoryCache
{
    public static final MainXmlObjectFactoryCache INSTANCE = new MainXmlObjectFactoryCache();
    
    private HashMap<String,MainXmlObjectFactory<?>> _classNameToFactory = new HashMap<String, MainXmlObjectFactory<?>>();
    
    public synchronized MainXmlObjectFactory<?>
    getFactory(TagNode  node)
    {
        String className = node.attributeValueN(ConventionConstants.LABEL_FOR_CLASS_NAME_OF_OBJECT);
        if (className == null)
        {
            return null;
        }
        MainXmlObjectFactory<?> factory = _classNameToFactory.get(className);
        if (factory != null)
        {
            return factory;
        }
        Class<?> clas;
        try
        {
            clas = Class.forName(className);
        }
        catch (Throwable e)
        {
            Throwable cause = e;
            while (cause.getCause() != null)
            {
                cause = cause.getCause();
            }
            throw new XmlTypedNodeException(node, className, "the class could not be loaded: " + cause);
        }
        Field field;
        try
        {
            field = clas.getField("FROM_XML_FACTORY");
        }
        catch (NoSuchFieldException e)
        {
            throw new XmlTypedNodeException(node, className, "the class does not have a FROM_XML_FACTORY field");
        }
        catch (SecurityException e)
        {
            throw new XmlTypedNodeException(node, className, "the FROM_XML_FACTORY field of the class is not accessible");
        }
        Object factoryObj;
        try
        {
            factoryObj = field.get(null);
        }
        catch (IllegalArgumentException e)
        {
            throw new InternalBugException(node._xmlContext, "Did not expect " + e);
        }
        catch (IllegalAccessException e)
        {
            throw new XmlTypedNodeException(node, className, "the value of the FROM_XML_FACTORY field of the class is not accessible");
        }
        catch (NullPointerException e)
        {
            throw new XmlTypedNodeException(node, className, "the FROM_XML_FACTORY field of the class is not static");
        }
        if (factoryObj == null)
        {
            throw new XmlTypedNodeException(node, className, "the value of the FROM_XML_FACTORY field of the class is null");
        }
        if (factoryObj instanceof MainXmlObjectFactory)
        {
            factory = (MainXmlObjectFactory<?>)factoryObj;
        }
        else
        {
            throw new XmlTypedNodeException(
                    node, className,
                    "class of FROM_XML_FACTORY is " + factoryObj.getClass().getName() + " which not a subclass of " + MainXmlObjectFactory.class.getName());
        }
        _classNameToFactory.put(className, factory);
        return factory;
    }
    
    public MainXmlObjectFactory<?>
    getSuperClassFactory(Class<?> clas,
                         String[] messageHolder)
    {
        if (clas.equals(Object.class))
        {
            messageHolder[0] = "class is " + clas.getName();
            return null;
        }
        Class<?> superClass = clas.getSuperclass();
        String superClassName = superClass.getName();
        if (superClass.equals(Object.class))
        {
            messageHolder[0] = "super class is " + superClassName;
            return null;
        }
        MainXmlObjectFactory<?> superClassFactory = getClassFactory(superClass, messageHolder);
        if (superClassFactory == null)
        {
            messageHolder[0] = "super class " + messageHolder[0];
        }
        return superClassFactory;
    }
    
    public MainXmlObjectFactory<?>
    getClassFactory(Class<?> clas,
                    String[] messageHolder)
    {
        String className = clas.getName();
        MainXmlObjectFactory<?> classFactory = _classNameToFactory.get(className);
        if (classFactory != null)
        {
            return classFactory;
        }
        Field field;
        try
        {
            field = clas.getField("FROM_XML_FACTORY");
        }
        catch (NoSuchFieldException e)
        {
            messageHolder[0] = className + " does not have a FROM_XML_FACTORY field";
            return null;
        }
        catch (SecurityException e)
        {
            messageHolder[0] = className + " has a FROM_XML_FACTORY field which is not accessible";
            return null;
        }
        Object factoryObj;
        try
        {
            factoryObj = field.get(null);
        }
        catch (IllegalArgumentException e)
        {
            messageHolder[0] = className + " has a FROM_XML_FACTORY but when getting value we got unexpected " + e;
            return null;
        }
        catch (IllegalAccessException e)
        {
            messageHolder[0] = className + " has a FROM_XML_FACTORY but the value is not accessible";
            return null;
        }
        catch (NullPointerException e)
        {
            messageHolder[0] = className + " has a FROM_XML_FACTORY but it is not static";
            return null;
        }
        if (factoryObj == null)
        {
            messageHolder[0] = className + " has a FROM_XML_FACTORY but its value is null";
            return null;
        }
        if (factoryObj instanceof MainXmlObjectFactory)
        {
            classFactory = (MainXmlObjectFactory<?>)factoryObj;
        }
        else
        {
            messageHolder[0] = className + " has a FROM_XML_FACTORY of class " +
                               factoryObj.getClass().getName() + " which not a subclass of " +
                               MainXmlObjectFactory.class.getName();
            return null;
        }
        _classNameToFactory.put(className, classFactory);
        return classFactory;
    }
}
