package rasmus_torkel.xml_basic.read.factory;

import java.io.File;
import java.util.ArrayList;

import rasmus_torkel.misc.ArrayMaker;
import rasmus_torkel.xml_basic.read.TagNode;
import rasmus_torkel.xml_basic.read.XmlReadOptions;
import rasmus_torkel.xml_basic.read.impl.XmlReader;
import rasmus_torkel.xml_basic.rw.TagNodeId;
import rasmus_torkel.xml_basic.rw.XmlNameSpace;
import rasmus_torkel.xml_basic.write.XmlSink;
import rasmus_torkel.xml_basic.write.XmlSinkWritable;

public abstract class XmlObjectFactory<T> implements XmlSinkWritable
{
    public final Class<?>  _supportedClass;
    
    /**
     * Constructor which determines {@link XmlObjectFactory#_supportedClass}
     * without a parameter to populate it from.
     * Subclasses using this constructor need to supply their
     * own version of {@link XmlObjectFactory#determineSupportedClass()}.
     */
    protected
    XmlObjectFactory()
    {
        _supportedClass = determineSupportedClass();
    }
    
    protected
    XmlObjectFactory(Class<?>  supportedClass)
    {
        _supportedClass = supportedClass;
    }
    
    /**
     * Implementations must extract information from the node, but must NOT invoke
     * verifyNoMoreChildren on the node. verifyNoMoreChildren should not be called
     * more than once on the same node and we want to have clarity about whose
     * job it is and it's the job of other parts of the factory.
     * <p>
     * This arrangement also makes it easier to handle inheritance because
     * typically implementation functions will call the constructor of T and there
     * may or may not be subclasses of T which pull out additional fields.
     * @param node
     * @return
     */
    public abstract T
    extractFromNode(TagNode node);
    
    protected Class<?>
    determineSupportedClass()
    {
        throw new RuntimeException(this.getClass().getName() + "does not support determineSupportedClass");
    }
    
    /**
     * Finds the next node on parent.
     * @param parentNode   The node we are extracting the next node from 
     * @param relativeName Of the next node. Null means any node will do.
     * @param nameSpace    Of the next node. Null means no name space
     * @return The next node if present or null.
     */
    public T
    nextOnParentN(TagNode      parentNode,
                  String       relativeName,
                  XmlNameSpace nameSpace)
    {
        TagNode node = nextChildNodeN(parentNode, relativeName, nameSpace);
        T object = extractFromNode(node);
        node.verifyNoMoreChildren();
        return object;
    }
    
    protected TagNode
    nextChildNodeN(TagNode      parentNode,
                   String       relativeName,
                   XmlNameSpace nameSpace)
    {
        TagNode node;
        if (relativeName != null)
        {
            node = parentNode.nextChildN(relativeName, nameSpace);
        }
        else
        {
            node = parentNode.nextChildN();
        }
        return node;
    }
    
    public T
    nextOnParentN(TagNode      parentNode,
                  String       simpleNodeName)
    {
        return nextOnParentN(parentNode, simpleNodeName, null);
    }
    
    public T
    nextOnParentN(TagNode      parentNode,
                  TagNodeId    id)
    {
        return nextOnParentN(parentNode, id._relativeName, id._nameSpace);
    }
    
    public T
    nextOnParentE(TagNode      parentNode,
                  String       relativeName,
                  XmlNameSpace nameSpace)
    {
        TagNode node = parentNode.nextChildE(relativeName, nameSpace);
        return extractFromNode(node);
    }
    
    public T
    nextOnParentE(TagNode      parentNode,
                  String       simpleName)
    {
        return nextOnParentE(parentNode, simpleName, null);
    }
    
    public T
    nextOnParentE(TagNode      parentNode,
                  TagNodeId    id)
    {
        return nextOnParentE(parentNode, id._relativeName, id._nameSpace);
    }
    
    public T
    nextOnParentD(TagNode      parentNode,
                  String       relativeName,
                  XmlNameSpace nameSpace,
                  T            defaultValue)
    {
        TagNode node = parentNode.nextChildN(relativeName, nameSpace);
        if (node == null)
        {
            return defaultValue;
        }
        return extractFromNode(node);
    }
    
    public T
    nextOnParentD(TagNode      parentNode,
                  String       simpleName,
                  T            defaultValue)
    {
        return nextOnParentD(parentNode, simpleName, null, defaultValue);
    }
    
    public T
    nextOnParentD(TagNode      parentNode,
                  TagNodeId    id,
                  T            defaultValue)
    {
        return nextOnParentD(parentNode, id._relativeName, id._nameSpace, defaultValue);
    }
    
    public ArrayList<T>
    extractListFromNode(TagNode      arrayNode)
    {
        return extractListFromNode(arrayNode, null, null);
    }
    
    public ArrayList<T>
    extractListFromNode(TagNode      arrayNode,
                        TagNodeId    elementId)
    {
        return extractListFromNode(arrayNode, elementId._relativeName, elementId._nameSpace);
    }
    
    public ArrayList<T>
    extractListFromNode(TagNode      arrayNode,
                        String       elementSimpleName)
    {
        return extractListFromNode(arrayNode, elementSimpleName, null);
    }
    
    public ArrayList<T>
    extractListFromNode(TagNode      arrayNode,
                        String       elementRelativeName,
                        XmlNameSpace elementNameSpace)
    {
        ArrayList<T> elementAl = new ArrayList<T>();
        boolean finished = false;
        while (!finished)
        {
            T element = nextOnParentN(arrayNode, elementRelativeName, elementNameSpace);
            if (element == null)
            {
                finished = true;
            }
            else
            {
                elementAl.add(element);
            }
        }
        arrayNode.verifyNoMoreChildren();
        return elementAl;
    }
    
    public T[]
    extractArrayFromNode(TagNode       arrayNode,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list = extractListFromNode(arrayNode);
        return arrayMaker.fromList(list);
    }
    
    public T[]
    extractArrayFromNode(TagNode       arrayNode,
                         TagNodeId     elementId,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list = extractListFromNode(arrayNode, elementId);
        return arrayMaker.fromList(list);
    }
    
    public T[]
    extractArrayFromNode(TagNode       arrayNode,
                         String        elementRelativeName,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list = extractListFromNode(arrayNode, elementRelativeName);
        return arrayMaker.fromList(list);
    }
    
    public T[]
    extractArrayFromNode(TagNode       arrayNode,
                         String        elementRelativeName,
                         XmlNameSpace  elementNameSpace,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list = extractListFromNode(arrayNode, elementRelativeName, elementNameSpace);
        return arrayMaker.fromList(list);
    }
    
    //-----------------------------------------------------
    // The nextArrayFromParentN functions could be written more simply and less repetitively but
    // doing it like this means we don't need separate tests for the corresponding list functions.
    
    public T[]
    nextArrayFromParentN(TagNode       arrayParentNode,
                         TagNodeId     arrayId,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list = nextListFromParentN(arrayParentNode, arrayId);
        if (list == null)
        {
            return null;
        }
        return arrayMaker.fromList(list);
    }
    
    public T[]
    nextArrayFromParentN(TagNode       arrayParentNode,
                         String        arrayName,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list = nextListFromParentN(arrayParentNode, arrayName);
        if (list == null)
        {
            return null;
        }
        return arrayMaker.fromList(list);
    }
    
    public T[]
    nextArrayFromParentN(TagNode       arrayParentNode,
                         String        arrayRelativeName,
                         XmlNameSpace  arrayNameSpace,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list = nextListFromParentN(arrayParentNode, arrayRelativeName, arrayNameSpace);
        if (list == null)
        {
            return null;
        }
        return arrayMaker.fromList(list);
    }
    
    public T[]
    nextArrayFromParentN(TagNode       arrayParentNode,
                         TagNodeId     arrayId,
                         TagNodeId     elementId,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list = nextListFromParentN(arrayParentNode, arrayId, elementId);
        if (list == null)
        {
            return null;
        }
        return arrayMaker.fromList(list);
    }
    
    public T[]
    nextArrayFromParentN(TagNode       arrayParentNode,
                         String        arrayName,
                         String        elementName,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list = nextListFromParentN(arrayParentNode, arrayName, elementName);
        if (list == null)
        {
            return null;
        }
        return arrayMaker.fromList(list);
    }
    
    public T[]
    nextArrayFromParentN(TagNode       arrayParentNode,
                         String        arrayRelativeName,
                         XmlNameSpace  arrayNameSpace,
                         String        elementRelativeName,
                         XmlNameSpace  elementNameSpace,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list =
                nextListFromParentN(
                        arrayParentNode,
                        arrayRelativeName, arrayNameSpace, elementRelativeName, elementNameSpace);
        if (list == null)
        {
            return null;
        }
        return arrayMaker.fromList(list);
    }
    
    //------------------------------------------------------------
    // The nextArrayFromParentE are written this way so that we don't need separate tests
    // for the corresponding list functions.
    
    public T[]
    nextArrayFromParentE(TagNode       arrayParentNode,
                         TagNodeId     arrayId,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list = nextListFromParentE(arrayParentNode, arrayId);
        return arrayMaker.fromList(list);
    }
    
    public T[]
    nextArrayFromParentE(TagNode       arrayParentNode,
                         String        arrayName,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list = nextListFromParentE(arrayParentNode, arrayName);
        return arrayMaker.fromList(list);
    }
    
    public T[]
    nextArrayFromParentE(TagNode       arrayParentNode,
                         String        arrayRelativeName,
                         XmlNameSpace  arrayNameSpace,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list = nextListFromParentE(arrayParentNode, arrayRelativeName, arrayNameSpace);
        return arrayMaker.fromList(list);
    }
    
    public T[]
    nextArrayFromParentE(TagNode       arrayParentNode,
                         TagNodeId     arrayId,
                         TagNodeId     elementId,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list = nextListFromParentE(arrayParentNode, arrayId, elementId);
        return arrayMaker.fromList(list);
    }
    
    public T[]
    nextArrayFromParentE(TagNode       arrayParentNode,
                         String        arrayName,
                         String        elementName,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list = nextListFromParentE(arrayParentNode, arrayName, elementName);
        return arrayMaker.fromList(list);
    }
    
    public T[]
    nextArrayFromParentE(TagNode       arrayParentNode,
                         String        arrayRelativeName,
                         XmlNameSpace  arrayNameSpace,
                         String        elementRelativeName,
                         XmlNameSpace  elementNameSpace,
                         ArrayMaker<T> arrayMaker)
    {
        ArrayList<T> list =
                nextListFromParentE(
                        arrayParentNode,
                        arrayRelativeName, arrayNameSpace, elementRelativeName, elementNameSpace);
        return arrayMaker.fromList(list);
    }
    
    //-------------------------------------------------------------------------------------------
    
    public ArrayList<T>
    nextListFromParentN(TagNode      listParentNode,
                        String       listRelativeName,
                        XmlNameSpace listNameSpace)
    {
        TagNode listNode = listParentNode.nextChildN(listRelativeName, listNameSpace);
        if (listNode == null)
        {
            return null;
        }
        return extractListFromNode(listNode);
    }
    
    public ArrayList<T>
    nextListFromParentN(TagNode      listParentNode,
                        String       listSimpleName)
    {
        return nextListFromParentN(listParentNode, listSimpleName, (XmlNameSpace)null);
    }
    
    public ArrayList<T>
    nextListFromParentN(TagNode      listParentNode,
                        TagNodeId    listId)
    {
        return nextListFromParentN(listParentNode,
                                   listId._relativeName,
                                   listId._nameSpace);
    }
    
    public ArrayList<T>
    nextListFromParentN(TagNode      listParentNode,
                        String       listRelativeName,
                        XmlNameSpace listNameSpace,
                        String       elementRelativeName,
                        XmlNameSpace elementNameSpace)
    {
        TagNode listNode = listParentNode.nextChildN(listRelativeName, listNameSpace);
        if (listNode == null)
        {
            return null;
        }
        return extractListFromNode(listNode, elementRelativeName, elementNameSpace);
    }
    
    public ArrayList<T>
    nextListFromParentN(TagNode      listParentNode,
                        String       listSimpleName,
                        String       elementSimpleName)
    {
        return nextListFromParentN(listParentNode, listSimpleName, null, elementSimpleName, null);
    }
    
    public ArrayList<T>
    nextListFromParentN(TagNode      listParentNode,
                        TagNodeId    listId,
                        TagNodeId    elementId)
    {
        return nextListFromParentN(listParentNode,
                                   listId._relativeName,
                                   listId._nameSpace,
                                   elementId._relativeName,
                                   elementId._nameSpace);
    }
    
    //-------------------------------------------------------------------------------------
    
    public ArrayList<T>
    nextListFromParentE(TagNode      listParentNode,
                        String       listRelativeName,
                        XmlNameSpace listNameSpace,
                        String       elementRelativeName,
                        XmlNameSpace elementNameSpace)
    {
        TagNode listNode = listParentNode.nextChildE(listRelativeName, listNameSpace);
        return extractListFromNode(listNode, elementRelativeName, elementNameSpace);
    }
    
    public ArrayList<T>
    nextListFromParentE(TagNode      listParentNode,
                        String       listSimpleName,
                        String       elementSimpleName)
    {
        return nextListFromParentE(listParentNode, listSimpleName, null, elementSimpleName, null);
    }
    
    public ArrayList<T>
    nextListFromParentE(TagNode      listParentNode,
                        TagNodeId    listId,
                        TagNodeId    elementId)
    {
        return nextListFromParentE(listParentNode,
                                   listId._relativeName,
                                   listId._nameSpace,
                                   elementId._relativeName,
                                   elementId._nameSpace);
    }
    
    public ArrayList<T>
    nextListFromParentE(TagNode      listParentNode,
                        String       listRelativeName,
                        XmlNameSpace listNameSpace)
    {
        TagNode listNode = listParentNode.nextChildE(listRelativeName, listNameSpace);
        return extractListFromNode(listNode);
    }
    
    public ArrayList<T>
    nextListFromParentE(TagNode      listParentNode,
                        String       listSimpleName)
    {
        return nextListFromParentE(listParentNode, listSimpleName, (XmlNameSpace)null);
    }
    
    public ArrayList<T>
    nextListFromParentE(TagNode      listParentNode,
                        TagNodeId    listId)
    {
        return nextListFromParentE(listParentNode,
                                   listId._relativeName,
                                   listId._nameSpace);
    }
    
    public T
    stringToObject(String              xmlString,
                   String              context)
    {
        return stringToObject(xmlString, context, XmlReadOptions.DEFAULT);
    }
    
    public T
    stringToObject(String              xmlString,
                   String              context,
                   XmlReadOptions      options)
    {
        TagNode node = XmlReader.xmlStringToRoot(xmlString, context, options);
        try
        {
            T object = extractFromNode(node);
            node.verifyNoMoreChildren();
            return object;
        }
        finally
        {
            node.closeReader();
        }
    }
    
    public T[]
    stringToArray(String              xmlString,
                  String              context,
                  ArrayMaker<T>       arrayMaker,
                  TagNodeId           arrayId)
    {
        return stringToArray(xmlString, context, arrayMaker, XmlReadOptions.DEFAULT, arrayId);
    }
    
    public T[]
    stringToArray(String              xmlString,
                  String              context,
                  ArrayMaker<T>       arrayMaker,
                  String              arrayName)
    {
        return stringToArray(xmlString, context, arrayMaker, XmlReadOptions.DEFAULT, arrayName);
    }
    
    public T[]
    stringToArray(String              xmlString,
                  String              context,
                  ArrayMaker<T>       arrayMaker,
                  String              arrayRelativeName,
                  XmlNameSpace        arrayNameSpace)
    {
        return stringToArray(xmlString, context, arrayMaker, XmlReadOptions.DEFAULT, arrayRelativeName, arrayNameSpace);
    }
    
    public T[]
    stringToArray(String              xmlString,
                  String              context,
                  ArrayMaker<T>       arrayMaker,
                  XmlReadOptions      options,
                  TagNodeId           arrayId)
    {
        return stringToArray(
                xmlString, context, arrayMaker, options, arrayId._relativeName, arrayId._nameSpace);
    }
    
    public T[]
    stringToArray(String              xmlString,
                  String              context,
                  ArrayMaker<T>       arrayMaker,
                  XmlReadOptions      options,
                  String              arrayName)
    {
        return stringToArray(xmlString, context, arrayMaker, options, arrayName, null);
    }
    
    public T[]
    stringToArray(String              xmlString,
                  String              context,
                  ArrayMaker<T>       arrayMaker,
                  XmlReadOptions      options,
                  String              arrayRelativeName,
                  XmlNameSpace        arrayNameSpace)
    {
        TagNode node = XmlReader.xmlStringToRoot(xmlString, context, options, arrayRelativeName, arrayNameSpace);
        try
        {
            T[] array = extractArrayFromNode(node, arrayMaker);
            return array;
        }
        finally
        {
            node.closeReader();
        }
    }
    
    public T
    fileToObject(File xmlFile)
    {
        return fileToObject(xmlFile, XmlReadOptions.DEFAULT);
    }
    
    public T
    fileToObject(File                xmlFile,
                 XmlReadOptions      options)
    {
        TagNode node = XmlReader.xmlFileToRoot(xmlFile, options);
        try
        {
            T object = extractFromNode(node);
            node.verifyNoMoreChildren();
            return object;
        }
        finally
        {
            node.closeReader();
        }
    }
    
    public T[]
    fileToArray(File                xmlFile,
                ArrayMaker<T>       arrayMaker,
                TagNodeId           arrayId)
    {
        return fileToArray(xmlFile, arrayMaker, XmlReadOptions.DEFAULT, arrayId);
    }
    
    public T[]
    fileToArray(File                xmlFile,
                ArrayMaker<T>       arrayMaker,
                String              arrayName)
    {
        return fileToArray(xmlFile, arrayMaker, XmlReadOptions.DEFAULT, arrayName);
    }
    
    public T[]
    fileToArray(File                xmlFile,
                ArrayMaker<T>       arrayMaker,
                String              arrayRelativeName,
                XmlNameSpace        arrayNameSpace)
    {
        return fileToArray(xmlFile, arrayMaker, XmlReadOptions.DEFAULT, arrayRelativeName, arrayNameSpace);
    }
    
    public T[]
    fileToArray(File                xmlFile,
                ArrayMaker<T>       arrayMaker,
                XmlReadOptions      options,
                TagNodeId           arrayId)
    {
        return fileToArray(xmlFile, arrayMaker, options, arrayId._relativeName, arrayId._nameSpace);
    }
    
    public T[]
    fileToArray(File                xmlFile,
                ArrayMaker<T>       arrayMaker,
                XmlReadOptions      options,
                String              arrayName)
    {
        return fileToArray(xmlFile, arrayMaker, options, arrayName, (XmlNameSpace)null);
    }
    
    public T[]
    fileToArray(File                xmlFile,
                ArrayMaker<T>       arrayMaker,
                XmlReadOptions      options,
                String              arrayRelativeName,
                XmlNameSpace        arrayNameSpace)
    {
        TagNode node = XmlReader.xmlFileToRoot(xmlFile, options, arrayRelativeName, arrayNameSpace);
        try
        {
            T[] array = extractArrayFromNode(node, arrayMaker);
            return array;
        }
        finally
        {
            node.closeReader();
        }
    }
    
    @Override
    public TagNodeId
    naturalTagNodeId()
    {
        return new TagNodeId(getClass().getSimpleName());
    }
    
    @Override
    public void
    mostToXml(XmlSink      xmlSink,
              String       relativeName,
              XmlNameSpace nameSpace)
    {
        xmlSink.sinkClassOfObjectAttribute(this);
        xmlSink.sinkSimpleNode("supportedClass", _supportedClass.getName());
    }
}
