/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2012/2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.xml_basic.read.impl;

import java.util.ArrayList;

import rasmus_torkel.text.read.ROTextPosition;
import rasmus_torkel.text.read.exception.DuplicateNodeException;
import rasmus_torkel.text.read.exception.EndOfNodeWronglyAssumedException;
import rasmus_torkel.text.read.exception.SurplusNodeException;
import rasmus_torkel.text.read.exception.UnorderedRetrievalException;
import rasmus_torkel.xml_basic.read.SupportedNode;
import rasmus_torkel.xml_basic.read.TagNode;
import rasmus_torkel.xml_basic.read.TextNode;
import rasmus_torkel.xml_basic.read.exception.XmlTextAndChildException;
import rasmus_torkel.xml_basic.rw.TagNodeId;
import rasmus_torkel.xml_basic.rw.XmlAttribute;
import rasmus_torkel.xml_basic.rw.XmlNameSpace;

/**
 * Root node or element node in regular XML Jargon.
 * 
 * @author Rasmus Torkel
 *
 */
public class TagNodeImpl extends TagNode
{
    // No HashMap for value look-up to save space.
    // Hopefully there wont be ridiculously long attribute lists.
    private XmlAttribute[]         _attributes;
    
    // Even indexes contain text nodes, odd indexes contain child nodes.
    private ArrayList<SupportedNode> _subnodeAl = new ArrayList<SupportedNode>();
    
    // Represents last thing retrieved by invoker.
    // 2 * n means invoker retrieved child n - 1
    // 2 * n + 1 means invoker retrieved text n - 1
    // Examples:
    // North cursor  The most forward thing retrieved
    // 0             nothing
    // 1             first text (text 0)
    // 2             first child node
    // 3             second text
    // 4             second child node
    private int                 _northCursor;
    
    private final boolean       _selfClosing;
    private boolean             _southFinished;
    private XmlReader           _reader;
    private ROTextPosition      _endPos;
    private ROTextPosition      _firstNonWhiteTextPos;
    private ROTextPosition      _firstChildPos;
    private boolean             _orderedRetrieval;
    
    TagNodeImpl(ROTextPosition textPos,
                TagNodeId      id,
                XmlAttribute[] attributes,
                TagNodeImpl    parent,
                boolean        selfClosing,
                XmlReader      reader)
    {
        super(textPos, reader._context, id, parent);
        _attributes = attributes;
        _selfClosing = selfClosing;
        _reader = reader;
        if (selfClosing)
        {
            _southFinished = true;
            _endPos = _reader.textPosition();
        }
        _orderedRetrieval = reader._readOptions._orderedRetrieval;
    }
    
    @Override
    public String
    toString()
    {
        return _id + ", starting at " + _textPos;
    }
    
    @Override
    public void
    closeReader()
    {
        _reader.close();
    }
    
    @Override
    public boolean
    isReaderOpen()
    {
        return _reader.isOpen();
    }
    
    @Override
    public void
    setOrderedRetrieval()
    {
        _orderedRetrieval = true;
    }
    
    @Override
    public int
    attributeQty()
    {
        return _attributes.length;
    }
    
    @Override
    public XmlAttribute
    attribute(int index)
    {
        return _attributes[index];
    }
    
    /*
    public String
    attributeValueN(String name)
    {
        for (int i = 0; i < _attributes.length; i++)
        {
            if (   name.equals(_attributes[i]._name)
                && _attributes[i]._nameSpace == null)
            {
                return _attributes[i]._value;
            }
        }
        return null;
    }*/
    
    public String
    attributeValueN(String       relativeName,
                    XmlNameSpace nameSpace)
    {
        for (int i = 0; i < _attributes.length; i++)
        {
            if (   relativeName.equals(_attributes[i]._name)
                && XmlNameSpace.isSame(nameSpace, _attributes[i]._nameSpace))
            {
                return _attributes[i]._value;
            }
        }
        return null;
    }
    
    /**
     * 
     * @return
     */
    public TextNode
    nextTextNodeN()
    {
        ROTextPosition readerPos = _reader.textPosition();
        if (   _orderedRetrieval
            && _endPos != null
            && _endPos.compareTo(readerPos) < 0)
        {
            if (_reader._readOptions._orderedRetrieval)
            {
                throw new UnorderedRetrievalException(
                        _xmlContext,
                        "Attempting to get text node within node " + this +
                        " which ended at " + this._endPos +
                        " while reading is already further ahead at " + readerPos +
                        "  and this contravenes the read options");
            }
            else
            {
                throw new UnorderedRetrievalException(
                        _xmlContext,
                        "Attempting to get text node within node " + this +
                        " which ended at " + this._endPos +
                        " while reading is already further ahead at " + readerPos +
                        "  and this is contrary to what the user specified for this node");
            }
        }
        int northTextQty = getNorthTextQty();
        int requiredSouthTextQty = northTextQty + 1;
        int requiredSubnodeQty = requiredSouthTextQty * 2 - 1;
        if (!ensureSouthQty(requiredSubnodeQty))
        {
            return null;
        }
        int nextTextIndex = requiredSubnodeQty - 1;
        TextNode text = (TextNode)_subnodeAl.get(nextTextIndex);
        _northCursor = requiredSubnodeQty;
        return text;
    }
    
    public TagNode
    nextChildN()
    {
        int northChildQty = getNorthChildQty();
        TagNode child = peepChildN();
        if (child != null)
        {
            _northCursor = 2 * northChildQty + 2;
        }
        else if (!_selfClosing)
        {
            // After the text after the last child node which we would have seen before we
            // realised that there are no more child nodes.
            _northCursor = 2 * northChildQty + 1;
        }
        return child;
    }
    
    public TagNode
    peepChildN()
    {
        ROTextPosition readerPos = _reader.textPosition();
        if (   _orderedRetrieval
            && _endPos != null
            && _endPos.compareTo(readerPos) < 0)
        {
            if (_reader._readOptions._orderedRetrieval)
            {
                throw new UnorderedRetrievalException(
                        _xmlContext,
                        "Attempting to get child node within node " + this +
                        " which ended at " + this._endPos +
                        " while reading is already further ahead at " + readerPos +
                        "  and this contravenes the read options");
            }
            else
            {
                throw new UnorderedRetrievalException(
                        _xmlContext,
                        "Attempting to get child node within node " + this +
                        " which ended at " + this._endPos +
                        " while reading is already further ahead at " + readerPos +
                        "  and this is contrary to what the invoker specified for this node");
            }
        }
        int northChildQty = getNorthChildQty();
        int requiredSouthChildQty = northChildQty + 1;
        int requiredSubnodeQty = requiredSouthChildQty * 2;
        if (!ensureSouthQty(requiredSubnodeQty))
        {
            return null;
        }
        int peepChildIndex = requiredSubnodeQty - 1;
        TagNode child = (TagNode)_subnodeAl.get(peepChildIndex);
        return child;
    }
    
    private int
    getNorthChildQty()
    {
        return _northCursor / 2;
    }
    
    private int
    getNorthTextQty()
    {
        return (_northCursor + 1) / 2;
    }
    
    public void
    verifyNoMoreChildren()
    {
        TagNode child = nextChildN();
        if (child != null)
        {
            throw new SurplusNodeException(_xmlContext,
                    "When verifyNoMoreChildren was invoked on " + this +
                    ", there was child node " + child);
        }
    }
    
    public void
    verifyNoMoreText()
    {
        TextNode textNode = nextTextNodeN();
        if (textNode != null)
        {
            throw new SurplusNodeException(_xmlContext,
                    "When verifyNoMoreText was invoked on " + this +
                    ", there was more text at " + textNode._textPos +
                    " which implies the existence of a surplus node");
        }
    }
    
    public ROTextPosition
    endPosition()
    {
        if (_endPos == null)
        {
            throw new EndOfNodeWronglyAssumedException(_xmlContext,
                    "endPosition was invoked on " + this + " before end of node was reached");
        }
        return _endPos;
    }
    
    private TextNode
    readText()
    {
        TextNode textNode = _reader.readText();
        boolean isWhiteSpace = XmlReader.isWhiteSpace(textNode._text);
        if (   _reader._readOptions._enforceChildrenXorText
            && _firstChildPos != null
            && !isWhiteSpace)
        {
            throw new XmlTextAndChildException(
                    _xmlContext, this, _firstChildPos, textNode);
        }
        if (!isWhiteSpace)
        {
            _firstNonWhiteTextPos = textNode._textPos;
        }

        _subnodeAl.add(textNode);
        return textNode;
    }
    
    private TagNode
    readChild()
    {
        TagNode child = _reader.getTagNode();
        if (child != null)
        {
            if (_firstNonWhiteTextPos != null && _reader._readOptions._enforceChildrenXorText)
            {
                throw new XmlTextAndChildException(
                        _xmlContext, this, _firstNonWhiteTextPos, child);
            }
            if (_firstChildPos == null)
            {
                _firstChildPos = child._textPos;
            }
            _subnodeAl.add(child);
        }
        else
        {
            _southFinished = true;
            _endPos = _reader.textPosition();
        }
        return child;
    }
    
    public void
    readRest()
    {
        ensureSouthQty(Integer.MAX_VALUE); // This will stop when we reach the end of our node
        if (_orderedRetrieval)
        {
            setLastNull();
        }
    }
    
    private void
    readLastRest()
    {
        int southQty = _subnodeAl.size();
        if (southQty >= 1)
        {
            int lastIndex = southQty - 1;
            if (isChildIndex(lastIndex))
            {
                TagNode child = (TagNode)_subnodeAl.get(lastIndex);
                child.readRest();
            }
        }
    }
    
    private void
    setLastNull()
    {
        int southQty = _subnodeAl.size();
        if (southQty >= 1)
        {
            int lastIndex = southQty - 1;
            _subnodeAl.set(lastIndex, null);
        }
    }
    
    private boolean
    ensureSouthQty(int desiredSouthQty)
    {
        int currentSouthQty = _subnodeAl.size();
        if (_southFinished)
        {
            return currentSouthQty >= desiredSouthQty;
        }
        for (int i = currentSouthQty; i < desiredSouthQty; i++)
        {
            SupportedNode subnode = ensureSouthOne(i);
            if (subnode == null)
            {
                return false;
            }
        }
        return true;
    }
    
    private SupportedNode
    ensureSouthOne(int index)
    {
        readLastRest();
        if (_orderedRetrieval)
        {
            setLastNull();
        }
        SupportedNode subnode;
        if (isChildIndex(index))
        {
            subnode = readChild();
        }
        else
        {
            subnode = readText();
        }
        return subnode;
    }
    
    public TagNode
    anyOrderChildN(String       relativeName,
                   XmlNameSpace nameSpace)
    {
        if (_reader._readOptions._orderedRetrieval)
        {
            throw new UnorderedRetrievalException(
                _xmlContext, "anyOrder operation invoked with ordered retrieval read option");
        }
        else if (_orderedRetrieval)
        {
            throw new UnorderedRetrievalException(
                _xmlContext, "anyOrder operation invoked despite invoker specifying order retrieval for this node");
        }
        ensureSouthQty(Integer.MAX_VALUE); // will obviously stop once we reach end
        int southQty = _subnodeAl.size();
        TagNode matchingChild = null;
        for (int i = 0; i < southQty; i++)
        {
            if (isChildIndex(i))
            {
                TagNode child = (TagNode)_subnodeAl.get(i);
                if (child.idMatches(relativeName, nameSpace))
                {
                    if (matchingChild == null)
                    {
                        _northCursor = i + 1;
                        matchingChild = child;
                    }
                    else
                    {
                        throw new DuplicateNodeException(_xmlContext,
                                this + " contains duplicate child nodes " + matchingChild +
                                " and " + child);
                    }
                }
            }
        }
        return matchingChild;
    }
    
    private boolean
    isChildIndex(int index)
    {
        return index % 2 == 1;
    }
    
    @Override
    public void
    verifyNoOtherChildren(TagNodeId... expectedIds)
    {
        if (_reader._readOptions._orderedRetrieval)
        {
            throw new UnorderedRetrievalException(
                _xmlContext, "verifyNoOtherChildren operation invoked with ordered retrieval read option");
        }
        else if (_orderedRetrieval)
        {
            throw new UnorderedRetrievalException(
                _xmlContext, "verifyNoOtherChildren operation despite invoker specifying ordered retrieval for this node");
        }
        readRest();
        int southQty = _subnodeAl.size();
        for (int childIndex = 1; childIndex < southQty; childIndex += 2)
        {
            TagNode child = (TagNode)_subnodeAl.get(childIndex);
            boolean childMatches = false;
            for (int expectedIdIndex = 0;
                 expectedIdIndex < expectedIds.length && !childMatches;
                 expectedIdIndex++)
            {
                TagNodeId expectedId = expectedIds[expectedIdIndex];
                if (child.idMatches(expectedId))
                {
                    childMatches = true;
                }
            }
            if (!childMatches)
            {
                throw new SurplusNodeException(_xmlContext, this + " has unexpected child " + child);
            }
        }
    }
}
