/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2012/2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.xml_basic.read.impl;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.util.ArrayList;

import rasmus_torkel.misc.CharUtil;
import rasmus_torkel.text.read.LookAheadReader;
import rasmus_torkel.text.read.ROTextPosition;
import rasmus_torkel.text.read.RWTextPosition;
import rasmus_torkel.text.read.exception.IncorrectInvocationReadException;
import rasmus_torkel.text.read.exception.InternalBugException;
import rasmus_torkel.text.read.exception.ReadIoException;
import rasmus_torkel.text.read.exception.WrongNodeException;
import rasmus_torkel.xml_basic.read.XmlReadOptions;
import rasmus_torkel.xml_basic.read.TagNode;
import rasmus_torkel.xml_basic.read.TextNode;
import rasmus_torkel.xml_basic.read.XmlNameSpaceAtLevel;
import rasmus_torkel.xml_basic.read.exception.XmlCharSetException;
import rasmus_torkel.xml_basic.read.exception.XmlSyntaxException;
import rasmus_torkel.xml_basic.rw.TagNodeId;
import rasmus_torkel.xml_basic.rw.XmlAttribute;
import rasmus_torkel.xml_basic.rw.XmlChars;
import rasmus_torkel.xml_basic.rw.XmlNameSpace;

/**
 * This is the class that is the starting point of XML reading. It has various static functions
 * for getting a root node. Invokers typically will not handle any objects of this class but
 * only nodes. However, the nodes have a reference back to object of this class that created them.
 * <p>
 * On closing the XmlReader (which leads to closing the reader for actual characters):
 * <ul>
 *   <li>If the XmlReader fails to get a root node, it's its responsibility to close automatically.</li>
 *   <li>If reading progresses all the way to the end, then closing will be automatic. However,
 *       it is not good enough to read the last child node of the root node. We actually need
 *       to look for another child node of the root node and not find it.</li>
 *   <li>Otherwise it's the invokers responsibility to close which may be done via the
 *       {@link rasmus_torkel.xml_basic.read.TagNode.closeReader()} function on any node.</li>
 *   <li>There is no harm in invoking the closing multiple times.</li>
 * </ul>
 * 
 * @author Torkel
 */
public class XmlReader
{
    private final LookAheadReader     _reader;
            final String              _context;
    private final RWTextPosition.View _textPos;
            final XmlReadOptions      _readOptions;
    
    private boolean            _gotRootNode;
    private ArrayList<TagNodeImpl> _tagStack = new ArrayList<TagNodeImpl>();
    private ArrayList<XmlNameSpaceAtLevel> _nameSpaceStack = new ArrayList<XmlNameSpaceAtLevel>();
    
    private static final String CSECTION_START = "<![CDATA[";
    private static final String CSECTION_END = "]]>";
    private static final String NAMESPACE_DECL_PREFIX = XmlNameSpace.XMLNS._prefix + ":";
    
    private static class TmpAttribute
    {
        final String _name;
        final String _value;
        
        public
        TmpAttribute(String name,
                     String value)
        {
            _name = name;
            _value = value;
        }
    }
    
    public
    XmlReader(Reader      reader,
              String      context,
              XmlReadOptions readOptions)
    {
        int biggestTokenSize = CSECTION_START.length();
        _reader = new LookAheadReader(reader, biggestTokenSize);
        _textPos = _reader._textPosView;
        _context = context;
        _readOptions = readOptions;
        stackInitialNameSpace(XmlNameSpace.XMLNS);
        stackInitialNameSpace(XmlNameSpace.XML);
    }
    
    private void
    stackInitialNameSpace(XmlNameSpace nameSpace)
    {
        XmlNameSpaceAtLevel nameSpaceAtLevel =
                new XmlNameSpaceAtLevel(nameSpace._prefix, nameSpace._name, -1);
        _nameSpaceStack.add(nameSpaceAtLevel);
    }
    
    /**
     * Gets non-name spaced root node.
     * @param name
     * @return 
     */
    public TagNodeImpl
    getRootNode(String name)
    {
        return getRootNode(name, null);
    }
    
    public TagNodeImpl
    getRootNode(String       relativeName,
                XmlNameSpace nameSpace)
    {
        TagNodeImpl rootNode = getRootNode();
        if (!rootNode.idMatches(relativeName, nameSpace))
        {
            TagNodeId id = new TagNodeId(relativeName, nameSpace);
            close();
            throw new WrongNodeException(_context,
                    "Got " + rootNode._id.comparisonName() + " at " + rootNode._textPos +
                    ", expected " + id.comparisonName());
        }
        return rootNode;
    }
    
    public TagNodeImpl
    getRootNode(TagNodeId id)
    {
        return getRootNode(id._relativeName, id._nameSpace);
    }
    
    /**
     * Gets root node without any preconception of what it should contain.
     * @return The root node.
     */
    public TagNodeImpl
    getRootNode()
    {
        try
        {
            if (_gotRootNode)
            {
                throw new IncorrectInvocationReadException(_context, "Attempt to get root node a second time");
            }
            TextNode preRootTextNode = readText();
            if (!isWhiteSpace(preRootTextNode._text))
            {
                throw makeSyntaxException(
                        "There is text preceding the root node at " + preRootTextNode._textPos);
            }
            return getTagNode();
        }
        catch (RuntimeException e)
        {
            close();
            throw e;
        }
    }
    
    /**
     * Closes the internal reader. Invoking any method on this class which requires more reading
     * is an error condition. Closing multiple times is permitted.
     */
    public void
    close()
    {
        try
        {
            _reader.close();
        }
        catch (IOException e)
        {
            throw new ReadIoException(_context, e.getMessage(), e);
        }
    }
    
    public boolean
    isOpen()
    {
        return _reader != null && _reader.isOpen();
    }
    
    /**
     * Gets the next tag node.
     * @return Next child tag node if we are already inside a node, otherwise root node. If there
     *         is no suitable tag node, we return null (presuming syntax is valid).
     */
    TagNodeImpl
    getTagNode()
    {
        try
        {
            // At this point we are either at end or in front of "<"
            if (have2CharsAfterText())
            {
                if (!_reader.startsWith("<"))
                {
                    throw new InternalBugException(_context, "Internal error");
                }
                if (_reader.startsWith("</", true))
                {
                    readEndTag();
                    return null;
                }
                else
                {
                    return readOpenTag();
                }
            }
            else
            {
                return null;
            }
        }
        catch (IOException e)
        {
            e.printStackTrace();
            throw makeIoException(e);
        }
    }
    
    private static final char[] textTerminators = new char[]{'&', '<'};
    
    /*
     * Postcondition: Read position will be either at end of input or at "<" with all possibilities
     * other than a start or end tag already scanned past.
     */
    TextNode
    readText()
    {
        try
        {
            StringBuilder textBuf = new StringBuilder();
            boolean finishedPreNode = false;
            ROTextPosition textStart = null;
            while (!finishedPreNode)
            {
                ROTextPosition textPossibleStart = null;
                if (textStart == null)
                {
                    textPossibleStart = textPosition();
                }
                _reader.consumeUntil(textTerminators, textBuf, false);
                if (textStart == null && textBuf.length() > 0)
                {
                    textStart = textPossibleStart;
                }
                if (_reader.consumedAll())
                {
                    finishedPreNode = true;
                }
                else if (_reader.startsWith("&"))
                {
                    if (textStart == null)
                    {
                        textStart = textPosition();
                    }
                    char ch = readUniCode();
                    textBuf.append(ch);
                }
                else if (_reader.startsWith("<?", true))
                {
                    ROTextPosition start = textPosition();
                    if (!_reader.consumeUntil("?>", null, true))
                    {
                        throw makeSyntaxException("Unterminated <? at " + start);
                    }
                }
                else if (_reader.startsWith("<!--", true))
                {
                    ROTextPosition start = textPosition();
                    if (!_reader.consumeUntil("-->", null, true))
                    {
                        throw makeSyntaxException("Unterminated comment at " + start);
                    }
                }
                else if (_reader.startsWith("<!DOCTYPE", true))
                {
                    ROTextPosition start = textPosition();
                    if (!_reader.consumeUntil(">", null, true))
                    {
                        throw makeSyntaxException("Unterminated document type declaration at " + start);
                    }
                }
                else if (_reader.startsWith(CSECTION_START, true))
                {
                    if (textStart == null)
                    {
                        textStart = textPosition();
                    }
                    ROTextPosition start = textPosition();
                    if (!_reader.consumeUntil(CSECTION_END, textBuf, true))
                    {
                        throw makeSyntaxException("Unterminated C Section at " + start);
                    }
                }
                else if (_reader.startsWith("<", false))
                {
                    if (textStart == null)
                    {
                        // text will be empty but we still want position
                        textStart = textPosition();
                    }
                    finishedPreNode = true;
                }
                else
                {
                    throw new InternalBugException(_context, "Internal error");
                }
            }
            String text = textBuf.toString();
            if (!_readOptions._valueCharSet.isIn(text))
            {
                int outCharPos = _readOptions._valueCharSet.firstOutCharPosition(text);
                char outCh = text.charAt(outCharPos);
                throw new XmlCharSetException(
                        _context,
                        "text at " + textStart +
                        " has character " + CharUtil.unicodeCharToStringForMessage(outCh) +
                        " which is not in value char set " + _readOptions._valueCharSet +
                        " which is specified in read options, position within text is " + (outCharPos + 1));
            }
            if (textStart == null)
            {
                textStart = textPosition();
            }
            return new TextNode(textStart, text);
        }
        catch (IOException e)
        {
            throw makeIoException(e);
        }
    }
    
    // We read all the text we needed to read, now we expect open or close tag or nothing at all.
    private boolean
    have2CharsAfterText() throws IOException
    {
        if (!_reader.haveNChars(2))
        {
            String message;
            if (!_gotRootNode)
            {
                message = "No root node";
            }
            else if (_tagStack.size() > 0)
            {
                StringBuilder buf = null;
                for (TagNodeImpl tagNode : _tagStack)
                {
                    if (buf == null)
                    {
                        buf = new StringBuilder(
                                "The following nodes are still open when end is encountered: ");
                        buf.append(tagNode);
                    }
                    else
                    {
                        buf.append(", ");
                        buf.append(tagNode);
                    }
                }
                message = buf.toString();
            }
            else
            {
                return false;
            }
            throw makeSyntaxException(message);
        }
        return true;
    }
    
    // Precondition: we already saw the </
    private void
    readEndTag() throws IOException
    {
        ROTextPosition endTagPos = textPosition();
        if (_tagStack.size() == 0)
        {
            throw makeSyntaxException(
                    "No node is open when close tag is encountered at " + endTagPos);
        }
        TagNodeImpl openNode = _tagStack.get(_tagStack.size() - 1);
        StringBuilder endTagBuf = new StringBuilder();
        if (!_reader.consumeUntil(">", endTagBuf, true))
        {
            throw makeSyntaxException(
                    "Close tag without > at " + endTagPos +
                    ", corresponding open node is " + openNode);
        }
        String endTagNameUntrimmed = endTagBuf.toString();
        String endTagName = endTagNameUntrimmed.trim();
        String openNodeFullName = openNode._id.displayName();
        if (!endTagName.equals(openNodeFullName))
        {
            throw makeSyntaxException(
                    "Close tag " + endTagName + " at " + endTagPos +
                    " fails to match open tag " + openNode);
        }
        _tagStack.remove(_tagStack.size() - 1);
        unstackNameSpaces();
        if (openNode._parent == null)
        {
            processPostRoot();
        }
    }
    
    private void
    unstackNameSpaces()
    {
        // We are working on the assumption that the tag node for which we are
        // unstacking name spaces will not be on the tag stack.
        // That way, we don't have to stack tag nodes without separate end tag.
        int unstackLevel = _tagStack.size();
        
        boolean keepGoing = true;
        while (keepGoing)
        {
            keepGoing = false;
            int lastNsIndex = _nameSpaceStack.size() - 1;
            if (lastNsIndex >= 0)
            {
                XmlNameSpaceAtLevel ns = _nameSpaceStack.get(lastNsIndex);
                if (ns._level >= unstackLevel)
                {
                    _nameSpaceStack.remove(lastNsIndex);
                    keepGoing = true;
                }
            }
        }
    }
    
    // Precondition: We saw the "<", we know it's not "<!".
    // We now expect an open tag because all other possibilities have already
    // been covered while reading the preceding text.
    private TagNodeImpl
    readOpenTag() throws IOException
    {
        ROTextPosition openTagPos = textPosition();
        _reader.advance();
        String tagName = readTagName(openTagPos);
        ArrayList<TmpAttribute> tmpAttributeAl = new ArrayList<TmpAttribute>();
        TmpAttribute tmpAttribute = null;
        do
        {
            tmpAttribute = readAttribute(openTagPos);
            if (tmpAttribute != null)
            {
                tmpAttributeAl.add(tmpAttribute);
            }
        } while (tmpAttribute != null);
        verifyNoPrefixClash();
        XmlAttribute[] attributes = new XmlAttribute[tmpAttributeAl.size()];
        for (int i = 0; i < attributes.length; i++)
        {
            tmpAttribute = tmpAttributeAl.get(i);
            String relativeName;
            XmlNameSpace nameSpace;
            int colonIndex = tmpAttribute._name.indexOf(':');
            if (colonIndex == -1)
            {
                nameSpace = null;
                relativeName = tmpAttribute._name;
            }
            else
            {
                String prefix = tmpAttribute._name.substring(0, colonIndex);
                relativeName = tmpAttribute._name.substring(colonIndex + 1);
                nameSpace = prefixToNameSpace(prefix);
                if (nameSpace == null)
                {
                    throw makeSyntaxException(
                            "Can not map prefix " + prefix + " of attribute name " + relativeName +
                            " for opening tag at " + openTagPos + " to a name space");
                }
                else if (nameSpace._name.equals(""))
                {
                    throw makeSyntaxException(
                            "Prefix " + prefix + " of attribute name " + relativeName +
                            " for opening tag at " + openTagPos + " has been undeclared");
                }
            }
            if (!_readOptions._nameCharSet.isIn(relativeName))
            {
                int outCharIndex = _readOptions._nameCharSet.firstOutCharPosition(relativeName);
                char outCh = relativeName.charAt(outCharIndex);
                throw new XmlCharSetException(
                        _context,
                        "Attribute name " + relativeName + " of node at " + openTagPos +
                        " has character " + CharUtil.unicodeCharToStringForMessage(outCh) +
                        " in position " + (outCharIndex + 1) +
                        " which is not in name char set " + _readOptions._nameCharSet +
                        " which is specified in read options");
            }
            attributes[i] = new XmlAttribute(relativeName, nameSpace, tmpAttribute._value);
        }
        boolean nodeFinished;
        char ch = _reader.read();
        if (ch == '/')
        {
            checkOpeningTagContinues(openTagPos);
            ch = _reader.read();
            if (ch != '>')
            {
                throw makeSyntaxException(
                    "Failed to find > at " + _textPos + " to end opening tag at " + openTagPos);
            }
            nodeFinished = true;
        }
        else if (ch == '>')
        {
            nodeFinished = false;
        }
        else
        {
            throw makeSyntaxException(
                "Failed to find opening tag end near " + _textPos.textPosition() +
                " to end open tag at " + openTagPos);
        }
        TagNodeImpl parentNode = null;
        if (_tagStack.size() > 0)
        {
            parentNode = _tagStack.get(_tagStack.size() - 1);
        }
        TagNodeId tagNodeId = tagNameToId(tagName, openTagPos);
        if (!_readOptions._nameCharSet.isIn(tagNodeId._relativeName))
        {
            int outCharIndex = _readOptions._nameCharSet.firstOutCharPosition(tagNodeId._relativeName);
            char outCh = tagNodeId._relativeName.charAt(outCharIndex);
            throw new XmlCharSetException(
                    _context,
                    "Name " + tagNodeId._relativeName + " of node at " + openTagPos +
                    " has character " + CharUtil.unicodeCharToStringForMessage(outCh) +
                    " in position " + (outCharIndex + 1) +
                    " which is not in name char set " + _readOptions._nameCharSet +
                    " which is specified in read options");
        }
        if (tagNodeId._nameSpace != null && tagNodeId._nameSpace._name.equals(""))
        {
            throw makeSyntaxException(
                    "Name " + tagNodeId._relativeName + " of node at " + openTagPos +
                    " has prefix " + tagNodeId._nameSpace._prefix + " which has been undeclared");
        }
        TagNodeImpl tagNode =
            new TagNodeImpl(openTagPos, tagNodeId, attributes, parentNode, nodeFinished, this);
        if (nodeFinished)
        {
            unstackNameSpaces();
        }
        else
        {
            _tagStack.add(tagNode);
        }
        _gotRootNode = true;
        return tagNode;
    }
    
    ROTextPosition
    textPosition()
    {
        return _reader._textPosView.textPosition();
    }
    
    private String
    readTagName(ROTextPosition openTagPos) throws IOException
    {
        char ch = _reader.peep();
        if (!isFirstTagChar(ch))
        {
            String chS = CharUtil.unicodeCharToStringForMessage(ch);
            throw makeSyntaxException(
                    "Illegal tag name after < at " + openTagPos +
                    ", following character " + chS + " is not a legal XML name starter");
        }
        _reader.advance();
        StringBuilder tagNameBuf = new StringBuilder();
        tagNameBuf.append(ch);
        boolean inTagName = true;
        int colonQty = 0;
        while (inTagName)
        {
            checkOpeningTagContinues(openTagPos);
            ch = _reader.peep();
            if (isTagChar(ch))
            {
                tagNameBuf.append(ch);
                _reader.advance();
                if (ch == ':')
                {
                    colonQty++;
                }
            }
            else
            {
                inTagName = false;
            }
        }
        String tagName = tagNameBuf.toString();
        if (colonQty > 1)
        {
            throw makeSyntaxException(
                    "Multiple colons for in node tag name " + tagName + " at " + openTagPos);
        }
        if (colonQty == 1)
        {
            int colonIndex = tagName.indexOf(':');
            int relNameIndex = colonIndex + 1;
            if (relNameIndex < tagName.length())
            {
                char firstRelNameChar = tagName.charAt(relNameIndex);
                if (!XmlChars.isNameStartChar(firstRelNameChar))
                {
                    throw makeSyntaxException(
                            "Relative name starts with unacceptable first character " +
                            CharUtil.unicodeCharToStringForMessage(firstRelNameChar) +
                            " in node tag name " + tagName + " at " + openTagPos);
                }
            }
        }
        if (!Character.isWhitespace(ch) && ch != '/' && ch != '>')
        {
            throw makeSyntaxException(
                    "Tag name " + tagName + " for node at " + openTagPos +
                    " is followed by unacceptable character " + CharUtil.unicodeCharToStringForMessage(ch));
        }
        return tagName;
    }
    
    private TagNodeId
    tagNameToId(String         tagName,
                ROTextPosition openTagPos)
    {
        String prefix;
        String relativeName;
        int colonIndex = tagName.indexOf(':');
        if (colonIndex == -1)
        {
            prefix = null;
            relativeName = tagName;
        }
        else
        {
            prefix = tagName.substring(0, colonIndex);
            relativeName = tagName.substring(colonIndex + 1);
        }
        XmlNameSpace nameSpace = prefixToNameSpace(prefix);
        if (nameSpace == null && prefix != null)
        {
            throw makeSyntaxException(
                    "Can not map prefix " + prefix + " of name " + tagName +
                    " for opening tag at " + openTagPos + " to a name space");
        }
        return new TagNodeId(relativeName, nameSpace);
    }
    
    private XmlNameSpace
    prefixToNameSpace(String prefix)
    {
        for (int nameSpaceIndex = _nameSpaceStack.size() - 1; nameSpaceIndex >= 0; nameSpaceIndex--)
        {
            XmlNameSpace nameSpace = _nameSpaceStack.get(nameSpaceIndex);
            if (prefix == null)
            {
                if (nameSpace._prefix == null)
                {
                    return nameSpace;
                }
            }
            else if (nameSpace._prefix.equals(prefix))
            {
                return nameSpace;
            }
        }
        return null;
    }
    
    private String
    readAttributeName(ROTextPosition openTagPos) throws IOException
    {
        char ch = _reader.peep();
        if (!isFirstAttributeChar(ch))
        {
            return null;
        }
        _reader.advance();
        StringBuilder tagNameBuf = new StringBuilder();
        tagNameBuf.append(ch);
        boolean inTagName = true;
        int colonQty = 0;
        while (inTagName)
        {
            checkOpeningTagContinues(openTagPos);
            ch = _reader.peep();
            if (isAttributeChar(ch))
            {
                tagNameBuf.append(ch);
                _reader.advance();
                if (ch == ':')
                {
                    colonQty++;
                }
            }
            else
            {
                inTagName = false;
            }
        }
        String tagName = tagNameBuf.toString();
        if (colonQty > 1)
        {
            throw makeSyntaxException(
                    "Multiple colons for attribute name " + tagName + " belonging to node starting at " + openTagPos);
        }
        if (colonQty == 1)
        {
            int colonIndex = tagName.indexOf(':');
            int relNameIndex = colonIndex + 1;
            if (relNameIndex < tagName.length())
            {
                char firstRelNameChar = tagName.charAt(relNameIndex);
                if (!XmlChars.isNameStartChar(firstRelNameChar))
                {
                    throw makeSyntaxException(
                            "Relative name starts with unacceptable first character " +
                            CharUtil.unicodeCharToStringForMessage(firstRelNameChar) +
                            " in attribute name " + tagName + " at " + openTagPos);
                }
            }
        }
        return tagName;
    }
    
    private String
    readAttributeValue(ROTextPosition openTagPos) throws IOException
    {
        char quoteChar = _reader.peep();
        if (quoteChar != '\'' && quoteChar != '"')
        {
            throw makeSyntaxException(
                  "Unquoted attribute value at " + _textPos + " for opening tag at " + openTagPos);
        }
        _reader.advance();
        StringBuilder buf = new StringBuilder();
        boolean finished = false;
        while (!finished)
        {
            checkOpeningTagContinues(openTagPos);
            char ch = _reader.peep();
            if (ch == '&')
            {
                char chu = readUniCode();
                buf.append(chu);
            }
            else if (ch == '<')
            {
                throw makeSyntaxException(
                    "Unexpected < at " + _textPos + " in attribute value for opening tag at " +
                    openTagPos);
            }
            else if (ch == quoteChar)
            {
                _reader.advance();
                finished = true;
            }
            else
            {
                buf.append(ch);
                _reader.advance();
            }
        }
        return buf.toString();
    }
    
    private TmpAttribute
    readAttribute(ROTextPosition openTagPos) throws IOException
    {
        _reader.advancePastWhitespace();
        checkOpeningTagContinues(openTagPos);
        String attributeName = readAttributeName(openTagPos);
        if (attributeName == null)
        {
            return null;
        }
        _reader.advancePastWhitespace();
        checkOpeningTagContinues(openTagPos);
        char ch = _reader.peep();
        if (ch != '=')
        {
            throw makeSyntaxException(
                    "Failed to find = at " + _textPos + " for opening tag at " + openTagPos +
                    ", attribute " + attributeName + ", instead found unacceptable character " +
                    CharUtil.unicodeCharToStringForMessage(ch));
        }
        _reader.advance();
        _reader.advancePastWhitespace();
        checkOpeningTagContinues(openTagPos);
        String attributeValue = readAttributeValue(openTagPos);
        if (!_readOptions._valueCharSet.isIn(attributeValue))
        {
            int outCharPos = _readOptions._valueCharSet.firstOutCharPosition(attributeValue);
            char outCh = attributeValue.charAt(outCharPos);
            throw new XmlCharSetException(
                    _context,
                    "Attribute " + attributeName + " for opening tag at " + openTagPos +
                    " has character inconsistent with read options: " +
                    CharUtil.unicodeCharToStringForMessage(outCh) +
                    ", position within value is " + (outCharPos + 1) +
                    ", charSet is " + _readOptions._valueCharSet);
        }
        TmpAttribute attribute = new TmpAttribute(attributeName, attributeValue);
        considerNameSpace(attribute);
        return attribute;
    }
    
    private void
    verifyNoPrefixClash()
    {
        int nameSpaceQty = _nameSpaceStack.size();
        if (nameSpaceQty < 2)
        {
            return;
        }
        for (int cmpLeftIndex = nameSpaceQty - 2; cmpLeftIndex >= 0; cmpLeftIndex--)
        {
            XmlNameSpaceAtLevel cmpLeft = _nameSpaceStack.get(cmpLeftIndex);
            if (cmpLeft._level < _tagStack.size())
            // The tag node for the namespaces currently being checked is not yet on the stack.
            {
                // We are down to the previous level without having found a problem.
                return;
            }
            for (int cmpRightIndex = cmpLeftIndex + 1; cmpRightIndex < nameSpaceQty; cmpRightIndex++)
            {
                XmlNameSpaceAtLevel cmpRight = _nameSpaceStack.get(cmpRightIndex);
                if (cmpLeft._prefix.equals(cmpRight._prefix) && !cmpLeft._name.equals(cmpRight._name))
                {
                    throw makeSyntaxException(
                            "name spaces " + cmpLeft._name + " and " + cmpRight._name +
                            " are both mapped to same prefix " + cmpLeft._prefix);
                }
            }
        }
    }
    
    private void
    processPostRoot()
    {
        try
        {
            while (_reader.haveChar())
            {
                char ch = _reader.read();
                if (!Character.isWhitespace(ch))
                {
                    throw makeSyntaxException("Non whitespace after root near " + textPosition());
                }
            }
        }
        catch (IOException e)
        {
            
        }
    }
    
    private void
    considerNameSpace(TmpAttribute   attribute)
    {
        String prefix;
        if (attribute._name.equals(XmlNameSpace.XMLNS._prefix))
        {
            prefix = null;
        }
        else if (attribute._name.startsWith(NAMESPACE_DECL_PREFIX))
        {
            prefix = attribute._name.substring(NAMESPACE_DECL_PREFIX.length());
        }
        else
        {
            return;
        }
        
        // We are working on the assumption that the TagNode has not yet been created.
        // That's because we want to have the name space available when we call the
        // TagNodeImpl constructor.
        XmlNameSpaceAtLevel ns = new XmlNameSpaceAtLevel(prefix, attribute._value, _tagStack.size());
        _nameSpaceStack.add(ns);
    }
    
    private char
    readUniCode() throws IOException
    {
        try
        {
            if (_reader.startsWith("&amp;", true))
            {
                return '&';
            }
            else if (_reader.startsWith("&gt;", true))
            {
                return '>';
            }
            else if (_reader.startsWith("&lt;", true))
            {
                return '<';
            }
            else if (_reader.startsWith("&quot;", true))
            {
                return '"';
            }
            else if (_reader.startsWith("&apos;", true))
            {
                return '\'';
            }
            else if (_reader.startsWith("&#", true))
            {
                int total = 0;
                if (   _reader.startsWith("x", true)
                    || _reader.startsWith("X", true))
                {
                    int digitQty = 0;
                    while (   !_reader.startsWith(";")
                           && !_reader.consumedAll())
                    {
                        total *= 16;
                        digitQty++;
                        char ch = _reader.peep();
                        int digit;
                        if (ch >= '0' && ch <= '9')
                        {
                            digit = ch - '0';
                        }
                        else if (ch >= 'a' && ch <= 'f')
                        {
                            digit = ch - 'a' + 10;
                        }
                        else if (ch >= 'A' && ch <= 'F')
                        {
                            digit = ch - 'A' + 10;
                        }
                        else
                        {
                            throw makeSyntaxException(
                                    "Character does not match hexadecimal unicode digit at " + _textPos);
                        }
                        _reader.advance();
                        total += digit;
                    }
                    if (digitQty > 4)
                    {
                        throw makeSyntaxException(
                                "To many digits for hexadecimal unicode near " + _textPos);
                    }
                }
                else
                {
                    while (   !_reader.startsWith(";")
                           && !_reader.consumedAll())
                    {
                        total *= 10;
                        char ch = _reader.peep();
                        int digit;
                        if (ch >= '0' && ch <= '9')
                        {
                            digit = ch - '0';
                            _reader.advance();
                        }
                        else
                        {
                            throw makeSyntaxException(
                                    "Character does not match decimal unicode digit at " + _textPos);
                        }
                        total += digit;
                        if (total > Short.MAX_VALUE)
                        {
                            throw makeSyntaxException(
                                    "Too large decimal unicode near " + _textPos);
                        }
                    }
                }
                if (_reader.consumedAll())
                {
                    throw makeSyntaxException("Finished before reaching ; in unicode at " + _textPos);
                }
                _reader.advance();
                return (char)total;
            }
            else
            {
                throw makeSyntaxException("Bad Unicode at " + _textPos);
            }
        }
        catch (IOException e)
        {
            throw makeIoException(e);
        }
    }
    
    private boolean
    isFirstAttributeChar(char ch)
    {
        return isFirstTagChar(ch);
    }
    
    private boolean
    isAttributeChar(char ch)
    {
        return isTagChar(ch);
    }
    
    private boolean
    isFirstTagChar(char ch)
    {
        return XmlChars.isNameStartChar(ch);
    }
    
    private boolean
    isTagChar(char ch)
    {
        return ch == ':' || XmlChars.isNameChar(ch);
    }
    
    public static boolean
    isWhiteSpace(String s)
    {
        int sLength = s.length();
        for (int i = 0; i < sLength; i++)
        {
            char ch = s.charAt(i);
            if (!Character.isWhitespace(ch))
            {
                return false;
            }
        }
        return true;
    }
    
    private void
    checkOpeningTagContinues(ROTextPosition openTagPos) throws IOException
    {
        if (_reader.consumedAll())
        {
            throw makeSyntaxException("Incomplete opening tag at " + openTagPos);
        }
    }
    
    ReadIoException
    makeIoException(IOException e)
    {
        return new ReadIoException(_context, e.toString() + " at " + _textPos, e);
    }
    
    XmlSyntaxException
    makeSyntaxException(String message)
    {
        try
        {
            _reader.close();
        }
        catch (IOException e)
        {
            // There is not much we can do. We certainly don't want to replace the original exception.
        }
        return new XmlSyntaxException(_context, message);
    }
    
    public static TagNode
    readerToRoot(Reader         reader,
                 String         context,
                 XmlReadOptions readOptions)
    {
        XmlReader xReader = new XmlReader(reader, context, readOptions);
        TagNode tagNode = xReader.getRootNode();
        return tagNode;
    }
    
    public static TagNode
    readerToRoot(Reader         reader,
                 String         context,
                 XmlReadOptions readOptions,
                 TagNodeId      expectedId)
    {
        XmlReader xReader = new XmlReader(reader, context, readOptions);
        TagNode tagNode = xReader.getRootNode(expectedId);
        return tagNode;
    }
    
    public static TagNode
    readerToRoot(Reader       reader,
                 String       context,
                 XmlReadOptions  readOptions,
                 String       expectedRelativeName,
                 XmlNameSpace expectedNameSpace)
    {
        XmlReader xReader = new XmlReader(reader, context, readOptions);
        TagNode tagNode = xReader.getRootNode(expectedRelativeName, expectedNameSpace);
        return tagNode;
    }
    
    public static TagNode
    xmlStringToRoot(String      xml,
                    String      context,
                    XmlReadOptions readOptions)
    {
        StringReader sReader = new StringReader(xml);
        return readerToRoot(sReader, context, readOptions);
    }
    
    public static TagNode
    xmlStringToRoot(String      xml,
                    String      context,
                    XmlReadOptions readOptions,
                    TagNodeId   expectedId)
    {
        StringReader sReader = new StringReader(xml);
        return readerToRoot(sReader, context, readOptions, expectedId);
    }
    
    public static TagNode
    xmlStringToRoot(String       xml,
                    String       context,
                    XmlReadOptions  readOptions,
                    String       expectedRelativeName,
                    XmlNameSpace expectedNameSpace)
    {
        StringReader sReader = new StringReader(xml);
        return readerToRoot(
                sReader, context, readOptions, expectedRelativeName, expectedNameSpace);
    }
    
    public static TagNode
    xmlStringToRoot(String         xml,
                    String         context,
                    XmlReadOptions readOptions,
                    String         expectedRelativeName)
    {
        return xmlStringToRoot(xml, context, readOptions, expectedRelativeName, null);
    }
    
    public static TagNode
    xmlFileToRoot(File        xmlFile,
                  String      xmlContext,
                  XmlReadOptions readOptions)
    {
        FileReader fReader = fileToReader(xmlFile, xmlContext);
        return readerToRoot(fReader, xmlContext, readOptions);
    }
    
    public static TagNode
    xmlFileToRoot(File        xmlFile,
                  String      xmlContext,
                  XmlReadOptions readOptions,
                  TagNodeId   expectedId)
    {
        FileReader fReader = fileToReader(xmlFile, xmlContext);
        return readerToRoot(fReader, xmlContext, readOptions, expectedId);
    }
    
    public static TagNode
    xmlFileToRoot(File         xmlFile,
                  String      xmlContext,
                  XmlReadOptions  readOptions,
                  String       expectedRelativeName,
                  XmlNameSpace expectedNameSpace)
    {
        FileReader fReader = fileToReader(xmlFile, xmlContext);
        return readerToRoot(
                fReader, xmlContext, readOptions, expectedRelativeName, expectedNameSpace);
    }
    
    private static FileReader
    fileToReader(File   file,
                 String xmlContext)
    {
        try
        {
            return new FileReader(file);
        }
        catch (FileNotFoundException e)
        {
            throw new ReadIoException(xmlContext, e.toString(), e);
        }
    }
    
    public static TagNode
    xmlFileToRoot(File        xmlFile,
                  XmlReadOptions readOptions)
    {
        String xmlContext = xmlFile.getName();
        return xmlFileToRoot(xmlFile, xmlContext, readOptions);
    }
    
    public static TagNode
    xmlFileToRoot(File        xmlFile,
                  XmlReadOptions readOptions,
                  TagNodeId   expectedId)
    {
        String xmlContext = xmlFile.getName();
        return xmlFileToRoot(xmlFile, xmlContext, readOptions, expectedId);
    }
    
    public static TagNode
    xmlFileToRoot(File         xmlFile,
                  XmlReadOptions  readOptions,
                  String       expectedRelativeName,
                  XmlNameSpace expectedNameSpace)
    {
        String xmlContext = xmlFile.getName();
        return xmlFileToRoot(
                xmlFile, xmlContext, readOptions, expectedRelativeName, expectedNameSpace);
    }
    
    public static TagNode
    xmlFileToRoot(File            xmlFile,
                  XmlReadOptions  readOptions,
                  String          expectedRootName)
    {
        return xmlFileToRoot(xmlFile, readOptions, expectedRootName, null);
    }
}
