/*
 * Copyright Rasmus Torkel 2012
 * Author    Rasmus Torkel 2012
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.xml_basic.rw;

import java.util.List;

import rasmus_torkel.misc.Constants;

public class XmlAttribute implements Comparable<XmlAttribute>
{
    public final String       _name;
    public final XmlNameSpace _nameSpace;
    public final String       _value;
    
    /**
     * If this attribute defines a name space, this field is the name space.
     * Otherwise this field is null.
     */
    public final XmlNameSpace _definedNameSpace;
    
    public
    XmlAttribute(String name,
                 String value)
    {
        this(name, null, value);
    }
    
    public
    XmlAttribute(String       name,
                 XmlNameSpace nameSpace,
                 String       value)
    {
        _name = name;
        if (_name.contains(":"))
        {
            throw new RuntimeException("name " + name + " contains :");
        }
        if (_name.contains(Constants.NULL_STRING))
        {
            throw new RuntimeException("name" + name + " contains null char");
        }
        _nameSpace = nameSpace;
        _value = value;
        if (_nameSpace != null && _nameSpace._name.equals(XmlNameSpace.XMLNS._name))
        {
            _definedNameSpace = new XmlNameSpace(name, value);
        }
        else
        {
            _definedNameSpace = null;
        }
    }
    
    /**
     * This implementation performs NO encoding and is basically just for test and debugging
     * purposes.
     * That's because when we do serious XML writing, we need encoding options
     * which are not available here.
     */
    @Override
    public String
    toString()
    {
        int initialSize = _name.length() + _value.length() + _value.length() / 4 + 16;
        StringBuilder buf = new StringBuilder(initialSize);
        if (_nameSpace != null)
        {
            buf.append(_nameSpace._name);
            buf.append(':');
        }
        buf.append(_name);
        buf.append('=');
        buf.append(_value);
        return buf.toString();
    }

    @Override
    public int
    compareTo(XmlAttribute a2)
    {
        if (_definedNameSpace != null && a2._definedNameSpace == null)
        {
            return -1;
        }
        if (a2._definedNameSpace != null && _definedNameSpace == null)
        {
            return 1;
        }
        int nameCmp = _name.compareTo(a2._name);
        if (nameCmp != 0)
        {
            return nameCmp;
        }
        
        // The following should not occur, but that does not mean that we are prepared
        // to sort badly if it does occur.
        return _value.compareTo(a2._value);
    }
    
    @Override
    public boolean
    equals(Object other)
    {
        if (other == null)
        {
            return false;
        }
        if (!(other instanceof XmlAttribute))
        {
            return false;
        }
        XmlAttribute otherAttribute = (XmlAttribute)other;
        return _name.equals(otherAttribute._name) && _value.equals(otherAttribute._value);
    }
    
    public boolean
    idEquals(XmlAttribute otherAttr)
    {
        if (!_name.equals(otherAttr._name))
        {
            return false;
        }
        else if (_nameSpace == null && otherAttr._nameSpace == null)
        {
            return true;
        }
        else if (_nameSpace == null || otherAttr._nameSpace == null)
        {
            return false;
        }
        else
        {
            return _nameSpace._name.equals(otherAttr._nameSpace._name);
        }
    }
    
    public static XmlAttribute[]
    listToAttribute(List<XmlAttribute> attributeList)
    {
        XmlAttribute[] attributes = new XmlAttribute[attributeList.size()];
        attributeList.toArray(attributes);
        return attributes;
    }
}
