/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2012/2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.xml_basic.rw;

import rasmus_torkel.misc.StringUtil;

/**
 * Represents basic XML name space information
 * 
 * @author Rasmus Torkel
 */
public class XmlNameSpace
{
    /**
     * The prefix for node names in the name space.
     * <p>
     * When reading, this is the prefix actually encountered.
     * When writing, this is the preferred prefix to be used. The actual prefix
     * may be different if there is another namespace with the same prefix.
     * Therefore, the prefix is not used for testing whether two namespaces are
     * the same.
     */
    public final String _prefix;
    
    public final String _name;
    
    /**
     * For situations while reading where there may be a name space or there may not be and if
     * there is it does not matter what it is.
     */
    public static final XmlNameSpace ANY =
        new XmlNameSpace(
                "any",
                "Not_a_real_namespace__pass_into_read_methods_to_indicate_namespace_doesnt_matter");
    
    public static final XmlNameSpace XMLNS =
        new XmlNameSpace("xmlns", "http://www.w3.org/2000/xmlns/");
    
    public static final XmlNameSpace XML =
        new XmlNameSpace("xml", "http://www.w3.org/XML/1998/namespace");
    
    public
    XmlNameSpace(String prefix,
                 String name)
    {
        _prefix = prefix;
        _name = name;
        if (   _prefix != null
            && !XmlChars.NAME_SET.isIn(_prefix))
        {
            String prefixPhrase = StringUtil.namedValueToPlainPrintablePhrase("name space prefix", _prefix);
            String outReason = XmlChars.NAME_SET.outReason(prefix);
            throw new RuntimeException(prefixPhrase + " " + outReason);
        }
    }
    
    @Override
    public String
    toString()
    {
        return _name;
    }
    
    @Override
    public boolean
    equals(Object other)
    {
        if (other == null)
        {
            return false;
        }
        if (!(other instanceof XmlNameSpace))
        {
            return false;
        }
        XmlNameSpace otherNameSpace = (XmlNameSpace)other;
        return otherNameSpace._name.equals(_name);
    }
    
    @Override
    public int
    hashCode()
    {
        return _name.hashCode();
    }
    
    public static boolean
    isSame(XmlNameSpace ns1,
           XmlNameSpace ns2)
    {
        if (ns1 == null && ns2 == null)
        {
            return true;
        }
        if (ns1 == null || ns2 == null)
        {
            return false;
        }
        if (ns1._name.equals(ns2._name))
        {
            return true;
        }
        return false;
    }
    
    public String
    makeDeclaration()
    {
        return "xmlns:" + _prefix + "=\"" + _name + "\"";
    }
}
