/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.xml_basic.tool;

import java.io.File;
import java.io.PrintStream;

import rasmus_torkel.xml_basic.read.TagNode;
import rasmus_torkel.xml_basic.read.TextNode;
import rasmus_torkel.xml_basic.read.XmlReadOptions;
import rasmus_torkel.xml_basic.read.impl.XmlReader;
import rasmus_torkel.xml_basic.rw.XmlAttribute;

/**
 * Class for comparing two XMLs. It can be used both as a library and as a stand-alone tool.
 * <p>
 * All the areEquivalent methods have a few points in common:
 * <ul>
 *   <li>At least for this version we have no options for setting equivalence criteria.
 *       Such options may be added later if the need arises.</li>
 *   <li>Output will go to PrintStream. Only the println(String) method is ever used, which should
 *       help if anybody wants to supply an implementation which does more then just print the
 *       messages.</li>
 * </ul>
 * Below are the criteria which are used to determine equivalence.
 * <ul>
 *   <li>Attributes may be in any order but must be otherwise identical.</li>
 *   <li>All the child (tag) nodes must be equivalent.</li>
 *   <li>Text nodes are only equivalent if they are identical except that empty or whitespace-only
 *       nodes are equivalent even if the whitespace differs.</li>
 *   <li>A tag node consisting of a single self-closing tag is not equivalent to a tag node
 *       consisting of a start tag and a matching end tag and nothing in between.</li>
 *   <li>Tag nodes with an identical set of child nodes but in different order are not
 *       equivalent.</li> 
 * </ul>
 * 
 * @author Torkel
 *
 */
public class XmlComparer
{
    private static final XmlReadOptions READ_OPTIONS =
        new XmlReadOptions(false, true, XmlReadOptions.MOST_LENIENT_CHAR_SET);
    
    /**
     * Main function for use as a stand-alone tool.
     * @param args The two names of the files to be compared.
     */
    public static void
    main(String[] args)
    {
        if (args.length != 2)
        {
            System.err.println(
                    "Need exactly two arguments which are the names of the files to be compared for equivalence");
            System.exit(1);
        }
        String fileName1 = args[0];
        String fileName2 = args[1];
        
        File file1 = new File(fileName1);
        File file2 = new File(fileName2);
        areEquivalent(file1, file2, System.out);
    }
    
    public static boolean
    areEquivalent(File        xmlFile1,
                  File        xmlFile2,
                  PrintStream diffStream)
    {
        String context1 = xmlFile1.getName();
        String context2 = xmlFile2.getName();
        return areEquivalent(xmlFile1, context1, xmlFile2, context2, diffStream);
    }
    
    public static boolean
    areEquivalent(File        xmlFile1,
                  String      context1,
                  File        xmlFile2,
                  String      context2,
                  PrintStream diffStream)
    {
        TagNode node1 = XmlReader.xmlFileToRoot(xmlFile1, context1, READ_OPTIONS);
        TagNode node2 = XmlReader.xmlFileToRoot(xmlFile2, context2, READ_OPTIONS);
        return areEquivalent(node1, node2, diffStream);
    }
    
    public static boolean
    areEquivalent(String      xml1,
                  String      context1,
                  String      xml2,
                  String      context2,
                  PrintStream diffStream)
    {
        TagNode node1 = XmlReader.xmlStringToRoot(xml1, context1, READ_OPTIONS);
        TagNode node2 = XmlReader.xmlStringToRoot(xml2, context2, READ_OPTIONS);
        return areEquivalent(node1, node2, diffStream);
    }
    
    public static boolean
    areEquivalent(File        xmlFile1,
                  String      xml2,
                  String      context2,
                  PrintStream diffStream)
    {
        String context1 = xmlFile1.getName();
        return areEquivalent(xmlFile1, context1, xml2, context2, diffStream);
    }
    
    public static boolean
    areEquivalent(File        xmlFile1,
                  String      context1,
                  String      xml2,
                  String      context2,
                  PrintStream diffStream)
    {
        TagNode node1 = XmlReader.xmlFileToRoot(xmlFile1, context1, READ_OPTIONS);
        TagNode node2 = XmlReader.xmlStringToRoot(xml2, context2, READ_OPTIONS);
        return areEquivalent(node1, node2, diffStream);
    }
    
    public static boolean
    areEquivalent(TagNode     node1,
                  TagNode     node2,
                  PrintStream diffStream)
    {
        if (!node1._id.equals(node2._id))
        {
            diffStream.println(nodeToString(node1) + " does not match " + nodeToString(node2));
            return false;
        }
        boolean areEquivalent = true;
        if (!haveEquivalentAttributes(node1, node2, diffStream))
        {
            areEquivalent = false;
        }
        boolean finished = false;
        while (!finished)
        {
            TextNode text1 = node1.nextTextNodeN();
            TextNode text2 = node2.nextTextNodeN();
            if (text1 == null || text2 == null)
            {
                finished = true;
                if (!(text1 == null && text2 == null))
                {
                    areEquivalent = false;
                    dealWithContentVsNoContentNode(node1, text1, node2, diffStream);
                }
            }
            else
            {
                boolean haveNonSpace1 = !isWhiteSpace(text1._text);
                boolean haveNonSpace2 = !isWhiteSpace(text2._text);
                if (!haveNonSpace1 || !haveNonSpace2)
                {
                    if (haveNonSpace1 || haveNonSpace2)
                    {
                        areEquivalent = false;
                        dealWithNonSpaceVsEmptyOrSpaceText(node1, text1, haveNonSpace1, node2, text2, diffStream);
                    }
                }
                else if (!text1._text.equals(text2._text))
                {
                    areEquivalent = false;
                    diffStream.println(
                            textToString(text1, node1) + " and " +
                            textToString(text2, node2) + " have different non-space text");
                }
            }
            if (!finished)
            {
                TagNode child1 = node1.nextChildN();
                TagNode child2 = node2.nextChildN();
                if (child1 == null && child2 == null)
                {
                    finished = true;
                }
                else if (child1 == null || child2 == null)
                {
                    areEquivalent = false;
                    finished = true;
                    TagNode finishedNode;
                    TagNode unfinishedNode;
                    TagNode child;
                    if (child1 == null)
                    {
                        finishedNode = node1;
                        unfinishedNode = node2;
                        child = child2;
                    }
                    else
                    {
                        finishedNode = node2;
                        unfinishedNode = node1;
                        child = child1;
                    }
                    diffStream.println(
                            "Where " + nodeToString(unfinishedNode) + " has child node " + child +
                            ", " + nodeToString(finishedNode) + " is finished");
                }
                else if (!areEquivalent(child1, child2, diffStream))
                {
                    areEquivalent = false;
                    finished = true;
                }
            }
        }
        return areEquivalent;
    }
    
    /**
     * Checks if both TagNodes have equivalent attributes.
     * Equivalent means equal except for order.
     * See also class header.
     * @param node1      First node whose attributes are to be compared
     * @param node2      Second node whose attributes are to be compared
     * @param diffStream The println(String) function is used to explain any differences
     * @return true if equivalent and false otherwise
     */
    public static boolean
    haveEquivalentAttributes(TagNode     node1,
                             TagNode     node2,
                             PrintStream diffStream)
    {
        boolean areEquivalent = true;
        XmlAttribute[] attributes1 = node1.sortedAttributes();
        XmlAttribute[] attributes2 = node2.sortedAttributes();
        if (attributes1.length != attributes2.length)
        {
            diffStream.println(
                    nodeToString(node1) + " has " + attributes1.length + " attributes and " +
                    nodeToString(node2) + " has " + attributes2.length);
            areEquivalent = false;
        }
        int cursor1 = 0;
        int cursor2 = 0;
        while (cursor1 < attributes1.length && cursor1 < attributes2.length)
        {
            if (cursor1 >= attributes1.length)
            {
                printExtraAttribute(node2, attributes2[cursor2], node1, diffStream);
                cursor2++;
                areEquivalent = false;
            }
            else if (cursor2 >= attributes1.length)
            {
                printExtraAttribute(node1, attributes1[cursor1], node2, diffStream);
                cursor1++;
                areEquivalent = false;
            }
            else
            {
                XmlAttribute attribute1 = attributes1[cursor1];
                XmlAttribute attribute2 = attributes2[cursor2];
                if (attribute1._name.equals(attribute2._name))
                {
                    if (!attribute1._value.equals(attribute2._value))
                    {
                        diffStream.println(
                                "Attribute " + attribute1._name + " has value " + attribute1._value +
                                " for " + nodeToString(node1) +
                                " but value " + attribute2._value + " for " + nodeToString(node2));
                        areEquivalent = false;
                    }
                    cursor1++;
                    cursor2++;
                }
                else if (attribute1.compareTo(attribute2) < 0)
                {
                    printExtraAttribute(node1, attribute1, node2, diffStream);
                    cursor1++;
                    areEquivalent = false;
                }
                else
                {
                    printExtraAttribute(node2, attribute2, node1, diffStream);
                    cursor2++;
                    areEquivalent = false;
                }
            }
        }
        return areEquivalent;
    }
    
    private static void
    dealWithContentVsNoContentNode(TagNode     node1,
                                   TextNode    text1,
                                   TagNode     node2,
                                   PrintStream diffStream)
    {
        TagNode nullTextNode;
        TagNode otherNode;
        if (text1 == null)
        {
            nullTextNode = node1;
            otherNode = node2;
        }
        else
        {
            nullTextNode = node2;
            otherNode = node1;
        }
        diffStream.println(nodeToString(nullTextNode) + " is a /> node but not " + nodeToString(otherNode));
    }
    
    private static void
    dealWithNonSpaceVsEmptyOrSpaceText(TagNode     node1,
                                       TextNode    text1,
                                       boolean     haveNonSpace1,
                                       TagNode     node2,
                                       TextNode    text2,
                                       PrintStream diffStream)
    {
        TextNode nonSpaceText;
        TextNode otherText;
        TagNode  nonSpaceTextContainingNode;
        TagNode  otherTextContainingNode;
        if (haveNonSpace1)
        {
            nonSpaceText = text1;
            otherText = text2;
            nonSpaceTextContainingNode = node1;
            otherTextContainingNode = node2;
        }
        else
        {
            nonSpaceText = text2;
            otherText = text1;
            nonSpaceTextContainingNode = node2;
            otherTextContainingNode = node1;
        }
        if (otherText._text.length() == 0)
        {
            diffStream.println(
                    textToString(nonSpaceText, nonSpaceTextContainingNode) +
                    " has non space characters but " +
                    textToString(otherText, otherTextContainingNode) + " is empty");
        }
        else
        {
            diffStream.println(
                    textToString(nonSpaceText, nonSpaceTextContainingNode) +
                    " has non space characters but " +
                    textToString(otherText, otherTextContainingNode) + " has only space characters");
        }
    }
    
    private static String
    nodeToString(TagNode node)
    {
        if (node._xmlContext == null)
        {
            return node.toString();
        }
        else
        {
            return node + " (" + node._xmlContext + ")";
        }
    }
    
    private static String
    textToString(TextNode text,
                 TagNode  textContainingNode)
    {
        if (textContainingNode._xmlContext == null)
        {
            return "text at " + text._textPos;
        }
        else
        {
            return "text at " + text._textPos + " (" + textContainingNode._xmlContext + ")";
        }
    }
    
    private static void
    printExtraAttribute(TagNode      extraAttributeNode,
                        XmlAttribute extraAttribute,
                        TagNode      otherNode,
                        PrintStream  diffStream)
    {
        diffStream.println(
            nodeToString(extraAttributeNode) + " has attribute " + extraAttribute +
            " which " + nodeToString(otherNode) + " does not have");
    }
    
    private static boolean
    isWhiteSpace(String s)
    {
        int sLength = s.length();
        for (int i = 0; i < sLength; i++)
        {
            char ch = s.charAt(i);
            if (!Character.isWhitespace(ch))
            {
                return false;
            }
        }
        return true;
    }
}
