/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.xml_basic.tool;

import java.io.File;

import rasmus_torkel.xml_basic.read.XmlReadOptions;
import rasmus_torkel.xml_basic.read.TagNode;
import rasmus_torkel.xml_basic.read.TextNode;
import rasmus_torkel.xml_basic.read.impl.XmlReader;
import rasmus_torkel.xml_basic.rw.XmlAttribute;
import rasmus_torkel.xml_basic.write.XmlSink;

/**
 * This class is both a library and a stand-alone tool.
 * 
 * @author Torkel
 *
 */
public class XmlReformatter
{
    /**
     * Main function for use as a stand-alone tool. Reformats to output file if
     * output file name is supplied and to standard output otherwise.
     * @param args Contains one or two elements.
     *             First element is the name of the input file. The second element
     *             is optional. If supplied, it is the name of the output file.
     */
    public static void
    main(String[] args)
    {
        if (args.length == 1)
        {
            String inFileName = args[0];
            File inFile = new File(inFileName);
            String outXml = XmlReformatter.reformat(inFile);
            System.out.print(outXml);
        }
        else if (args.length == 2)
        {
            String inFileName = args[0];
            String outFileName = args[1];
            File inFile = new File(inFileName);
            File outFile = new File(outFileName);
            XmlReformatter.reformat(inFile, outFile);
        }
        else
        {
            System.out.println("Need one or two arguments, inputFileName and optional outputFileName");
        }
    }
    
    public static String
    reformat(String inXml,
             String context)
    {
        TagNode node = XmlReader.xmlStringToRoot(inXml, context, XmlReadOptions.DEFAULT);
        XmlSink xmlSink = new XmlSink();
        reformat(node, xmlSink);
        return xmlSink.toString();
    }
    
    public static String
    reformat(File inFile)
    {
        TagNode node = XmlReader.xmlFileToRoot(inFile, XmlReadOptions.DEFAULT);
        XmlSink xmlSink = new XmlSink();
        reformat(node, xmlSink);
        return xmlSink.toString();
    }
    
    public static void
    reformat(File                       inFile,
             File                       outFile)
    {
        TagNode node = XmlReader.xmlFileToRoot(inFile, XmlReadOptions.DEFAULT);
        XmlSink xmlSink = new XmlSink(outFile);
        reformat(node, xmlSink);
    }
    
    public static void
    reformat(TagNode    node,
             XmlSink    xmlSink)
    {
        xmlSink.startNode(node._id);
        int attributeQty = node.attributeQty();
        for (int i = 0; i < attributeQty; i++)
        {
            XmlAttribute attribute = node.attribute(i);
            
            // NameSpace declaration is a little bit tricky, so there is some of special treatment
            // required. See explanations for the XmlSink functions.
            if (attribute._definedNameSpace == null)
            {
                xmlSink.sinkAttribute(attribute);
            }
            else
            {
                xmlSink.declareNameSpace(attribute._definedNameSpace);
            }
        }
        TextNode textNode = node.nextTextNodeN();
        boolean finished = textNode == null;
        if (finished)
        {
            xmlSink.closeTagAndNode();
            return;
        }
        int childQty = 0;
        while (!finished)
        {
            TagNode child = node.nextChildN();
            if (child == null)
            {
                finished = true;
            }
            else
            {
                childQty++;
                reformat(child, xmlSink);
            }
        }
        if (childQty == 0 && textNode != null)
        {
            xmlSink.sinkText(textNode._text);
        }
        xmlSink.closeNode();
    }
}
