package rasmus_torkel.xml_basic.write;

import java.io.File;
import java.util.ArrayList;

import rasmus_torkel.misc.StringUtil;
import rasmus_torkel.set.chars.CharSet;
import rasmus_torkel.set.strings.JustifyingStringSet;
import rasmus_torkel.text.encode.CharEncodeOptions;
import rasmus_torkel.text.encode.UnsupportedCharacterEncodeException;
import rasmus_torkel.text.write.IndentingLineSink;
import rasmus_torkel.text.write.IndentingLineSinkOptions;
import rasmus_torkel.xml_basic.rw.XmlChars;

/**
 * This is a cut down version of XmlSink which does not support attributes.
 * That implies that it also does not support name spaces.
 * <p>
 * While the regular XmlSink performs quite well, we want to see if we can
 * make it even faster.
 * <p>
 * Caution: Not yet fully tested.
 * 
 * @author Rasmus Torkel
 *
 */
public class NoAttrXmlSink
{
    public static class Options
    {
        public final CharSet                  _nameCharSet;
        public final CharEncodeOptions        _encodeOptions;
        public final IndentingLineSinkOptions _indentingLineSinkOptions;
        
        public
        Options(CharSet                  nameCharSet,
                CharEncodeOptions        encodeOptions,
                IndentingLineSinkOptions indentingLineSinkOptions)
        {
            _nameCharSet = nameCharSet;
            _encodeOptions = encodeOptions;
            _indentingLineSinkOptions = indentingLineSinkOptions;
        }
        
        public
        Options(CharEncodeOptions        encodeOptions,
                IndentingLineSinkOptions indentingLineSinkOptions)
        {
            this(encodeOptions._coreChars, encodeOptions, indentingLineSinkOptions);
        }
    }
    
    public static final Options DEFAULT_OPTIONS = new Options(CharEncodeOptions.DEFAULT, IndentingLineSinkOptions.DEFAULT);
    
    private final IndentingLineSink    _lineSink;
    private final XmlTextEncoder       _textNodeEncoder;
    //private final XmlTextEncoder       _singleQuotedEncoder;
    //private final XmlTextEncoder       _doubleQuotedEncoder;
    private final CharSet              _nameCharSet;
    
    private static class NodeRecord
    {
        public String                  _nodeName;
        public boolean                 _didChild;
    }
    
    private StringBuilder _pieceBuilder = new StringBuilder(64);
    
    private ArrayList<NodeRecord>   _nodeRecordAl = new ArrayList<NodeRecord>();
    private ArrayList<NodeRecord>   _spareNodeRecordCache = new ArrayList<NodeRecord>();
    
    /**
     * These are the possible states before after and in between user operations
     */
    private static enum State
    {
        beforeRoot,
        inStartTag,
        betweenTags,
        afterRoot
    }
    
    private State _state = State.beforeRoot;
    
    public
    NoAttrXmlSink(File file)
    {
        this(file, DEFAULT_OPTIONS);
    }
    
    public
    NoAttrXmlSink(File    file,
                  Options options)
    {
        this(new IndentingLineSink(file, options._indentingLineSinkOptions), options);
    }
    
    /**
     * Constructor for sink that builds up an XML string using default options. 
     */
    public
    NoAttrXmlSink()
    {
        this(DEFAULT_OPTIONS);
    }
    
    /**
     * Constructor for sink that builds up an XML string.
     * @param writeOptions
     */
    public
    NoAttrXmlSink(Options options)
    {
        this(new IndentingLineSink(options._indentingLineSinkOptions), options);
    }
    
    public
    NoAttrXmlSink(IndentingLineSink lineSink)
    {
        this(lineSink, DEFAULT_OPTIONS);
    }
    
    public
    NoAttrXmlSink(IndentingLineSink lineSink,
                  Options           options)
    {
        _lineSink = lineSink;
        _textNodeEncoder = XmlTextEncoder.makeTextNodeEncoder(options._encodeOptions);
        _nameCharSet = options._nameCharSet;
    }
    
    public void
    startNode(String nodeName)
    {
        checkNameAgainstSets(nodeName, "startNode");
        if (_state == State.inStartTag)
        {
            closeStartTag();
        }
        else if (_state == State.afterRoot)
        {
            throw new RuntimeException("startNode not valid for state " + _state);
        }
        if (_nodeRecordAl.size() > 0)
        {
            NodeRecord parentNodeRecord = _nodeRecordAl.get(_nodeRecordAl.size() - 1);
            if (!parentNodeRecord._didChild)
            {
                _lineSink.endLine();
                parentNodeRecord._didChild = true;
            }
        }
        _lineSink.sinkPiece("<");
        _lineSink.sinkPiece(nodeName);
        saveNodeName(nodeName);
        _lineSink.incrementLevel();
        _state = State.inStartTag;
    }
    
    private void
    checkNameAgainstSets(String              relativeName,
                         String              methodName)
    {
        checkNameAgainstSet(relativeName, XmlChars.NAME_SET, methodName);
        checkNameAgainstSet(relativeName, _nameCharSet, methodName);
    }
    
    private void
    checkNameAgainstSet(String              relativeName,
                        JustifyingStringSet nameSet,
                        String              methodName)
    {
        if (nameSet.isIn(relativeName))
        {
            return;
        }
        String namePhrase = StringUtil.namedValueToPlainPrintablePhrase("name", relativeName);
        String outReason = nameSet.outReason(relativeName);
        throw new RuntimeException(methodName + " called with " + namePhrase + " which " + outReason); 
    }
    
    private void
    closeStartTag()
    {
        _lineSink.sinkPiece(">");
        _state = State.betweenTags;
    }
    
    public void
    sinkText(String text) throws UnsupportedCharacterEncodeException
    {
        if (_state == State.inStartTag)
        {
            closeStartTag();
        }
        else if (_state == State.betweenTags)
        {
        }
        else
        {
            throw new RuntimeException("sinkText not valid for state " + _state);
        }
        resetPieceBuilder();
        try
        {
            _textNodeEncoder.encode(_pieceBuilder, text);
        }
        catch (UnsupportedCharacterEncodeException e)
        {
            String textSoFar = toString();
            throw new UnsupportedCharacterEncodeException(
                    e._ch, e._position, "between-XML-tag text", e._textPiece, textSoFar, e._valueCharSet);
        }
        String encodedText = _pieceBuilder.toString();
        _lineSink.sinkPiece(encodedText);
    }
    
    public void
    closeNode()
    {
        if (_state == State.inStartTag)
        {
            closeStartTag();
        }
        else if (_state == State.betweenTags)
        {
        }
        else
        {
            throw new RuntimeException("closeNode not valid for state " + _state);
        }
        _lineSink.decrementLevel();
        NodeRecord nodeRecord = popNodeRecord();
        _lineSink.sinkPiece("</");
        _lineSink.sinkPiece(nodeRecord._nodeName);
        _lineSink.sinkPiece(">");
        _lineSink.endLine();
        finaliseCloseNode();
    }
    
    private void
    finaliseCloseNode()
    {
        if (_lineSink.level() == 0)
        {
            _state = State.afterRoot;
            _lineSink.close();
        }
        else
        {
            _state = State.betweenTags;
        }
    }
    
    private NodeRecord
    popNodeRecord()
    {
        int nodeNameIndex = _nodeRecordAl.size() - 1;
        NodeRecord nodeRecord = _nodeRecordAl.get(nodeNameIndex);
        _nodeRecordAl.remove(nodeNameIndex);
        _spareNodeRecordCache.add(nodeRecord);
        return nodeRecord;
    }
    
    /**
     * Closes the start tag such that no end tag is required.
     */
    public void
    closeTagAndNode()
    {
        if (_state != State.inStartTag)
        {
            throw new RuntimeException("closeTagAndNode not valid for state " + _state);
        }
        int nodeNameIndex = _nodeRecordAl.size() - 1;
        _nodeRecordAl.remove(nodeNameIndex);
        _lineSink.decrementLevel();
        _lineSink.sinkPiece("/>");
        _lineSink.endLine();
        finaliseCloseNode();
    }
    
    public void
    endLine()
    {
        _lineSink.endLine();
    }
    
    public void
    abandon()
    {
        if (_state != State.afterRoot)
        {
            _lineSink.close();
        }
    }
    
    private void
    saveNodeName(String nodeName)
    {
        NodeRecord nodeRecord;
        int spareQty = _spareNodeRecordCache.size();
        if (spareQty > 0)
        {
            nodeRecord = _spareNodeRecordCache.remove(spareQty - 1);
        }
        else
        {
            nodeRecord = new NodeRecord();
        }
        nodeRecord._nodeName = nodeName;
        nodeRecord._didChild = false;
        _nodeRecordAl.add(nodeRecord);
    }
    
    private void
    resetPieceBuilder()
    {
        _pieceBuilder.delete(0, _pieceBuilder.length());
    }
    
    @Override
    public String
    toString()
    {
        return _lineSink.toString();
    }
    
    // Above here are the essential methods
    //--------------------------------------------------
    // Below here are the convenience methods
    
    public void
    sinkSimpleNode(String nodeName,
                   String contents)
    {
        if (contents == null)
        {
            return;
        }
        sinkSimpleStartTag(nodeName);
        sinkText(contents);
        closeNode();
    }
    
    public void
    sinkSimpleNode(String nodeName,
                   Object object)
    {
        sinkSimpleNode(nodeName, object.toString());
    }
    
    public void
    sinkSimpleNode(String nodeName,
                   int    contents)
    {
        sinkSimpleNode(nodeName, String.valueOf(contents));
    }
    
    public void
    sinkSimpleNode(String nodeName,
                   long   contents)
    {
        sinkSimpleNode(nodeName, String.valueOf(contents));
    }
    
    public void
    sinkSimpleNode(String nodeName,
                   float  contents)
    {
        sinkSimpleNode(nodeName, String.valueOf(contents));
    }
    
    public void
    sinkSimpleNode(String nodeName,
                   double contents)
    {
        sinkSimpleNode(nodeName, String.valueOf(contents));
    }
    
    public void
    sinkSimpleNode(String  nodeName,
                   boolean contents)
    {
        sinkSimpleNode(nodeName, String.valueOf(contents));
    }
    
    public void
    sinkOpenTag(String  nodeName)
    {
        startNode(nodeName);
        closeStartTag();
    }
    
    //--------------------------------------------------------------
    
    /**
     * Sinks array of objects by calling toString on them to get string representation.
     * @param relativeArrayName
     * @param arrayNameSpace
     * @param elementRelativeName
     * @param elementNameSpace
     * @param contents
     */
    public void
    sinkToStringArrayNode(String       relativeArrayName,
                          String       elementRelativeName,
                          Object[]     contents)
    {
        if (contents == null)
        {
            return;
        }
        sinkOpenTag(relativeArrayName);
        for (Object element : contents)
        {
            sinkSimpleNode(elementRelativeName, element);
        }
        closeNode();
    }
    
    public void
    sinkSimpleStartTag(String       relativeName)
    {
        startNode(relativeName);
        closeStartTag();
    }
}
