/*
 * Copyright Rasmus Torkel 2015
 * Author    Rasmus Torkel 2015
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.xml_basic.write;

import java.util.ArrayDeque;
import java.util.HashMap;
import java.util.HashSet;

import rasmus_torkel.xml_basic.rw.XmlAttribute;
import rasmus_torkel.xml_basic.rw.XmlNameSpace;

/**
 * This class tracks the use of name spaces and associated prefixes as name spaces go in and out of
 * scope. This class also resolves problems that arise when different name spaces have the same
 * preferred prefix. It is intended to assist classes which are writing or outputting XML.
 * <p>
 * I made a design decision to maintain prefix to name space mappings for the life of the class.
 * So of a name space is declared with a preferred and available prefix and then goes out of
 * scope, then even if it is later declared with a different preferred prefix, it will get the
 * prefix declared initially. This does not change the actual data, XML files which only differ
 * in the prefixes used for name spaces are equivalent.
 * <p>
 * What this class does to prefixes should make no difference to the programmer because the
 * programmer always specifies name spaces using the XmlNameSpace type.
 * 
 * @author Rasmus Torkel
 *
 */
class XmlNameSpaceTracker
{
    public static class NameSpaceAcceptance
    {
        public final XmlNameSpace _nameSpace;
        public final boolean      _needToDeclare;
        
        public
        NameSpaceAcceptance(XmlNameSpace nameSpace,
                            boolean      needToDeclare)
        {
            _nameSpace = nameSpace;
            _needToDeclare = needToDeclare;
        }
        
        public String
        nodeName(String relativeName)
        {
            return _nameSpace._prefix + ":" + relativeName;
        }
        
        public XmlAttribute
        nameSpaceAttribute()
        {
            return new XmlAttribute(_nameSpace._prefix,
                                    XmlNameSpace.XMLNS,
                                    _nameSpace._name);
        }
    }
    
    private static class TrackedXmlNameSpace
    {
        public final XmlNameSpace _nameSpace;
        public final int          _level;
        
        public
        TrackedXmlNameSpace(XmlNameSpace nameSpace,
                            int          level)
        {
            _nameSpace = nameSpace;
            _level = level;
        }
    }
    
    // Root level is 0, therefore before we see increment for root node, we are at level -1.
    private int                            _level = -1;
    
    // These two structures are only for in-scope name spaces.
    private ArrayDeque<TrackedXmlNameSpace>      _trackedNameSpaceAl =
            new ArrayDeque<XmlNameSpaceTracker.TrackedXmlNameSpace>();
    private HashSet<XmlNameSpace>               _inScopeNameSpaceHs = new HashSet<XmlNameSpace>();
    
    // These two mappings are not reduced as nodes go out of scope.
    private final HashMap<XmlNameSpace, XmlNameSpace> _nominalToNs =
            new HashMap<XmlNameSpace, XmlNameSpace>();
    private final HashMap<String, XmlNameSpace> _actualPrefixToNs =
            new HashMap<String, XmlNameSpace>();
    
    public
    XmlNameSpaceTracker()
    {
        notifyNameSpace(XmlNameSpace.XML);
        notifyNameSpace(XmlNameSpace.XMLNS);
    }
    
    /**
     * Calling this method informs the instance that we went into a node.
     */
    public void
    incrementLevel()
    {
        _level++;
    }
    
    /**
     * Calling this method informs the instance that we left a node.
     * This may result in name spaces going out of scope.
     */
    public void
    decrementLevel()
    {
        if (_level == -1)
        {
            throw new RuntimeException("decrementLevel called while outside root node");
        }
        _level--;
        for (;;)
        {
            TrackedXmlNameSpace trackedNameSpace = _trackedNameSpaceAl.peek();
            if (trackedNameSpace == null)
            {
                return;
            }
            else if (trackedNameSpace._level <= _level)
            {
                return;
            }
            else
            {
                _trackedNameSpaceAl.pop();
                _inScopeNameSpaceHs.remove(trackedNameSpace._nameSpace);
            }
        }
    }
    
    /**
     * Calling this method informs the instance that the invoker is referring to
     * an XML name space. This could be because a node with a name in a namespace
     * is to be written or because of a potential explicit declaration.
     * Be sure to be familiar with the returned name space acceptance.
     * @param notifiedNameSpace The name space that the invoker is going to refer to
     * @return name space acceptance. IMPORTANT 1: invoker should only write the name space
     * returned here as opposed to the original name space. IMPORTANT 2: Invoker should
     * declare name space (from acceptance) if and only if indicated in the acceptance.
     */
    public NameSpaceAcceptance
    notifyNameSpace(XmlNameSpace notifiedNameSpace)
    {
        XmlNameSpace savedNameSpace = _nominalToNs.get(notifiedNameSpace);
        boolean needToDeclare = !_inScopeNameSpaceHs.contains(notifiedNameSpace);
        if (savedNameSpace != null)
        {
            if (needToDeclare)
            {
                return trackNameSpace(savedNameSpace);
            }
            else
            {
                return new NameSpaceAcceptance(savedNameSpace, false);
            }
        }
        
        XmlNameSpace savableNameSpace;
        String actualPrefix = actualPrefixForDeclarableNameSpace(notifiedNameSpace);
        if (actualPrefix == notifiedNameSpace._prefix)
        {
            savableNameSpace = notifiedNameSpace;
        }
        else
        {
            savableNameSpace = new XmlNameSpace(actualPrefix, notifiedNameSpace._name);
        }
        _actualPrefixToNs.put(actualPrefix, savableNameSpace);
        _nominalToNs.put(savableNameSpace, savableNameSpace);
        return trackNameSpace(savableNameSpace);
    }
    
    private String
    actualPrefixForDeclarableNameSpace(XmlNameSpace notifiedNameSpace)
    {
        for (String prefixCandidate = notifiedNameSpace._prefix; ; prefixCandidate = prefixCandidate + "_")
        {
            XmlNameSpace savableNameSpace = _actualPrefixToNs.get(prefixCandidate);
            if (savableNameSpace == null)
            {
                return prefixCandidate;
            }
        }
    }
    
    private NameSpaceAcceptance
    trackNameSpace(XmlNameSpace savableNameSpace)
    {
        _inScopeNameSpaceHs.add(savableNameSpace);
        TrackedXmlNameSpace trackedNameSpace = new TrackedXmlNameSpace(savableNameSpace, _level);
        _trackedNameSpaceAl.push(trackedNameSpace);
        return new NameSpaceAcceptance(savableNameSpace, true);
    }
}
