/*
 * Copyright Rasmus Torkel 2014
 * Author    Rasmus Torkel 2014
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.xml_basic.write;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;

import rasmus_torkel.convention.ConventionConstants;
import rasmus_torkel.misc.StringUtil;
import rasmus_torkel.set.chars.CharSet;
import rasmus_torkel.set.strings.JustifyingStringSet;
import rasmus_torkel.text.encode.UnsupportedCharacterEncodeException;
import rasmus_torkel.text.write.IndentingLineSink;
import rasmus_torkel.xml_basic.rw.TagNodeId;
import rasmus_torkel.xml_basic.rw.XmlAttribute;
import rasmus_torkel.xml_basic.rw.XmlChars;
import rasmus_torkel.xml_basic.rw.XmlNameSpace;
import rasmus_torkel.xml_basic.write.XmlNameSpaceTracker.NameSpaceAcceptance;

/**
 * This is the class for writing XML. XML can be written various destinations, including files and strings.
 * <p>
 * When writing a complete XML file, this sink will close the underlying writer automatically when the root
 * node is completed. Of course, explicit closing is also possible using the function abandon.
 * <p>
 * The basic sequence of writing a node is startNode,sinkAttribute zero or more times,do various contents,closeNode.
 * Contents consist of more nodes or text (sinkText).
 *  
 * @author Rasmus Torkel
 *
 */
public class XmlSink
{
    private final IndentingLineSink    _lineSink;
    private final XmlTextEncoder       _textNodeEncoder;
    private final XmlTextEncoder       _singleQuotedEncoder;
    private final XmlTextEncoder       _doubleQuotedEncoder;
    private final AttributeLineOptions _attributeLineOptions;
    private final boolean              _attributeSortingOn;
    private final CharSet              _nameCharSet;
    private final XmlNameSpaceTracker  _nameSpaceTracker = new XmlNameSpaceTracker();
    
    private static class NodeRecord
    {
        public String                  _nodeName;
        public boolean                 _didChild;
    }
    
    private StringBuilder _pieceBuilder = new StringBuilder(64);
    
    private ArrayList<NodeRecord>   _nodeRecordAl = new ArrayList<NodeRecord>();
    private ArrayList<NodeRecord>   _spareNodeRecordCache = new ArrayList<NodeRecord>();
    private ArrayList<XmlAttribute> _attributeAl = new ArrayList<XmlAttribute>();
    
    /**
     * These are the possible states before after and in between user operations
     */
    private static enum State
    {
        beforeRoot,
        inStartTag,
        betweenTags,
        afterRoot
    }
    
    private State _state = State.beforeRoot;
    
    /**
     * Constructor for instance that write XML to a file
     * @param file destination for the XML
     */
    public
    XmlSink(File file)
    {
        this(file, XmlWriteOptions.DEFAULT);
    }

    /**
     * Constructor for instance that write XML to a file
     * @param file        destination for the XML
     * @param writeOption
     */
    public
    XmlSink(File            file,
            XmlWriteOptions writeOptions)
    {
        this(new IndentingLineSink(file, writeOptions._indentingLineSinkOptions), writeOptions);
    }
    
    /**
     * Constructor for sink that builds up an XML string using default options.
     * Later, you can obtain the XML string by calling the toString functions.
     */
    public
    XmlSink()
    {
        this(XmlWriteOptions.DEFAULT);
    }
    
    /**
     * Constructor for sink that builds up an XML string.
     * Later, you can obtain the XML string by calling the toString functions.
     * @param writeOptions
     */
    public
    XmlSink(XmlWriteOptions writeOptions)
    {
        this(new IndentingLineSink(writeOptions._indentingLineSinkOptions), writeOptions);
    }

    /**
     * Constructor.
     * @param lineSink     Make sure this doesn't already have stuff in it, or you could end up with invalid XNL.
     */
    public
    XmlSink(IndentingLineSink lineSink)
    {
        this(lineSink, XmlWriteOptions.DEFAULT);
    }
    
    /**
     * Constructor.
     * @param lineSink     Make sure this doesn't already have stuff in it, or you could end up with invalid XNL.
     * @param writeOptions
     */
    public
    XmlSink(IndentingLineSink lineSink,
            XmlWriteOptions   writeOptions)
    {
        _lineSink = lineSink;
        _textNodeEncoder = XmlTextEncoder.makeTextNodeEncoder(writeOptions._encodeOptions);
        _singleQuotedEncoder = XmlTextEncoder.makeSingleQuotedEncoder(writeOptions._encodeOptions);
        _doubleQuotedEncoder = XmlTextEncoder.makeDoubleQuotedEncoder(writeOptions._encodeOptions);
        _attributeLineOptions = writeOptions._attributeLineOptions;
        _attributeSortingOn = writeOptions._attributesSortingOn;
        _nameCharSet = writeOptions._nameCharSet;
    }
    
    /**
     * Starts a start tag for a node. Closes start tag for enclosing node if necessary.
     * @param relativeName
     * @param nameSpace
     */
    public void
    startNode(String       relativeName,
              XmlNameSpace nameSpace)
    {
        checkNameAgainstSets(relativeName, nameSpace, "startNode");
        if (_state == State.inStartTag)
        {
            closeStartTag();
        }
        else if (_state == State.afterRoot)
        {
            throw new RuntimeException("startNode not valid for state " + _state);
        }
        _nameSpaceTracker.incrementLevel();
        String nodeName;
        XmlNameSpaceTracker.NameSpaceAcceptance acceptance = null;
        if (nameSpace == null)
        {
            nodeName = relativeName;
        }
        else
        {
            acceptance = _nameSpaceTracker.notifyNameSpace(nameSpace);
            nodeName = acceptance.nodeName(relativeName);
        }
        if (_nodeRecordAl.size() > 0)
        {
            NodeRecord parentNodeRecord = _nodeRecordAl.get(_nodeRecordAl.size() - 1);
            if (!parentNodeRecord._didChild)
            {
                _lineSink.endLine();
                parentNodeRecord._didChild = true;
            }
        }
        _lineSink.sinkPiece("<");
        _lineSink.sinkPiece(nodeName);
        saveNodeName(nodeName);
        _lineSink.incrementLevel();
        _attributeAl.clear();
        if (acceptance != null && acceptance._needToDeclare)
        {
            XmlAttribute attr = acceptance.nameSpaceAttribute();
            internalDeclareAttribute(attr);
        }
        _state = State.inStartTag;
    }
    
    private void
    checkNameAgainstSets(String              relativeName,
                         XmlNameSpace        nameSpace,
                         String              methodName)
    {
        checkNameAgainstSet(relativeName, nameSpace, XmlChars.NAME_SET, methodName);
        checkNameAgainstSet(relativeName, nameSpace, _nameCharSet, methodName);
    }
    
    private void
    checkNameAgainstSet(String              relativeName,
                        XmlNameSpace        nameSpace,
                        JustifyingStringSet nameSet,
                        String              methodName)
    {
        if (nameSet.isIn(relativeName))
        {
            return;
        }
        String nameDesc = nameSpace == null ? "name" : "relative name";
        String namePhrase = StringUtil.namedValueToPlainPrintablePhrase(nameDesc, relativeName);
        String outReason = nameSet.outReason(relativeName);
        throw new RuntimeException(methodName + " called with " + namePhrase + " which " + outReason); 
    }
    
    private void
    closeStartTag()
    {
        internalSinkAttributes();
        _lineSink.sinkPiece(">");
        _state = State.betweenTags;
    }
    
    /**
     * Sinks the sort of text that goes between start and end tag.
     * If the state is such that a start tag is open, the start
     * tag will be automatically closed.
     * @param text Text to go between start and end tag.
     * @throws UnsupportedCharacterEncodeException
     */
    public void
    sinkText(String text) throws UnsupportedCharacterEncodeException
    {
        if (_state == State.inStartTag)
        {
            closeStartTag();
        }
        else if (_state == State.betweenTags)
        {
        }
        else
        {
            throw new RuntimeException("sinkText not valid for state " + _state);
        }
        resetPieceBuilder();
        try
        {
            _textNodeEncoder.encode(_pieceBuilder, text);
        }
        catch (UnsupportedCharacterEncodeException e)
        {
            String textSoFar = toString();
            throw new UnsupportedCharacterEncodeException(
                    e._ch, e._position, "between-XML-tag text", e._textPiece, textSoFar, e._valueCharSet);
        }
        String encodedText = _pieceBuilder.toString();
        _lineSink.sinkPiece(encodedText);
    }
    
    /**
     * Closes the currently open node. If called for the rootNode, the underlying Writer will automatically be closed.
     */
    public void
    closeNode()
    {
        if (_state == State.inStartTag)
        {
            closeStartTag();
        }
        else if (_state == State.betweenTags)
        {
        }
        else
        {
            throw new RuntimeException("closeNode not valid for state " + _state);
        }
        _lineSink.decrementLevel();
        NodeRecord nodeRecord = popNodeRecord();
        _lineSink.sinkPiece("</");
        _lineSink.sinkPiece(nodeRecord._nodeName);
        _lineSink.sinkPiece(">");
        _lineSink.endLine();
        finaliseCloseNode();
    }
    
    private void
    finaliseCloseNode()
    {
        _nameSpaceTracker.decrementLevel();
        if (_lineSink.level() == 0)
        {
            _state = State.afterRoot;
            _lineSink.close();
        }
        else
        {
            _state = State.betweenTags;
        }
    }
    
    private NodeRecord
    popNodeRecord()
    {
        int nodeNameIndex = _nodeRecordAl.size() - 1;
        NodeRecord nodeRecord = _nodeRecordAl.get(nodeNameIndex);
        _nodeRecordAl.remove(nodeNameIndex);
        _spareNodeRecordCache.add(nodeRecord);
        return nodeRecord;
    }
    
    /**
     * Closes the start tag such that no end tag is required.
     */
    public void
    closeTagAndNode()
    {
        if (_state != State.inStartTag)
        {
            throw new RuntimeException("closeTagAndNode not valid for state " + _state);
        }
        int nodeNameIndex = _nodeRecordAl.size() - 1;
        _nodeRecordAl.remove(nodeNameIndex);
        internalSinkAttributes();
        _lineSink.decrementLevel();
        _lineSink.sinkPiece("/>");
        _lineSink.endLine();
        finaliseCloseNode();
    }
    
    /**
     * Closes underlying writer. Do not continue writing after calling
     * this function.
     * <p>
     * This function is named as it is to avoid confusion with the
     * functions that close node and start tag.
     */
    public void
    abandon()
    {
        if (_state != State.afterRoot)
        {
            _lineSink.close();
        }
    }
    
    private void
    saveNodeName(String nodeName)
    {
        NodeRecord nodeRecord;
        int spareQty = _spareNodeRecordCache.size();
        if (spareQty > 0)
        {
            nodeRecord = _spareNodeRecordCache.remove(spareQty - 1);
        }
        else
        {
            nodeRecord = new NodeRecord();
        }
        nodeRecord._nodeName = nodeName;
        nodeRecord._didChild = false;
        _nodeRecordAl.add(nodeRecord);
    }
    
    /**
     * Sinks an attribute, other than a name space declaration.
     * <p>
     * Name space declaration is a tricky business. Because of prefix clashes,
     * a prefix might need to be assigned which is different to the one that
     * was specified.
     * As this function writes attributes exactly as specified, it is not
     * suitable for name space declarations.
     * 
     * @param attribute
     */
    public void
    sinkAttribute(XmlAttribute attribute)
    {
        if (_state != State.inStartTag)
        {
            throw new RuntimeException("sinkAttribute not valid for state " + _state);
        }
        checkNameAgainstSets(attribute._name, attribute._nameSpace, "sinkAttribute");
        if (!XmlChars.NAME_SET.isIn(attribute._name))
        {
            throw new RuntimeException("sinkAttribute called with name " + attribute._name + " which does not comply with name rules");
        }
        if (!_nameCharSet.isIn(attribute._name))
        {
            throw new RuntimeException(
                    "sinkAttribute called with name " + attribute._name +
                    " which is inconsistent with name char set " + _nameCharSet);
        }
        if (attribute._name.contains(":"))
        {
            throw new RuntimeException("sinkAttribute called with name " + attribute._name + " which contains :");
        }
        if (attribute._nameSpace != null)
        {
            if (attribute._nameSpace.equals(XmlNameSpace.XMLNS))
            {
                throw new RuntimeException(
                        "sinkAttribute used to declare name space " + attribute._value +
                        " with prefix " + attribute._name +
                        ", sinkAttribute must not be used to declare name space, use declareNameSpace instead");
            }
            NameSpaceAcceptance acceptance = _nameSpaceTracker.notifyNameSpace(attribute._nameSpace);
            if (acceptance._needToDeclare)
            {
                XmlAttribute nameSpaceAttribute = acceptance.nameSpaceAttribute();
                internalDeclareAttribute(nameSpaceAttribute);
            }
            if (!acceptance._nameSpace._prefix.equals(attribute._nameSpace._prefix))
            {
                attribute = new XmlAttribute(attribute._name, acceptance._nameSpace, attribute._value);
            }
        }
        for (XmlAttribute existingAttr : _attributeAl)
        {
            if (attribute.idEquals(existingAttr))
            {
                throw new RuntimeException("Two attributes with same identification: " + existingAttr + " and " + attribute);
            }
        }
        _attributeAl.add(attribute);
    }
    
    /**
     * Declares a name space (unless the name space is already declared).
     * <p>
     * There are a few things you should know about declaring name spaces:
     * <ul>
     *   <li>This class does not allow multiple name spaces with the same
     *       prefix, even if the name spaces don't have overlapping scope.
     *       It's not an error condition. This class will simply give you
     *       a different prefix. The prefix supplied with the name space
     *       should be considered a preferred prefix.</li>
     *   <li>Name space declaration is automatic. You sink a node or an
     *       attribute with a name space, and this class will declare the
     *       name space for you, if it hasn't been declared already.
     *       Therefore, there is never an absolute need to invoke it from
     *       outside this class.
     *       However, if there are separate subtrees that use it, it might
     *       be neater to declare it explicitly for a common ancestor node,
     *       rather than have this class declare it separately for each
     *       subtree that needs it.</li>
     * </ul>
     * @param nameSpace
     */
    public void
    declareNameSpace(XmlNameSpace nameSpace)
    {
        if (_state == State.beforeRoot)
        {
            throw new RuntimeException("Declaration of name space " + nameSpace + " before root");
        }
        if (_state == State.afterRoot)
        {
            throw new RuntimeException("Declaration of name space " + nameSpace + " after root");
        }
        if (nameSpace._name.equals(""))
        {
            throw new RuntimeException(
                    "Declaration of name space with empty name amounts to undeclaring name space which is not supported");
        }
        XmlNameSpaceTracker.NameSpaceAcceptance acceptance = _nameSpaceTracker.notifyNameSpace(nameSpace);
        if (acceptance != null && acceptance._needToDeclare)
        {
            XmlAttribute nameSpaceAttribute = acceptance.nameSpaceAttribute();
            internalDeclareAttribute(nameSpaceAttribute);
        }
    }
    
    private void
    internalDeclareAttribute(XmlAttribute attribute)
    {
        _attributeAl.add(attribute);
    }
    
    private String
    encodeAttributeValue(String attributeValue) throws UnsupportedCharacterEncodeException
    {
        resetPieceBuilder();
        encodeAttributeValue(_pieceBuilder, attributeValue);
        return _pieceBuilder.toString();
    }
    
    private void
    resetPieceBuilder()
    {
        _pieceBuilder.delete(0, _pieceBuilder.length());
    }
    
    private void
    encodeAttributeValue(StringBuilder buf,
                         String        attributeValue) throws UnsupportedCharacterEncodeException
    {
        int quotQty = countChar(attributeValue, '"');
        int aposQty = countChar(attributeValue, '\'');
        char quoteChar;
        XmlTextEncoder encoder;
        if (quotQty > aposQty)
        {
            quoteChar = '\'';
            encoder = _singleQuotedEncoder;
        }
        else
        {
            quoteChar = '"';
            encoder = _doubleQuotedEncoder;
        }
        buf.append(quoteChar);
        encoder.encode(buf, attributeValue);
        buf.append(quoteChar);
    }
    
    private static int
    countChar(String s,
              char   ch)
    {
        int qty = 0;
        int sLength = s.length();
        for (int i = 0; i < sLength; i++)
        {
            char sCh = s.charAt(i);
            if (sCh == ch)
            {
                qty++;
            }
        }
        return qty;
    }
    
    private void
    internalSinkAttributes()
    {
        boolean needSplitAttributes = false;
        if (_attributeLineOptions._alwaysSplitRootTag && _nodeRecordAl.size() == 1)
        {
            needSplitAttributes = true;
        }
        else if (_attributeAl.size() >= _attributeLineOptions._attributeQtySplitThreshold)
        {
            needSplitAttributes = true;
        }
        else if (_attributeLineOptions._alwaysSplitXmlnsTag)
        {
            for (XmlAttribute attribute : _attributeAl)
            {
                if (attribute._name.startsWith("xmlns:"))
                {
                    needSplitAttributes = true;
                }
            }
        }
        if (_attributeSortingOn)
        {
            XmlAttribute[] attributes = XmlAttribute.listToAttribute(_attributeAl);
            Arrays.sort(attributes);
            for (XmlAttribute attribute : attributes)
            {
                internalSinkAttribute(attribute, needSplitAttributes);
            }
        }
        else
        {
            for (XmlAttribute attribute : _attributeAl)
            {
                internalSinkAttribute(attribute, needSplitAttributes);
            }
        }
    }
    
    private void
    internalSinkAttribute(XmlAttribute attribute,
                          boolean      needSeparateLine)
    {
        if (needSeparateLine)
        {
            _lineSink.endLine();
            //_lineSink.sinkPiece(_lineSink._indent._indent);
        }
        else
        {
            _lineSink.sinkPiece(" ");
        }
        if (attribute._nameSpace != null)
        {
            _lineSink.sinkPiece(attribute._nameSpace._prefix);
            _lineSink.sinkPiece(":");
        }
        _lineSink.sinkPiece(attribute._name);
        _lineSink.sinkPiece("=");
        String encodedValue;
        try
        {
            encodedValue = encodeAttributeValue(attribute._value);
        }
        catch (UnsupportedCharacterEncodeException e)
        {
            throw new UnsupportedCharacterEncodeException(
                    e._ch, e._position,
                    "value of attribute named " + attribute._name, e._textPiece, toString(), e._valueCharSet);
        }
        _lineSink.sinkPiece(encodedValue);
    }
    
    /**
     * When building up a string as opposed to writing a file,
     * you can obtain the XML string through this function.
     */
    @Override
    public String
    toString()
    {
        return _lineSink.toString();
    }
    
    // Above here are the essential methods
    //--------------------------------------------------
    // Below here are the convenience methods
    
    /**
     * Sinks an attribute which is not in a name space.
     * @param name  of the attribute
     * @param value of the attribute
     * @throws UnsupportedCharacterEncodeException
     */
    public void
    sinkAttribute(String name,
                  String value) throws UnsupportedCharacterEncodeException
    {
        if (name.contains(":"))
        {
            throw new RuntimeException("sinkAttribute called such that there was no name space but name contains :, for attributes in name space, use a sinkAttribute function that allows name space");
        }
        if (value == null)
        {
            return;
        }
        XmlAttribute attribute = new XmlAttribute(name, value);
        sinkAttribute(attribute);
    }
    
    /**
     * Sinks an attribute which is not in a name space.
     * @param name  of the attribute
     * @param value object on which the toString function is called to obtain the value text.
     * @throws UnsupportedCharacterEncodeException
     */
    public void
    sinkAttribute(String name,
                  Object value)
    {
        if (value == null)
        {
            return;
        }
        String valueS = value.toString();
        sinkAttribute(name, valueS);
    }

    /**
     * Sinks an attribute which is not in a name space.
     * @param name  of the attribute
     * @param value of the attribute
     * @throws UnsupportedCharacterEncodeException
     */
    public void
    sinkAttribute(String name,
                  int    value)
    {
        String valueS = String.valueOf(value);
        sinkAttribute(name, valueS);
    }

    /**
     * Sinks an attribute which is not in a name space.
     * @param name  of the attribute
     * @param value of the attribute
     * @throws UnsupportedCharacterEncodeException
     */
    public void
    sinkAttribute(String name,
                  long   value)
    {
        String valueS = String.valueOf(value);
        sinkAttribute(name, valueS);
    }

    /**
     * Sinks an attribute which is not in a name space.
     * @param name  of the attribute
     * @param value of the attribute
     * @throws UnsupportedCharacterEncodeException
     */
    public void
    sinkAttribute(String name,
                  float  value)
    {
        String valueS = String.valueOf(value);
        sinkAttribute(name, valueS);
    }

    /**
     * Sinks an attribute which is not in a name space.
     * @param name  of the attribute
     * @param value of the attribute
     * @throws UnsupportedCharacterEncodeException
     */
    public void
    sinkAttribute(String name,
                  double value)
    {
        String valueS = String.valueOf(value);
        sinkAttribute(name, valueS);
    }

    /**
     * Sinks an attribute which is not in a name space.
     * @param name  of the attribute
     * @param value of the attribute
     * @throws UnsupportedCharacterEncodeException
     */
    public void
    sinkAttribute(String  name,
                  boolean value)
    {
        String valueS = String.valueOf(value);
        sinkAttribute(name, valueS);
    }
    
    /**
     * Makes a simple node, consisting of open tag, value, close tag.
     * @param nodeName name of the node
     * @param contents value between the tags. Can be null. if so, the method does nothing.
     */
    public void
    sinkSimpleNode(String nodeName,
                   String contents)
    {
        if (contents == null)
        {
            return;
        }
        startNode(nodeName);
        //sinkSimpleStartTag(nodeName);
        sinkText(contents);
        closeNode();
    }
    
    /**
     * Makes a simple node, consisting of open tag, value, close tag.
     * @param nodeName name of the node
     * @param object source of value, value obtained through toString function.
     *               object may be null. If so, the method does nothing.
     */
    public void
    sinkSimpleNode(String nodeName,
                   Object object)
    {
        if (object == null)
        {
            return;
        }
        sinkSimpleNode(nodeName, object.toString());
    }
    
    /**
     * Makes a simple node, consisting of open tag, value, close tag.
     * @param nodeName name of the node
     * @param contents value between the tags. Can be null. if so, the method does nothing.
     */
    public void
    sinkSimpleNode(String nodeName,
                   int    contents)
    {
        sinkSimpleNode(nodeName, String.valueOf(contents));
    }

    /**
     * Makes a simple node, consisting of open tag, value, close tag.
     * @param nodeName name of the node
     * @param contents value between the tags. Can be null. if so, the method does nothing.
     */
    public void
    sinkSimpleNode(String nodeName,
                   long   contents)
    {
        sinkSimpleNode(nodeName, String.valueOf(contents));
    }

    /**
     * Makes a simple node, consisting of open tag, value, close tag.
     * @param nodeName name of the node
     * @param contents value between the tags. Can be null. if so, the method does nothing.
     */
    public void
    sinkSimpleNode(String nodeName,
                   float  contents)
    {
        sinkSimpleNode(nodeName, String.valueOf(contents));
    }

    /**
     * Makes a simple node, consisting of open tag, value, close tag.
     * @param nodeName name of the node
     * @param contents value between the tags. Can be null. if so, the method does nothing.
     */
    public void
    sinkSimpleNode(String nodeName,
                   double contents)
    {
        sinkSimpleNode(nodeName, String.valueOf(contents));
    }

    /**
     * Makes a simple node, consisting of open tag, value, close tag.
     * @param nodeName name of the node
     * @param contents value between the tags. Can be null. if so, the method does nothing.
     */
    public void
    sinkSimpleNode(String  nodeName,
                   boolean contents)
    {
        sinkSimpleNode(nodeName, String.valueOf(contents));
    }

    /**
     * Makes a simple node, consisting of open tag, value, close tag.
     * @param nodeName name of the node
     * @param contents value between the tags. Can be null. if so, the method does nothing.
     */
    public void
    sinkSimpleNode(String  nodeName,
                   char    contents)
    {
        sinkSimpleNode(nodeName, String.valueOf(contents));
    }
    
    /**
     * Causes a structured object to write itself to our XML.
     * @param nodeName name of the node that will contain the structured object
     * @param object   the object that writes itself to our XML.
     */
    public void
    sinkNode(String          nodeName,
             XmlSinkWritable object)
    {
        sinkNode(nodeName, null, object);
    }

    /**
     * 
     * @param nodeName
     */
    /*
     * Discontinued. Use startNode
    public void
    sinkSimpleStartTag(String nodeName)
    {
        sinkSimpleStartTag(nodeName, null);
    }*/

    /**
     * Starts a node. Use closeNode to close.
     * @param nodeName
     */
    public void
    startNode(String  nodeName)
    {
        startNode(nodeName, null);
    }

    /* Discontinued. Use startNode
    public void
    sinkOpenTag(String  nodeName)
    {
        startNode(nodeName);
        closeStartTag();
    }*/
    
    /**
     * Records class of object in attribute {@see ConventionConstants.LABEL_FOR_CLASS_NAME_OF_OBJECT}.
     * This method exists to support higher level XML functions like creating objects from XML through factories
     * while allowing inheritance.
     * See <a href="http://rasmustorkel.com/rasmus_torkel_xml.html">overview document</a>.
     * 
     * @param object The object whose class we want to record in an attribute.
     */
    public void
    sinkClassOfObjectAttribute(Object object)
    {
        String className = object.getClass().getName();
        sinkAttribute(ConventionConstants.LABEL_FOR_CLASS_NAME_OF_OBJECT, className);
    }
    
    /**
     * Sinks a simple node consisting of open tag, contents, close tag.
     * @param relativeName for the node
     * @param nameSpace    for the node
     * @param contents     for between open and close tag
     */
    public void
    sinkSimpleNode(String       relativeName,
                   XmlNameSpace nameSpace,
                   String       contents)
    {
        if (contents == null)
        {
            return;
        }
        startNode(relativeName, nameSpace);
        sinkText(contents);
        closeNode();
    }

    /**
     * Sinks a simple node consisting of open tag, string representation of object, close tag.
     * @param relativeName for the node
     * @param nameSpace    for the node
     * @param object       String representation goes between open and close tag
     */
    public void
    sinkSimpleNode(String       relativeName,
                   XmlNameSpace nameSpace,
                   Object       object)
    {
        if (object == null)
        {
            return;
        }
        sinkSimpleNode(relativeName, nameSpace, object.toString());
    }

    /**
     * Sinks a simple node consisting of open tag, contents, close tag.
     * @param relativeName for the node
     * @param nameSpace    for the node
     * @param contents     for between open and close tag
     */
    public void
    sinkSimpleNode(String       relativeName,
                   XmlNameSpace nameSpace,
                   int          contents)
    {
        sinkSimpleNode(relativeName, nameSpace, String.valueOf(contents));
    }

    /**
     * Sinks a simple node consisting of open tag, contents, close tag.
     * @param relativeName for the node
     * @param nameSpace    for the node
     * @param contents     for between open and close tag
     */
    public void
    sinkSimpleNode(String       relativeName,
                   XmlNameSpace nameSpace,
                   long         contents)
    {
        sinkSimpleNode(relativeName, nameSpace, String.valueOf(contents));
    }

    /**
     * Sinks a simple node consisting of open tag, contents, close tag.
     * @param relativeName for the node
     * @param nameSpace    for the node
     * @param contents     for between open and close tag
     */
    public void
    sinkSimpleNode(String       relativeName,
                   XmlNameSpace nameSpace,
                   float        contents)
    {
        sinkSimpleNode(relativeName, nameSpace, String.valueOf(contents));
    }

    /**
     * Sinks a simple node consisting of open tag, contents, close tag.
     * @param relativeName for the node
     * @param nameSpace    for the node
     * @param contents     for between open and close tag
     */
    public void
    sinkSimpleNode(String       relativeName,
                   XmlNameSpace nameSpace,
                   double       contents)
    {
        sinkSimpleNode(relativeName, nameSpace, String.valueOf(contents));
    }

    /**
     * Sinks a simple node consisting of open tag, contents, close tag.
     * @param relativeName for the node
     * @param nameSpace    for the node
     * @param contents     for between open and close tag
     */
    public void
    sinkSimpleNode(String       relativeName,
                   XmlNameSpace nameSpace,
                   boolean      contents)
    {
        sinkSimpleNode(relativeName, nameSpace, String.valueOf(contents));
    }
    

    /**
     * Causes a structured object to write itself to our XML.
     * The object gets to choose the name of the node and possibly a name space.
     * @param object   the object that writes itself to our XML.
     */
    public void
    sinkNode(XmlSinkWritable object)
    {
        sinkNode(null, null, object);
    }

    /**
     * Causes a structured object to write itself to our XML.
     * @param relativeName of the node that will contain the structured object.
     *                     Can be null, then the object will chose it's own
     *                     node name/id.
     * @param nameSpace of the node that will contain the structured object
     * @param object    the object that writes itself to our XML.
     */
    public void
    sinkNode(String          relativeName,
             XmlNameSpace    nameSpace,
             XmlSinkWritable object)
    {
        if (relativeName == null)
        {
            TagNodeId id = object.naturalTagNodeId();
            startNode(id);
        }
        else
        {
            startNode(relativeName, nameSpace);
        }
        object.mostToXml(this, relativeName, nameSpace);
        closeNode();
    }
    
    //--------------------------------------------------------------
    
    /**
     * Sinks array of objects by calling toString on them to get string representation.
     * @param arrayNodeName
     * @param elementNodeName
     * @param contents
     */
    public void
    sinkToStringArrayNode(String   arrayNodeName,
                          String   elementNodeName,
                          Object[] contents)
    {
        sinkToStringArrayNode(arrayNodeName, null, elementNodeName, null, contents);
    }
    
    /**
     * Sinks array of objects by calling toString on them to get string representation.
     * @param arrayId
     * @param elementId
     * @param contents
     */
    public void
    sinkToStringArrayNode(TagNodeId arrayId,
                          TagNodeId elementId,
                          Object[]  contents)
    {
        sinkToStringArrayNode(
                arrayId._relativeName, arrayId._nameSpace, elementId._relativeName, elementId._nameSpace,
                contents);
    }
    
    /**
     * Sinks array of objects by calling toString on them to get string representation.
     * @param relativeArrayName
     * @param arrayNameSpace
     * @param elementRelativeName
     * @param elementNameSpace
     * @param contents
     */
    public void
    sinkToStringArrayNode(String       relativeArrayName,
                          XmlNameSpace arrayNameSpace,
                          String       elementRelativeName,
                          XmlNameSpace elementNameSpace,
                          Object[]     contents)
    {
        if (contents == null)
        {
            return;
        }
        startNode(relativeArrayName, arrayNameSpace);
        for (Object element : contents)
        {
            sinkSimpleNode(elementRelativeName, elementNameSpace, element);
        }
        closeNode();
    }
    
    //-----------------------------------------------------------------------
    
    /**
     * Sinks an array node of elements that write themselves to XML.
     * @param arrayId  for the array node. For the node names/ids of elements,
     *                 the natural id of the elements will be used.
     * @param objects  will write themselves to XML
     */
    public void
    sinkArrayNode(TagNodeId         arrayId,
                  XmlSinkWritable[] objects)
    {
        sinkArrayNode(arrayId._relativeName, arrayId._nameSpace, objects);
    }

    /**
     * Sinks an array node of elements that write themselves to XML.
     * @param arrayNodeName  for the array node. For the node names/ids of elements,
     *                       the natural id of the elements will be used.
     * @param objects  will write themselves to XML
     */
    public void
    sinkArrayNode(String            arrayNodeName,
                  XmlSinkWritable[] contents)
    {
        sinkArrayNode(arrayNodeName, (XmlNameSpace)null, contents);
    }

    /**
     * Sinks an array node of elements that write themselves to XML.
     * @param arrayRelativeName  for the array node. For the node names/ids of elements,
     *                           the natural id of the elements will be used.
     * @param arrayNameSpace     for the array node
     * @param objects  will write themselves to XML
     */
    public void
    sinkArrayNode(String            arrayRelativeName,
                  XmlNameSpace      arrayNameSpace,
                  XmlSinkWritable[] contents)
    {
        if (contents == null)
        {
            return;
        }
        startNode(arrayRelativeName, arrayNameSpace);
        for (XmlSinkWritable element : contents)
        {
            sinkNode(element);
        }
        closeNode();
    }
    
    //------------------------------------------------------------------------------
    
    /*
     * Discontinued. Use startNode.
     * 
     * Sinks a start tag without attributes, closing the startTag immediately.
     * Note: You can just call startNode, the only difference is that this function closes
     * the start tag immediately while startNode only does so when you write something
     * else (other than an attribute).
     * @param relativeName
     * @param nameSpace
     *
    public void
    sinkSimpleStartTag(String       relativeName,
                       XmlNameSpace nameSpace)
    {
        startNode(relativeName, nameSpace);
        closeStartTag();
    }*/
    
    /* Discontinued. use startNode
    public void
    sinkOpenTag(String       relativeName,
                XmlNameSpace nameSpace)
    {
        startNode(relativeName, nameSpace);
        closeStartTag();
    }*/
    
    /**
     * Sinks a simple node, consisting of open tag, contents, close tag
     * @param id       of the node
     * @param contents for between the tags
     */
    public void
    sinkSimpleNode(TagNodeId id,
                   String    contents)
    {
        sinkSimpleNode(id._relativeName, id._nameSpace, contents);
    }
    
    /**
     * Sinks a simple node, consisting of open tag, string representation of object, close tag.
     * @param id     of the node
     * @param object string representation goes between open and close tag
     */
    public void
    sinkSimpleNode(TagNodeId id,
                   Object    object)
    {
        if (object == null)
        {
            return;
        }
        sinkSimpleNode(id._relativeName, id._nameSpace, object.toString());
    }

    /**
     * Sinks a simple node, consisting of open tag, contents, close tag
     * @param id       of the node
     * @param contents for between the tags
     */
    public void
    sinkSimpleNode(TagNodeId id,
                   int       contents)
    {
        sinkSimpleNode(id._relativeName, id._nameSpace, String.valueOf(contents));
    }

    /**
     * Sinks a simple node, consisting of open tag, contents, close tag
     * @param id       of the node
     * @param contents for between the tags
     */
    public void
    sinkSimpleNode(TagNodeId id,
                   long      contents)
    {
        sinkSimpleNode(id._relativeName, id._nameSpace, String.valueOf(contents));
    }

    /**
     * Sinks a simple node, consisting of open tag, contents, close tag
     * @param id       of the node
     * @param contents for between the tags
     */
    public void
    sinkSimpleNode(TagNodeId id,
                   float     contents)
    {
        sinkSimpleNode(id._relativeName, id._nameSpace, String.valueOf(contents));
    }

    /**
     * Sinks a simple node, consisting of open tag, contents, close tag
     * @param id       of the node
     * @param contents for between the tags
     */
    public void
    sinkSimpleNode(TagNodeId id,
                   double    contents)
    {
        sinkSimpleNode(id._relativeName, id._nameSpace, String.valueOf(contents));
    }

    /**
     * Sinks a simple node, consisting of open tag, contents, close tag
     * @param id       of the node
     * @param contents for between the tags
     */
    public void
    sinkSimpleNode(TagNodeId id,
                   boolean   contents)
    {
        sinkSimpleNode(id._relativeName, id._nameSpace, String.valueOf(contents));
    }

    /**
     * Causes a structured object to write itself to our XML.
     * @param id       for the node that will contain the structured object
     * @param object   the object that writes itself to our XML.
     */
    public void
    sinkNode(TagNodeId     id,
             XmlSinkWritable object)
    {
        sinkNode(id._relativeName, id._nameSpace, object);
    }
    
    /*
     * Discontinued. Use startNode.
    public void
    sinkSimpleStartTag(TagNodeId id)
    {
        sinkSimpleStartTag(id._relativeName, id._nameSpace);
    }*/
    
    /**
     * Starts a node.
     * @param id for the node
     */
    public void
    startNode(TagNodeId id)
    {
        startNode(id._relativeName, id._nameSpace);
    }
    
    /* Discontinued. Use startNode
    public void
    sinkOpenTag(TagNodeId id)
    {
        startNode(id._relativeName, id._nameSpace);
        closeStartTag();
    }*/
    
    /**
     * Converts an object to an XML string using the object's XML functionality.
     * @param object for conversion to XML
     * @return The XML string representation of the object
     */
    public static String
    objectToString(XmlSinkWritable object)
    {
        return objectToString(object, XmlWriteOptions.DEFAULT);
    }

    /**
     * Converts an object to an XML string using the object's XML functionality.
     * @param object for conversion to XML
     * @return The XML string representation of the object
     */
    public static String
    objectToString(XmlSinkWritable object,
                   XmlWriteOptions options)
    {
        XmlSink sink = new XmlSink(options);
        try
        {
            sink.sinkNode(object);
            return sink.toString();
        }
        finally
        {
            sink.abandon();
        }
    }
    
    /**
     * Converts an array of objects to an XML string using the objects's XML
     * functionality.
     * @param arrayId for the array node
     * @param objects to be converted to XML
     * @return XML string representation of array
     */
    public static String
    arrayToString(TagNodeId         arrayId,
                  XmlSinkWritable[] objects)
    {
        return arrayToString(arrayId._relativeName, arrayId._nameSpace, objects);
    }
    
    /**
     * Converts an array of objects to an XML string using the objects's XML
     * functionality.
     * @param arrayName for the array node
     * @param objects   to be converted to XML
     * @return XML string representation of array
     */
    public static String
    arrayToString(String            arrayName,
                  XmlSinkWritable[] objects)
    {
        return arrayToString(arrayName, (XmlNameSpace)null, objects);
    }
    
    /**
     * Converts an array of objects to an XML string using the objects's XML
     * functionality.
     * @param arrayRelativeName for the array node
     * @param arrayNameSpace    for the array node
     * @param objects to be converted to XML
     * @return XML string representation of array
     */
    public static String
    arrayToString(String            arrayRelativeName,
                  XmlNameSpace      arrayNameSpace,
                  XmlSinkWritable[] objects)
    {
        return arrayToString(arrayRelativeName, arrayNameSpace, objects, XmlWriteOptions.DEFAULT);
    }

    /**
     * Converts an array of objects to an XML string using the objects's XML
     * functionality.
     * @param arrayId for the array node
     * @param objects to be converted to XML
     * @param options
     * @return XML string representation of array
     */
    public static String
    arrayToString(TagNodeId         arrayId,
                  XmlSinkWritable[] objects,
                  XmlWriteOptions   options)
    {
        return arrayToString(arrayId._relativeName, arrayId._nameSpace, objects, options);
    }

    /**
     * Converts an array of objects to an XML string using the objects's XML
     * functionality.
     * @param arrayName for the array node
     * @param objects   to be converted to XML
     * @param options
     * @return XML string representation of array
     */
    public static String
    arrayToString(String            arrayName,
                  XmlSinkWritable[] objects,
                  XmlWriteOptions   options)
    {
        return arrayToString(arrayName, (XmlNameSpace)null, objects, options);
    }

    /**
     * Converts an array of objects to an XML string using the objects's XML
     * functionality.
     * @param arrayRelativeName for the array node
     * @param arrayNameSpace    for the array node
     * @param objects to be converted to XML
     * @param options
     * @return XML string representation of array
     */
    public static String
    arrayToString(String            arrayRelativeName,
                  XmlNameSpace      arrayNameSpace,
                  XmlSinkWritable[] objects,
                  XmlWriteOptions   options)
    {
        XmlSink sink = new XmlSink(options);
        try
        {
            sink.sinkArrayNode(arrayRelativeName, arrayNameSpace, objects);
            return sink.toString();
        }
        finally
        {
            sink.abandon();
        }
    }
    
    /**
     * Makes an XML file which contains the object's XML representation,
     * utilising the object's XML functionality.
     * @param object  to be converted to XML
     * @param xmlFile will contain the XML
     */
    public static void
    objectToFile(XmlSinkWritable object,
                 File            xmlFile)
    {
        objectToFile(object, xmlFile, XmlWriteOptions.DEFAULT);
    }

    /**
     * Makes an XML file which contains the object's XML representation,
     * utilising the object's XML functionality.
     * @param object  to be converted to XML
     * @param xmlFile will contain the XML
     * @param options
     */
    public static void
    objectToFile(XmlSinkWritable object,
                 File            xmlFile,
                 XmlWriteOptions options)
    {
        XmlSink sink = new XmlSink(xmlFile, options);
        try
        {
            sink.sinkNode(object);
        }
        finally
        {
            sink.abandon();
        }
    }

    /**
     * Makes an XML file which contains the objects's XML representation,
     * utilising the objects's XML functionality.
     * @param arrayId for the arrayNode
     * @param objects to be converted to XML
     * @param xmlFile will contain the XML
     */
    public static void
    arrayToFile(TagNodeId         arrayId,
                XmlSinkWritable[] objects,
                File              xmlFile)
    {
        XmlSink sink = new XmlSink(xmlFile);
        try
        {
            sink.sinkArrayNode(arrayId._relativeName, arrayId._nameSpace, objects);
        }
        finally
        {
            sink.abandon();
        }
    }

    /**
     * Makes an XML file which contains the objects's XML representation,
     * utilising the objects's XML functionality.
     * @param arrayName for the arrayNode
     * @param objects to be converted to XML
     * @param xmlFile will contain the XML
     */
    public static void
    arrayToFile(String            arrayName,
                XmlSinkWritable[] objects,
                File              xmlFile)
    {
        XmlSink sink = new XmlSink(xmlFile);
        sink.sinkArrayNode(arrayName, (XmlNameSpace)null, objects);
    }

    /**
     * Makes an XML file which contains the objects's XML representation,
     * utilising the objects's XML functionality.
     * @param arrayRelativeName for the arrayNode
     * @param arrayNameSpace for the arrayNode
     * @param objects to be converted to XML
     * @param xmlFile will contain the XML
     */
    public static void
    arrayToFile(String            arrayRelativeName,
                XmlNameSpace      arrayNameSpace,
                XmlSinkWritable[] objects,
                File              xmlFile)
    {
        arrayToFile(arrayRelativeName, arrayNameSpace, objects, xmlFile, XmlWriteOptions.DEFAULT);
    }

    /**
     * Makes an XML file which contains the objects's XML representation,
     * utilising the objects's XML functionality.
     * @param arrayId for the arrayNode
     * @param objects to be converted to XML
     * @param xmlFile will contain the XML
     * @param options
     */
    public static void
    arrayToFile(TagNodeId         arrayId,
                XmlSinkWritable[] objects,
                File              xmlFile,
                XmlWriteOptions   options)
    {
        XmlSink sink = new XmlSink(xmlFile, options);
        try
        {
            sink.sinkArrayNode(arrayId._relativeName, arrayId._nameSpace, objects);
        }
        finally
        {
            sink.abandon();
        }
    }

    /**
     * Makes an XML file which contains the objects's XML representation,
     * utilising the objects's XML functionality.
     * @param arrayName for the arrayNode
     * @param objects to be converted to XML
     * @param xmlFile will contain the XML
     * @param options
     */
    public static void
    arrayToFile(String            arrayName,
                XmlSinkWritable[] objects,
                File              xmlFile,
                XmlWriteOptions   options)
    {
        XmlSink sink = new XmlSink(xmlFile, options);
        try
        {
            sink.sinkArrayNode(arrayName, (XmlNameSpace)null, objects);
        }
        finally
        {
            sink.abandon();
        }
    }

    /**
     * Makes an XML file which contains the objects's XML representation,
     * utilising the objects's XML functionality.
     * @param arrayRelativeName for the arrayNode
     * @param arrayNameSpace for the arrayNode
     * @param objects to be converted to XML
     * @param xmlFile will contain the XML
     * @param options
     */
    public static void
    arrayToFile(String            arrayRelativeName,
                XmlNameSpace      arrayNameSpace,
                XmlSinkWritable[] objects,
                File              xmlFile,
                XmlWriteOptions   options)
    {
        XmlSink sink = new XmlSink(xmlFile, options);
        try
        {
            sink.sinkArrayNode(arrayRelativeName, arrayNameSpace, objects);
        }
        finally
        {
            sink.abandon();
        }
    }
}
