/*
 * Copyright Rasmus Torkel 2013
 * Author    Rasmus Torkel 2013
 * 
 * You may only use this software subject to the terms and conditions which either you or your
 * organisation would have downloaded with this software and which should have been placed
 * somewhere accessible to you. These conditions are likely to
 * -- allow you or your organisation to determine whether you or your organisation
 *    may use this software for free or not and if not how to pay and how much;
 * -- prohibit or restrict republishing of this software and transmission to other organisations,
 *    even related companies within the same group;
 * -- prohibit the renaming of this class and the containing package.
 *    
 * Neither you nor your organisation may modify this header. However, you or your organisation
 * may modify anything below this header. If so, you or your organisation must place a comment
 * to that effect immediately below this header.
 */
package rasmus_torkel.xml_basic.write;

import rasmus_torkel.set.chars.CharSet;
import rasmus_torkel.set.chars.DefinedCharSets;
import rasmus_torkel.text.encode.CharEncodeOptions;
import rasmus_torkel.text.write.IndentingLineSinkOptions;
import rasmus_torkel.xml_basic.rw.XmlChars;

public class XmlWriteOptions
{
    public final CharSet                  _nameCharSet;
    public final CharEncodeOptions        _encodeOptions;
    public final AttributeLineOptions     _attributeLineOptions;
    public final boolean                  _attributesSortingOn;
    public final IndentingLineSinkOptions _indentingLineSinkOptions;
    
    public static final boolean DEFAULT_ATTRIBUTES_SORTING_ON = false;
    public static final CharEncodeOptions DEFAULT_CHAR_ENCODE_OPTIONS =
        new CharEncodeOptions(DefinedCharSets.ENGLISH_KEYBOARD_NON_CONTROL,
                              XmlChars.VALUE_CHAR_SET);
    public static final XmlWriteOptions DEFAULT =
        new XmlWriteOptions(DEFAULT_CHAR_ENCODE_OPTIONS,
                            AttributeLineOptions.DEFAULT,
                            DEFAULT_ATTRIBUTES_SORTING_ON,
                            IndentingLineSinkOptions.DEFAULT);
    
    public static final XmlWriteOptions ONE_LINE_PER_TAG = DEFAULT.diffAttributeLineOptions(AttributeLineOptions.NO_SPLIT);
    public static final XmlWriteOptions ONE_LINE_PER_ATTRIBUTE = DEFAULT.diffAttributeLineOptions(AttributeLineOptions.SPLIT);
    
    public static final XmlWriteOptions DONT_ESCAPE_SPACE = DEFAULT.diffEncodeOptions(CharEncodeOptions.DONT_ESCAPE_SPACE, false);
    
    public
    XmlWriteOptions(CharEncodeOptions        encodeOptions,
                    AttributeLineOptions     attributeLineOptions,
                    boolean                  attributesSortingOn,
                    IndentingLineSinkOptions indentingLineSinkOptions)
    {
        this(encodeOptions._coreChars, encodeOptions, attributeLineOptions, attributesSortingOn, indentingLineSinkOptions);
    }
    
    public
    XmlWriteOptions(CharSet                  nameCharSet,
                    CharEncodeOptions        encodeOptions,
                    AttributeLineOptions     attributeLineOptions,
                    boolean                  attributesSortingOn,
                    IndentingLineSinkOptions indentingLineSinkOptions)
    {
        _nameCharSet = nameCharSet;
        _encodeOptions = encodeOptions;
        _attributeLineOptions = attributeLineOptions;
        _attributesSortingOn = attributesSortingOn;
        _indentingLineSinkOptions = indentingLineSinkOptions;
    }
    
    @Override
    public String
    toString()
    {
        String sortStr = _attributesSortingOn ? "attributesSortingOn" : "attributesSortingOff";
        return "{ " + _encodeOptions + ", " +
                      _attributeLineOptions + ", " +
                      sortStr + ", " +
                      _indentingLineSinkOptions + " }";
    }
    
    public String
    toXml()
    {
        return toXml(this);
    }
    
    public String
    toXml(XmlWriteOptions optionsForConvertingOptions)
    {
        XmlSink xmlSink = new XmlSink(optionsForConvertingOptions);
        toXml(xmlSink, "XmlWriteOptions");
        return xmlSink.toString();
    }
    
    public void
    toXml(XmlSink xmlSink, String fieldName)
    {
        xmlSink.startNode(fieldName);
        xmlSink.sinkAttribute("type", getClass().getName());
        {
            xmlSink.startNode("encodeOptions");
            xmlSink.sinkAttribute("type", _encodeOptions.getClass().getName());
            xmlSink.sinkSimpleNode("coreChars", _encodeOptions._coreChars);
            xmlSink.sinkSimpleNode("allChars", _encodeOptions._allChars);
            xmlSink.sinkSimpleNode("spaceEscapeOption", _encodeOptions._spaceEscapeOption);
            xmlSink.closeNode();
        }
        _attributeLineOptions.toXml(xmlSink, "attributeLineOptions");
        xmlSink.sinkSimpleNode("attributesSortingOn", _attributesSortingOn);
        {
            xmlSink.startNode("indentingLineSinkOptions");
            xmlSink.sinkAttribute("type", _indentingLineSinkOptions.getClass().getName());
            xmlSink.sinkSimpleNode("indent", _indentingLineSinkOptions._indent);
            xmlSink.sinkSimpleNode("lineEndOption", _indentingLineSinkOptions._lineEndOption);
            xmlSink.closeNode();
        }
        xmlSink.closeNode();
    }
    
    public XmlWriteOptions
    diffNameCharSet(CharSet nameCharSet)
    {
        return new XmlWriteOptions(
                nameCharSet, _encodeOptions, _attributeLineOptions, _attributesSortingOn, _indentingLineSinkOptions);
    }
    
    public XmlWriteOptions
    diffAllCharSet(CharSet allCharSet)
    {
        CharEncodeOptions diffEncodeOptions = CharEncodeOptions.DEFAULT.diffAllChars(allCharSet);
        return diffEncodeOptions(diffEncodeOptions, false);
    }
    
    public XmlWriteOptions
    diffEncodeOptions(CharEncodeOptions diffEncodeOptions,
                      boolean           wantNameCharSetReplacement)
    {
        CharSet nameCharSet;
        if (wantNameCharSetReplacement)
        {
            nameCharSet = diffEncodeOptions._coreChars;
        }
        else
        {
            nameCharSet = _nameCharSet;
        }
        return new XmlWriteOptions(
                nameCharSet,
                diffEncodeOptions,
                _attributeLineOptions,
                _attributesSortingOn,
                _indentingLineSinkOptions);
    }
    
    public XmlWriteOptions
    diffAttributeLineOptions(AttributeLineOptions diffAttributeLineOptions)
    {
        return new XmlWriteOptions(
                _encodeOptions,
                diffAttributeLineOptions,
                _attributesSortingOn,
                _indentingLineSinkOptions);
    }
    
    public XmlWriteOptions
    diffAttributeSortOption(boolean diffAttributesSortingOn)
    {
        return new XmlWriteOptions(
                _encodeOptions,
                _attributeLineOptions,
                diffAttributesSortingOn,
                _indentingLineSinkOptions);
    }
    
    public XmlWriteOptions
    diffIndentingLineSinkOptions(IndentingLineSinkOptions diffIndentingLineSinkOptions)
    {
        return new XmlWriteOptions(
                _encodeOptions,
                _attributeLineOptions,
                _attributesSortingOn,
                diffIndentingLineSinkOptions);
    }
}
